/*++
This file contains a 'Sample Driver' and is licensed as such
under the terms of your license agreement with Intel or your
vendor.  This file may be modified by the user, subject to
the additional terms of the license agreement
--*/


/*++
Copyright (c)  2009-2018 Intel Corporation. All rights reserved
This software and associated documentation (if any) is furnished
under a license and may only be used or copied in accordance
with the terms of the license. Except as permitted by such
license, no part of this software or documentation may be
reproduced, stored in a retrieval system, or transmitted in any
form or by any means without the express written consent of
Intel Corporation.

Module Name:

FwUpdateLib.c

Abstract:

FW Update Local Library Code
--*/


#include <types.h>
#include "Fwu_Common.h"
#include "typedef.h"
#include "common.h"
#include "FWUpdateLib.h"

#include "Heci.h"

#include "cse_nvm_defs.h"
#include "cse_manifest.h"
#include "fw_caps_msgs.h"
#include "heci1_cse_gs1_def.h"

#include <Protocol/PciRootBridgeIo.h>
#include <Library/MemoryAllocationLib.h>
#include <Library/BaseMemoryLib.h>


//#define _DEBUG
#ifdef _DEBUG
#define DebugPrint Print
#else
#define DebugPrint(...)
#endif

#define MAX_ERROR_CODE (sizeof(ErrorMap)/sizeof(FwuCseErrorMap))

#define HBM_PROTOCOL_ADDRESS                 0       // HECI bus message protocol address
#define CLIENT_HOST_ADDRESS                  1       // HECI Client host address

///////////////////////////////// GUIDS  ////////////////////////////////////
#define LOCAL_FW_UPDATE_GUID \
{ \
    0x309dcde8, 0xccb1, 0x4062, 0x8f, 0x78, 0x60, 0x1, 0x15, 0xa3, 0x43, 0x27 \
}

#define HECI_FLOW_CTRL_MSG_LEN 8

#define MKHI_SUPPORT
#define HECI_ADDRESS_FIX_MKHI 7 // fixed HECI address for HCI client: heci_address_fix_mkhi \fw\src\os\include\heci_types.h

#define PARTITION_NAME_UPDT  0x54445055
#define FWUPD_SKIP_UPDT_SIZE 0
#define MANIFEST_BUFFER 0x1000
#define INVALID_DATA_FORMAT_VERSION 0

//New for SKL
#define GetTypedValue(type,value) (type)(*((type*)&value))

#define GetVersionMajor(value)  GetTypedValue(UINT16, *((UINT8*)value + 0))
#define GetVersionMinor(value)  GetTypedValue(UINT16, *((UINT8*)value + 2))
#define GetVersionHotfix(value) GetTypedValue(UINT16, *((UINT8*)value + 4))
#define GetVersionBuild(value)  GetTypedValue(UINT16, *((UINT8*)value + 6))

#ifndef MODULE_NAME_BUP_MANIFEST
#define MODULE_NAME_FTPR_MANIFEST "FTPR.man"
#endif

#ifndef MODULE_NAME_RBEP_MANIFEST
#define MODULE_NAME_RBEP_MANIFEST "RBEP.man"
#endif

#define MODULE_NAME_NFTP_MANIFEST "NFTP.man"
#define MODULE_NAME_PMCP_MANIFEST "PMCP.man"
#define MODULE_NAME_OEMP_MANIFEST "OEMP.man"
#define MODULE_NAME_WCOD_MANIFEST "WCOD.man"
#define MODULE_NAME_LOCL_MANIFEST "LOCL.man"
#define MODULE_NAME_MDMV_MANIFEST "MDMV.man"
#define MODULE_NAME_ISHC_MANIFEST "ISHC.man"
#define MODULE_NAME_IUNP_MANIFEST "IUNP.man"
#define MODULE_NAME_UPDT_MANIFEST "UPDT.man"

#define MODULE_NAME_FTPR "FTPR"
#define MODULE_NAME_NFTP "NFTP"
#define MODULE_NAME_RBEP "RBEP"
#define MODULE_NAME_PMCP "PMCP"
#define MODULE_NAME_OEMP "OEMP"
#define MODULE_NAME_WCOD "WCOD"
#define MODULE_NAME_LOCL "LOCL"
#define MODULE_NAME_MDMV "MDMV"
#define MODULE_NAME_ISHC "ISHC"
#define MODULE_NAME_IUNP "IUNP"

#define MAX_CPD_ENTRY_NAME_SIZE (12)

// Used for full FW Update number of expected partitions 3 (FTP, NFTP, RBEP)
#define NUM_OF_MANDATORY_PARTITIONS 3
#define NUM_ME_CLIENTS    2
#define PARTITION_NAME_UPDT  0x54445055

/** Disable FWU_GET_IPU_PT_ATTRB HECI message - Testing purposes only */
#define BYPASS_FWU_GET_IPU_PT_ATTRB (0)

#define PDT_SERVICE_PAYLOAD_SIZE  4096
#define PDT_UPDATE_DATA_TYPE 2

// Location of ME on PCI space
#if VP
#define ME_BUS  0
#define ME_DEV  14
#define ME_FUNC 0
#else
#define ME_BUS  0
#define ME_DEV  22
#define ME_FUNC 0
#endif

#define ERROR_NULL_CHECK      1
#define ME_PCI_FWSTATUS1_REG           0x40  // This is the offset for ME FW status register1
#define ME_PCI_FWSTATUS2_REG           0x48  // This is the offset for ME FW status register2
#define ME_PCI_FWSTATUS3_REG           0x60  // This is the offset for ME FW status register3
#define ME_PCI_FWSTATUS4_REG           0x64  // This is the offset for ME FW status register4
#define ME_PCI_FWSTATUS5_REG           0x68  // This is the offset for ME FW status register5
#define ME_PCI_FWSTATUS6_REG           0x6C  // This is the offset for ME FW status register6

typedef unsigned (*HeciDeviceIdCheckCallbackFn)(UINT32);

enum FwUpdPartitionName {
    FwUpdPartitionNameFtpr,
    FwUpdPartitionNameRbep,
    FwUpdPartitionNamePmcp,
    FwUpdPartitionNameOemp,
    FwUpdPartitionNameNftp,
    FwUpdPartitionNameLocl,
    FwUpdPartitionNameWcod,
    FwUpdPartitionNameMdmv,
    FwUpdPartitionNameIshc,
    FwUpdPartitionNameIunp,
    FwUpdPartitionNameUpdt,
    FwUpdPartitionNameFpsp,
    FwUpdPartitionNameCount
};

typedef enum _CLIENT_ID {
    FWU_CLIENT = 0,
} CLIENT_ID;

typedef struct {
    UINT8   ClientId;
    UINT8   HostAddress;
    UINT32  MaxMessageLength;
    BOOLEAN HaveFlowControl;
    BOOLEAN Connected;
} ME_CLIENT_PROPERTIES;

typedef struct _ImagePartition
{
    UINT32   PartitionName;    ///< Identifies the partition (FPT_PARTITION_NAME_xxx)
    UINT32   UpvVersion;
    UINT32   InstanceId;
    UINT32   OffsetFromFPT;    ///< Base of Partition from base of FPT.
    UINT32   PartitionLength;  ///< Length of partition in bytes
    UINT32 ImageOrder;
}ImagePartition;

typedef struct {
    UINT16  NumberOfElements;
    UINT16  DataType;
    UINT16  PayloadSize;
    UINT8   Payload[PDT_SERVICE_PAYLOAD_SIZE];
} PDT_UPDATE_DATA;

static BOOLEAN g_fwulib_heci_initialized = FALSE;

static const EFI_GUID  FWUGuid = LOCAL_FW_UPDATE_GUID;
static EFI_GUID  gEfiHeciProtocolGuid = EFI_HECI_PROTOCOL_GUID;
static ME_CLIENT_PROPERTIES MeClients[NUM_ME_CLIENTS];
static EFI_HECI_PROTOCOL* mHeci = NULL;
//static HW_SKU mHwSku = {0};
static FlashVersion mVer = {0};
static EFI_PCI_ROOT_BRIDGE_IO_PROTOCOL* gPciRootBrdgIo = NULL;
static fptt_ver1_t* g_fptt = NULL;

//static UINT32 mVendor = 0;
static UINT32 mChunkSize = 0;
static UINT32 mDataFormatVersion = 0;
static UINT32 mLastResetType = 0;
static UINT32 mLastFwUpdateStatus = 0;
static UINT16 mInterfaces = 0;
static UINT8* mpTxBuff = NULL; // general purpose message buffer
static UINT8* mpRxBuff = NULL; // general purpose message buffer
static BOOLEAN mInitialized = FALSE;

static CHAR8* FwUpdManifestFileNames[FwUpdPartitionNameCount] =
{
    MODULE_NAME_FTPR_MANIFEST,
    MODULE_NAME_NFTP_MANIFEST,
    MODULE_NAME_RBEP_MANIFEST,
    MODULE_NAME_PMCP_MANIFEST,
    MODULE_NAME_OEMP_MANIFEST,
    MODULE_NAME_WCOD_MANIFEST,
    MODULE_NAME_LOCL_MANIFEST,
    MODULE_NAME_MDMV_MANIFEST,
    MODULE_NAME_ISHC_MANIFEST,
    MODULE_NAME_IUNP_MANIFEST,
    MODULE_NAME_UPDT_MANIFEST
};

static FwuCseErrorMap ErrorMap[] = {
    {STATUS_UPDATE_SUCCESS, FWU_ERROR_SUCCESS},
    {STATUS_SIZE_ERROR, FWU_INVALID_IMG_LENGTH},
    {STATUS_ALREADY_STARTED, FWU_UPD_PROCESS},
    {STATUS_UPDATE_IMAGE_INVALID, FWU_IMG_HEADER},
    {STATUS_UPDATE_INTEGRITY_FAILURE, FWU_SGN_MISMATCH},
    {STATUS_UPDATE_SKU_MISMATCH, FWU_SKU_MISMATCH},
    {STATUS_UPDATE_FW_VERSION_MISMATCH, FWU_VER_MISMATCH},
    {STATUS_UPDATE_GENERAL_FAILURE, FWU_GENERAL},
    {STATUS_UPDATE_OUT_OF_RESOURCES, FWU_NO_MEMORY},
    {STATUS_UPDATE_AUDIT_POLICY_FAILURE, FWU_AUDIT_POLICY_FAILURE},
    {STATUS_UPDATE_ERROR_CREATING_FT, FWU_ERROR_CREATING_FT},
    {STATUS_UPDATE_SAL_NOTIFICATION_ERROR, FWU_SAL_NOTIFICATION_ERROR},
    {STATUS_UPDATE_IMG_LOADING, FWU_UPD_IMG_LOADING},
    {STATUS_UPDATE_IMG_AUTHENTICATING, FWU_UPD_IMG_AUTHENTICATING},
    {STATUS_UPDATE_IMG_PROCESSING, FWU_UPD_IMG_PROCESSING},
    {STATUS_UPDATE_CREATING_FT, FWU_UPD_CREATING_FT},
    {STATUS_UPDATE_UPDATING_CODE, FWU_UPD_UPDATING_CODE},
    {STATUS_UPDATE_UPDATING_NFT, FWU_UPD_UPDATING_NFT},
    {STATUS_UPDATE_FLASH_CODE_PARTITION_INVALID, FWU_FLASH_CODE_PARTITION_INVALID},
    {STATUS_UPDATE_FLASH_NFT_PARTITION_INVALID, FWU_FLASH_NFT_PARTITION_INVALID},
    {STATUS_UPDATE_ILLEGAL_IMAGE_LENGTH, FWU_ILLEGAL_IMAGE_LENGTH},
    {STATUS_UPDATE_NOT_READY, FWU_IME_NOT_READY},
    {STATUS_UPDATE_HOST_RESET_REQUIRED, FWU_HOST_RESET_REQUIRED},
    {STATUS_INVALID_GLUT, FWU_INVALID_GLUT},
    {STATUS_INVALID_OEM_ID, FWU_INVALID_OEM_ID},
    {STATUS_UPDATE_IMAGE_BLACKLISTED, FWU_UPDATE_IMAGE_BLACKLISTED},
    {STATUS_DOWNGRADE_NOT_ALLOWED_VCN_RESTRICTION, FWU_IMAGE_UNDER_VCN},
    {STATUS_UPDATE_IMAGE_VERSION_HISTORY_CHECK_FAILURE, FWU_IMAGE_VER_HIST_CHK_FAIL},
    {STATUS_UPDATE_DOWNGRADE_VETOED, FWU_DOWNGRADE_VETOED},
    {STATUS_UPDATE_WRITE_FILE_FAILURE, FWU_FW_WRITE_FILE_FAIL},
    {STATUS_UPDATE_READ_FILE_FAILURE, FWU_FW_READ_FILE_FAIL},
    {STATUS_UPDATE_DELETE_FILE_FAILURE, FWU_FW_DELETE_FILE_FAIL},
    {STATUS_UPDATE_PARTITION_LAYOUT_NOT_COMPATIBLE, FWU_PARTITION_LAYOUT_NOT_COMP},
    {STATUS_DOWNGRADE_NOT_ALLOWED_DATA_MISMATCHED, FWU_DOWNGRADE_NOT_ALLOWED_DATA_MISMATCH},
    {STATUS_UPDATE_FW_UPDATE_IS_DISABLED, FWU_LOCAL_DIS},
    {STATUS_UPDATE_PASSWORD_NOT_MATCHED, FWU_UPDATE_PASSWORD_NOT_MATCHED},
    {STATUS_UPDATE_PASSWORD_EXCEED_MAXIMUM_RETRY, FWU_UPDATE_PASSWORD_EXCEED_MAXIMUM_RETRY},
    {STATUS_UPDATE_INRECOVERY_MODE_RESTRICT_UPDATE_TO_ATTEMPTED_VERSION, FWU_UPDATE_PASSWORD_EXCEED_MAXIMUM_RETRY},
    {STATUS_UPDATE_UPV_VERSION_MISMATCHED, FWU_UPV_VERSION_MISMATCHED},
    {STATUS_UPDATE_INSTID_IS_NOT_EXPECTED_ID, FWU_INSTID_IS_NOT_EXPECTED_ID},
    {STATUS_UPDATE_INFO_NOT_AVAILABLE, FWU_INFO_NOT_AVAILABLE},
    {STATUS_UPDATE_REJ_IPU_FULL_UPDATE_NEEDED, FWU_REJ_IPU_FULL_UPDATE_NEEDED},
    {STATUS_UPDATE_IPU_NAMEID_NOT_FOUND, FWU_IPU_NAMEID_NOT_FOUND},
    {STATUS_UPDATE_RESTORE_POINT_INVALID, FWU_RESTORE_POINT_INVALID},
    {STATUS_UPDATE_RESTORE_POINT_VALID_BUT_NOT_LATEST, FWU_RESTORE_POINT_VALID_BUT_NOT_LATEST},
    {STATUS_UPDATE_RESTORE_POINT_OPERATION_NOT_ALLOWED, FWU_RESTORE_POINT_OPERATION_NOT_ALLOWED},
    {STATUS_DOWNGRADE_NOT_ALLOWED_SVN_RESTRICTION, FWU_DOWNGRADE_NOT_ALLOWED_SVN_RESTRICTION},
    {STATUS_DOWNGRADE_NOT_ALLOWED_VCN_RESTRICTION, FWU_DOWNGRADE_NOT_ALLOWED_VCN_RESTRICTION},
    {STATUS_INVALID_SVN, FWU_INVALID_SVN},
    {STATUS_UPDATE_OUT_OF_SVN_RESOURCES, FWU_OUT_OF_SVN_RESOURCES},
    {STATUS_UPDATE_REJECT_RESTORE_POINT_REQUEST_FLASH_IN_RECOVERY, FWU_REJECT_RESTORE_POINT_REQUEST_FLASH_IN_RECOVERY},
    {STATUS_UPDATE_REJECTED_BY_UPDATE_POLICY, FWU_REJECTED_BY_UPDATE_POLICY},
    {STATUS_UPDATE_REJECTED_INCOMPATIBLE_TOOL_USAGE, FWU_REJECTED_INCOMPATIBLE_TOOL_USAGE},
    {STATUS_UPDATE_REJECTED_CROSSPOINT_UPDATE_NOT_ALLOWED, FWU_REJECTED_CROSSPOINT_UPDATE_NOT_ALLOWED},
    {STATUS_UPDATE_REJECTED_CROSSHOTFIX_UPDATE_NOT_ALLOWED, FWU_REJECTED_CROSSHOTFIX_UPDATE_NOT_ALLOWED},
    {STATUS_UPDATE_REJECTED_CURRENT_FW_NOT_ELIGIBLE_FOR_UPDATE, FWU_REJECTED_CURRENT_FW_NOT_ELIGIBLE_FOR_UPDATE},
    {STATUS_UPDATE_REJECTED_WRONG_UPDATE_OPERATION, FWU_REJECTED_WRONG_UPDATE_OPERATION},
    {STATUS_UPDATE_REJECTED_WRONG_UPDATE_IMAGE_FOUND, FWU_REJECTED_WRONG_UPDATE_IMAGE_FOUND},
    {STATUS_UPDATE_REJECTED_IFR_UPDATE_NOT_ALLOWED, FWU_REJECTED_IFR_UPDATE_NOT_ALLOWED},
    {STATUS_UPDATE_FAILURE_OCCURRED_DURING_ROLLBACK, FWU_FAILURE_OCCURRED_DURING_ROLLBACK},
    {STATUS_UPDATE_REJECT_RESTORE_POINT_REQUEST_RESTART_NEEDED, FWU_REJECT_RESTORE_POINT_REQUEST_RESTART_NEEDED},
    {STATUS_UPDATE_PMC_INSTANCE, FWU_ERROR_PMC_INSTANCE},
    {STATUS_UPDATE_H_LP_MISMATCH, FWU_ERROR_H_LP_MISMATCH},
    {STATUS_UPDATE_REJECT_RESTORE_POINT_REQUEST_RESTART_NEEDED, FWU_ERROR_REJECT_RESTORE_POINT_REQUEST_RESTART_NEEDED},
    {STATUS_UPDATE_PMC_INSTANCE, FWU_ERROR_PMC_INSTANCE_MISMATCH},
    {STATUS_UPDATE_UPD_IMG_TOO_BIG, FWU_ERROR_UPD_IMG_TOO_BIG},
    {STATUS_UPDATE_INVALID_MANIFEST_SIZE, FWU_ERROR_INVALID_MANIFEST_SIZE},
    {STATUS_UPDATE_OPEN_LDR_VER_MAN_FAILED, FWU_ERROR_OPEN_LDR_VER_MAN_FAILED},
    {STATUS_UPDATE_OPEN_LDR_KEYS_FAILED, FWU_ERROR_OPEN_LDR_KEYS_FAILED},
    {STATUS_UPDATE_INSTALL_KEYS_FAILED, FWU_ERROR_INSTALL_KEYS_FAILED},
    {STATUS_UPDATE_UNINSTALL_KEYS_FAILED, FWU_ERROR_UNINSTALL_KEYS_FAILED},
    {STATUS_UPDATE_OPEN_LDR_INSTALL_FAILED, FWU_ERROR_OPEN_LDR_INSTALL_FAILED},
    {STATUS_UPDATE_INSTALL_FAILED, FWU_ERROR_INSTALL_FAILED},
    {STATUS_UPDATE_PREUPDATE_VFS, FWU_ERROR_PREUPDATE_VFS},
    {STATUS_UPDATE_SPAWN_FAILED, FWU_ERROR_SPAWN_FAILED},
    {STATUS_UPDATE_WAITPID_FAILED, FWU_ERROR_WAITPID_FAILED},
    {STATUS_UPDATE_PREUPDATE_FAILED, FWU_ERROR_PREUPDATE_FAILED},
    {STATUS_UPDATE_SKUMGR_FAILED, FWU_ERROR_SKUMGR_FAILED},
    {STATUS_UPDATE_CFGMGR_FAILED, FWU_ERROR_CFGMGR_FAILED},
    {STATUS_UPDATE_MAN_NOT_FOUND, FWU_ERROR_MAN_NOT_FOUND},
    {STATUS_UPDATE_CRYPTO_FAILED, FWU_ERROR_CRYPTO_FAILED},
    {STATUS_UPDATE_VER_MAN_FAILED_FTPR, FWU_ERROR_VER_MAN_FAILED_FTPR},
    {STATUS_UPDATE_VER_MAN_FAILED_NFTP, FWU_ERROR_VER_MAN_FAILED_NFTP},
    {STATUS_UPDATE_VER_MAN_FAILED_DLMP, FWU_ERROR_VER_MAN_FAILED_DLMP},
    {STATUS_UPDATE_VER_MAN_FAILED_RBEP, FWU_ERROR_VER_MAN_FAILED_RBEP},
    {STATUS_UPDATE_VER_MAN_FAILED_PMCP, FWU_ERROR_VER_MAN_FAILED_PMCP},
    {STATUS_UPDATE_VER_MAN_FAILED_OEMP, FWU_ERROR_VER_MAN_FAILED_OEMP},
    {STATUS_UPDATE_VER_MAN_FAILED_WCOD, FWU_ERROR_VER_MAN_FAILED_WCOD},
    {STATUS_UPDATE_VER_MAN_FAILED_LOCL, FWU_ERROR_VER_MAN_FAILED_LOCL},
    {STATUS_UPDATE_VER_MAN_FAILED_PAVP, FWU_ERROR_VER_MAN_FAILED_PAVP},
    {STATUS_UPDATE_VER_MAN_FAILED_IOMP, FWU_ERROR_VER_MAN_FAILED_IOMP},
    {STATUS_UPDATE_VER_MAN_FAILED_MGPP, FWU_ERROR_VER_MAN_FAILED_MGPP},
    {STATUS_UPDATE_VER_MAN_FAILED_TBTP, FWU_ERROR_VER_MAN_FAILED_TBTP},
    {STATUS_UPDATE_VER_MAN_FAILED_ISHC, FWU_ERROR_VER_MAN_FAILED_ISHC},
    {STATUS_UPDATE_VER_MAN_FAILED_IUNP, FWU_ERROR_VER_MAN_FAILED_IUNP},
    {STATUS_UPDATE_VER_MAN_FAILED_UPDT, FWU_ERROR_VER_MAN_FAILED_UPDT},
    {STATUS_UPDATE_GET_EXT_FAILED_FTPR, FWU_ERROR_GET_EXT_FAILED_FTPR},
    {STATUS_UPDATE_GET_EXT_FAILED_NFTP, FWU_ERROR_GET_EXT_FAILED_NFTP},
    {STATUS_UPDATE_GET_EXT_FAILED_DLMP, FWU_ERROR_GET_EXT_FAILED_DLMP},
    {STATUS_UPDATE_GET_EXT_FAILED_RBEP, FWU_ERROR_GET_EXT_FAILED_RBEP},
    {STATUS_UPDATE_GET_EXT_FAILED_PMCP, FWU_ERROR_GET_EXT_FAILED_PMCP},
    {STATUS_UPDATE_GET_EXT_FAILED_OEMP, FWU_ERROR_GET_EXT_FAILED_OEMP},
    {STATUS_UPDATE_GET_EXT_FAILED_WCOD, FWU_ERROR_GET_EXT_FAILED_WCOD},
    {STATUS_UPDATE_GET_EXT_FAILED_LOCL, FWU_ERROR_GET_EXT_FAILED_LOCL},
    {STATUS_UPDATE_GET_EXT_FAILED_PAVP, FWU_ERROR_GET_EXT_FAILED_PAVP},
    {STATUS_UPDATE_GET_EXT_FAILED_IOMP, FWU_ERROR_GET_EXT_FAILED_IOMP},
    {STATUS_UPDATE_GET_EXT_FAILED_MGPP, FWU_ERROR_GET_EXT_FAILED_MGPP},
    {STATUS_UPDATE_GET_EXT_FAILED_TBTP, FWU_ERROR_GET_EXT_FAILED_TBTP},
    {STATUS_UPDATE_GET_EXT_FAILED_ISHC, FWU_ERROR_GET_EXT_FAILED_ISHC},
    {STATUS_UPDATE_GET_EXT_FAILED_IUNP, FWU_ERROR_GET_EXT_FAILED_IUNP},
    {STATUS_UPDATE_GET_EXT_FAILED_UPDT, FWU_ERROR_GET_EXT_FAILED_UPDT},
    {STATUS_UPDATE_INTEGRITY_FAILED_FTPR, FWU_ERROR_INTEGRITY_FAILED_FTPR},
    {STATUS_UPDATE_INTEGRITY_FAILED_NFTP, FWU_ERROR_INTEGRITY_FAILED_NFTP},
    {STATUS_UPDATE_INTEGRITY_FAILED_DLMP, FWU_ERROR_INTEGRITY_FAILED_DLMP},
    {STATUS_UPDATE_INTEGRITY_FAILED_RBEP, FWU_ERROR_INTEGRITY_FAILED_RBEP},
    {STATUS_UPDATE_INTEGRITY_FAILED_PMCP, FWU_ERROR_INTEGRITY_FAILED_PMCP},
    {STATUS_UPDATE_INTEGRITY_FAILED_OEMP, FWU_ERROR_INTEGRITY_FAILED_OEMP},
    {STATUS_UPDATE_INTEGRITY_FAILED_WCOD, FWU_ERROR_INTEGRITY_FAILED_WCOD},
    {STATUS_UPDATE_INTEGRITY_FAILED_LOCL, FWU_ERROR_INTEGRITY_FAILED_LOCL},
    {STATUS_UPDATE_INTEGRITY_FAILED_PAVP, FWU_ERROR_INTEGRITY_FAILED_PAVP},
    {STATUS_UPDATE_INTEGRITY_FAILED_IOMP, FWU_ERROR_INTEGRITY_FAILED_IOMP},
    {STATUS_UPDATE_INTEGRITY_FAILED_MGPP, FWU_ERROR_INTEGRITY_FAILED_MGPP},
    {STATUS_UPDATE_INTEGRITY_FAILED_TBTP, FWU_ERROR_INTEGRITY_FAILED_TBTP},
    {STATUS_UPDATE_INTEGRITY_FAILED_ISHC, FWU_ERROR_INTEGRITY_FAILED_ISHC},
    {STATUS_UPDATE_INTEGRITY_FAILED_IUNP, FWU_ERROR_INTEGRITY_FAILED_IUNP},
    {STATUS_UPDATE_INTEGRITY_FAILED_UPDT, FWU_ERROR_INTEGRITY_FAILED_UPDT},
    {STATUS_UPDATE_PART_SIZE                    ,FWU_ERROR_PART_SIZE},
    {STATUS_UPDATE_BACKUP_OUTSIDE_NFTP          ,FWU_ERROR_BACKUP_OUTSIDE_NFTP},
    {STATUS_UPDATE_MAX_IUPS                     ,FWU_ERROR_MAX_IUPS},
    {STATUS_UPDATE_NOT_IUP                      ,FWU_ERROR_NOT_IUP},
    {STATUS_UPDATE_IUP_MISSING_UPDATE           ,FWU_ERROR_IUP_MISSING_UPDATE},
    {STATUS_UPDATE_PMC_MISSING_UPDATE           ,FWU_ERROR_PMC_MISSING_UPDATE},
    {STATUS_UPDATE_NOT_PARTIAL_IUP              ,FWU_ERROR_NOT_PARTIAL_IUP},
    {STATUS_UPDATE_PARTIAL_TCSS                 ,FWU_ERROR_PARTIAL_TCSS},
    {STATUS_UPDATE_FTPR_VER                     ,FWU_ERROR_FTPR_VER},
    {STATUS_UPDATE_FTPR_SVN                     ,FWU_ERROR_FTPR_SVN},
    {STATUS_UPDATE_FTPR_VCN                     ,FWU_ERROR_FTPR_VCN},
    {STATUS_UPDATE_FTPR_VER_MAJOR               ,FWU_ERROR_FTPR_VER_MAJOR},
    {STATUS_UPDATE_IUP_SVN                      ,FWU_ERROR_IUP_SVN},
    {STATUS_UPDATE_IUP_VCN                      ,FWU_ERROR_IUP_VCN},
    {STATUS_UPDATE_IMAGE_LEN                    ,FWU_ERROR_IMAGE_LEN},
    {STATUS_UPDATE_PV_BIT                       ,FWU_ERROR_PV_BIT},
    {STATUS_UPDATE_REVENUE                      ,FWU_ERROR_REVENUE},
    {STATUS_UPDATE_SVN_UPGRADE                  ,FWU_ERROR_SVN_UPGRADE},
    {STATUS_UPDATE_SVN_HOTFIX                   ,FWU_ERROR_SVN_HOTFIX},
    {STATUS_UPDATE_API_VER_MAJOR_FTPR           ,FWU_ERROR_API_VER_MAJOR_FTPR},
    {STATUS_UPDATE_API_VER_MAJOR_NFTP           ,FWU_ERROR_API_VER_MAJOR_NFTP},
    {STATUS_UPDATE_API_VER_MAJOR_DLMP           ,FWU_ERROR_API_VER_MAJOR_DLMP},
    {STATUS_UPDATE_API_VER_MAJOR_RBEP           ,FWU_ERROR_API_VER_MAJOR_RBEP},
    {STATUS_UPDATE_API_VER_MAJOR_PMCP           ,FWU_ERROR_API_VER_MAJOR_PMCP},
    {STATUS_UPDATE_API_VER_MAJOR_OEMP           ,FWU_ERROR_API_VER_MAJOR_OEMP},
    {STATUS_UPDATE_API_VER_MAJOR_WCOD           ,FWU_ERROR_API_VER_MAJOR_WCOD},
    {STATUS_UPDATE_API_VER_MAJOR_LOCL           ,FWU_ERROR_API_VER_MAJOR_LOCL},
    {STATUS_UPDATE_API_VER_MAJOR_PAVP           ,FWU_ERROR_API_VER_MAJOR_PAVP},
    {STATUS_UPDATE_API_VER_MAJOR_IOMP           ,FWU_ERROR_API_VER_MAJOR_IOMP},
    {STATUS_UPDATE_API_VER_MAJOR_MGPP           ,FWU_ERROR_API_VER_MAJOR_MGPP},
    {STATUS_UPDATE_API_VER_MAJOR_TBTP           ,FWU_ERROR_API_VER_MAJOR_TBTP},
    {STATUS_UPDATE_API_VER_MAJOR_ISHC           ,FWU_ERROR_API_VER_MAJOR_ISHC},
    {STATUS_UPDATE_API_VER_MAJOR_IUNP           ,FWU_ERROR_API_VER_MAJOR_IUNP},
    {STATUS_UPDATE_API_VER_MAJOR_UPDT           ,FWU_ERROR_API_VER_MAJOR_UPDT},
    {STATUS_UPDATE_IUP_MISSING_FLASH            ,FWU_ERROR_IUP_MISSING_FLASH},
    {STATUS_UPDATE_PARTITION_NOT_FOUND          ,FWU_ERROR_PARTITION_NOT_FOUND},
    {STATUS_UPDATE_ENGINEERING_MISMATCH         ,FWU_ERROR_ENGINEERING_MISMATCH},
    {STATUS_UPDATE_OEMP_MISSING                 ,FWU_ERROR_OEMP_MISSING},
    {STATUS_UPDATE_IUPS_NOT_COMPATIBLE          ,FWU_ERROR_IUPS_NOT_COMPATIBLE},
    {STATUS_UPDATE_OEMP_IN_UPDATE               ,FWU_ERROR_OEMP_IN_UPDATE},
    {STATUS_UPDATE_WRONG_IUP                    ,FWU_ERROR_WRONG_IUP},
    {STATUS_UPDATE_IMAGE_IUP_SIZE               ,FWU_ERROR_IMAGE_IUP_SIZE},
    {STATUS_UPDATE_OPEN_IUP                     ,FWU_ERROR_OPEN_IUP},
    {STATUS_UPDATE_SPI_IUP                      ,FWU_ERROR_SPI_IUP},
    {STATUS_UPDATE_RULE_FILE_INVALID            ,FWU_ERROR_RULE_FILE_INVALID},
    {STATUS_UPDATE_PWDMGR_FAILED                ,FWU_ERROR_PWDMGR_FAILED},
    {STATUS_UPDATE_PWR_FAILED                   ,FWU_ERROR_PWR_FAILED},
    {STATUS_UPDATE_SPI_FAILED                   ,FWU_ERROR_SPI_FAILED},
    {STATUS_UPDATE_RESTORE_POINT_ALREADY_STARTED,FWU_ERROR_RESTORE_POINT_ALREADY_STARTED},
    {STATUS_UPDATE_RESTORE_POINT_OFFSET_INVALID ,FWU_ERROR_RESTORE_POINT_OFFSET_INVALID},
    {STATUS_UPDATE_WRONG_HECI_MSG_LENGTH        ,FWU_ERROR_WRONG_HECI_MSG_LENGTH},
    {STATUS_UPDATE_ENV_INVALID                  ,FWU_ERROR_ENV_INVALID},
    {STATUS_UPDATE_WRONG_DATA_OPERATION         ,FWU_ERROR_WRONG_DATA_OPERATION},
    {STATUS_UPDATE_DATA_LENGTH_INVALID          ,FWU_ERROR_DATA_LENGTH_INVALID},
    {STATUS_UPDATE_WRONG_END_OPERATION          ,FWU_ERROR_WRONG_END_OPERATION},
    {STATUS_UPDATE_FLUSH_NVM_ERR                ,FWU_ERROR_FLUSH_NVM_ERR},
    {STATUS_UPDATE_IPK_FAILURE                  ,FWU_ERROR_IPK_FAILURE},
    {STATUS_UPDATE_PG_FAILURE                   ,FWU_ERROR_PG_FAILURE},
    {STATUS_UPDATE_INVALID_FLASH_IMAGE          ,FWU_ERROR_INVALID_FLASH_IMAGE},
    {STATUS_UPDATE_INVALID_UPDATE_IMAGE         ,FWU_ERROR_INVALID_UPDATE_IMAGE},
    {STATUS_SVN_TCB_ARB                         ,FWU_ERROR_SVN_TCB_ARB},
    {STATUS_UPDATE_LAST_STATUS_CODE, FWU_LAST_STATUS_CODE}
};

static EFI_STATUS GetFwVersion();
static EFI_STATUS ConnectFWU();
static EFI_STATUS DisConnectFWU ();

/*++
********************************************************************************
**
** FUNCTION:
**   ReadFwStatusRegValue( UINT32 * pFwStatusRegiset
**
** DESCRIPTION:
**   This function returns the sp.
**
** ARGUMENTS:
**   none
**
** RETURNS:
**   0/1 - Success/Failure
         FPT_ME_DISABLED
         FPT_ME_DISABLE_FAILURE
         FPT_HECI_COMM_ERROR
         ERROR_NULL_CHECK
**
********************************************************************************
--*/
static UINT32 ReadFwStatusRegValue(INT8 fwStatusReg,  UINT32 * pFwStatusRegData)
{
    UINT32      rtnStatus = SUCCESS;
    UINT8       tmpSpace[256] = {0}; // collect entire PCI Config Space data for the Bus/Device/Func requested
    UINT32*     pTmpReg = NULL;
    UINT64 Address = 0;

    if (NULL == pFwStatusRegData)
    {
        //Print(L"\n Failed at NULL == pFwStatusRegData \n");
        return ERROR_NULL_CHECK;
    }

    //Print(L"Entered ReadFwStatusRegValue Address\n");
    Address = EFI_PCI_ADDRESS(ME_BUS, ME_DEV, ME_FUNC, 0);

    //Print(L"\n Address EFI_PCI_ADDRESS(ME_BUS, ME_DEV, ME_FUNC, 0) : 0x%x \n ",Address);

#if 0
    //Unhide HECI to be able to read the FWSTS register.
    UnhideHeci( &heciDisabled );
#endif

    ZeroMem(tmpSpace, sizeof(tmpSpace));

    //Now read the ME FW status register
    //Bus:0, Device:22 & Function:0
    //Print("\n Beofre gPciRootIo \n");

    gPciRootBrdgIo->Pci.Read(gPciRootBrdgIo, EfiPciWidthUint32, Address, 64, tmpSpace);

    //Print("\n After gPciRootIo and abt to enter swithc\n");
    switch( fwStatusReg )
    {
        //Print("\n Enter switch \n");
    case 1:
        pTmpReg = (UINT32 *)&tmpSpace[ ME_PCI_FWSTATUS1_REG ];
        //Print(L"\n pTempReg Case 1 : 0x%08X \n, pTmpReg");
        break;

    case 2:
        pTmpReg = (UINT32*)&tmpSpace[ ME_PCI_FWSTATUS2_REG ];
        //Print(L"\n pTempReg Case 2 : 0x%08X \n, pTmpReg");
        break;

    case 3:
        pTmpReg = (UINT32 *)&tmpSpace[ ME_PCI_FWSTATUS3_REG ];
        //Print(L"\n pTempReg Case 3 : 0x%08X \n, pTmpReg");
        break;

    case 4:
        pTmpReg = (UINT32 *)&tmpSpace[ ME_PCI_FWSTATUS4_REG ];
        //Print(L"\n pTempReg Case 4 : 0x%08X \n, pTmpReg");
        break;

    case 5:
        pTmpReg = (UINT32 *)&tmpSpace[ ME_PCI_FWSTATUS5_REG ];
        //Print(L"\n pTempReg Case 5 : 0x%08X \n, pTmpReg");
        break;

    case 6:
        pTmpReg = (UINT32 *)&tmpSpace[ ME_PCI_FWSTATUS6_REG ];
        //Print(L"\n pTempReg Case 6 : 0x%08X \n, pTmpReg");
        break;

    default:
        //Print(L"\n Invalid Firmware Status Register requested. \n");
        pTmpReg = (UINT32 *)&tmpSpace[ 0 ];    //  need to assign something to the pointer
        rtnStatus = ERROR_NULL_CHECK;
    }

    *pFwStatusRegData = *pTmpReg;
    //Print(L"\n Requested Firmware Status Register: 0x%08X. \n",*pFwStatusRegData);

    //
    //       If the HECI was hidden, re-hide it
    //
#if 0 //In Palt_info this is dsiabled for now.
    if( FALSE != heciDisabled )
    {
        HideHeci();
    }
#endif
    //Print(L"\n status at end : %r \n", rtnStatus);
    return rtnStatus;
}  //    end func: ReadSpecificFwStatusRegValue()

static UINT32 FromCseToFwuErrorCode(UINT32 cseer)
{
    UINT32 i = 0;
    for(i = 0; i < MAX_ERROR_CODE; i++)
    {
        if(ErrorMap[i].Cse_error_code == cseer){
            return ErrorMap[i].Fwu_error_code;
        }
    }
    return FWU_LAST_STATUS_CODE;
}

/*++
    Routine Description:
    Checks if ME FW Update client is ready to accept an update
    Arguments:
    Returns:
    TRUE if ready
--*/
BOOLEAN
    IsUpdateStatusPending (
    IN  UINT32    status
    )

{
    switch (status)
    {
    case STATUS_UPDATE_IMG_LOADING: //fall through
    case STATUS_UPDATE_IMG_AUTHENTICATING: //fall through
    case STATUS_UPDATE_IMG_PROCESSING: //fall through
    case STATUS_UPDATE_CREATING_FT: //fall through
    case STATUS_UPDATE_UPDATING_CODE: //fall through
    case STATUS_UPDATE_UPDATING_NFT:
        return TRUE;
    case STATUS_UPDATE_HOST_RESET_REQUIRED:
        return FALSE;
    }

    return FALSE;
}

#ifdef _DEBUG
static VOID
    _ShowBuffer (
    UINT8  *Message,
    UINT32 Length
    )
    /*++

    Routine Description:

    For serial //DEBUGger used, it will show the buffer message to serila consol.

    Arguments:

    Message - the address point of buffer message
    Length  - message length

    Returns:

    None.
    --*/
{
    UINT32  LineBreak = 0;
    UINT32  Index = 0;

    while (Length-- > 0)
    {
        if (LineBreak == 0)
        {
            Print(L"%02x: ", (Index & 0xF0));
        }

        Print(L"%02x ", Message[Index++]);
        LineBreak++;
        if (LineBreak == 16)
        {
            Print(L"\n");
            LineBreak = 0;
        }

        if (LineBreak == 8) {
            Print(L"- ");
        }
    }

    Print(L"\n");

    return ;
}
#endif

/*++

Routine Description:

Read the flow control buffer as one
is required before we can send a request message to the ME.
Typically the ME will send one after we have sent a request message.
This function allows us to request one if we have not received one.

Arguments:

Returns:

--*/
static EFI_STATUS GetFlowControl (IN ME_CLIENT_PROPERTIES   *MeClient)
{
    EFI_STATUS              Status = EFI_SUCCESS;
    HBM_FLOW_CONTROL        FlowControlMessage = {0};
    UINT32                          Length = 0;
    UINT8                   i = 0;

    if (MeClient->HaveFlowControl)
    {
        return EFI_SUCCESS;
    }

    ZeroMem (&FlowControlMessage, sizeof(HBM_FLOW_CONTROL));
    //
    // Ask ME for flow control buffer
    //
    Length = sizeof (HBM_FLOW_CONTROL);
    Status = mHeci->ReadMsg (0, BLOCKING,(UINT32 *) &FlowControlMessage,&Length);

    DebugPrint(L"Read Msg from Get FC : 0x%08x\n", FlowControlMessage );


    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable to get Flow Control Message - %r\n", Status);
        return Status;
    }

    DebugPrint(L"Got ME Flow Control message via Request for %x\n", FlowControlMessage.MEAddress);

    for (i = 0; i < NUM_ME_CLIENTS; i++)
    {
        if (FlowControlMessage.MEAddress == MeClients[i].ClientId)
        {
            MeClients[i].HaveFlowControl = FALSE; // reset for next time as we will use this request immediately
        }
    }
    return Status;
}

/*++

Routine Description:

Before we wait for a message from the ME, we need to notify the ME
that we have a buffer ready. Since our code already has one, this is really
a clear to send message to the ME

Arguments:

Returns:

--*/
static EFI_STATUS SendFlowControl (IN ME_CLIENT_PROPERTIES  *MeClient)
{
    EFI_STATUS              Status = EFI_SUCCESS;
    HBM_FLOW_CONTROL        FlowControlMessage = {0};
    UINT32                          Length = 0;

    ZeroMem (&FlowControlMessage, sizeof(HBM_FLOW_CONTROL));

    //FlowControlMessage.Command.Command = FLOW_CONTROL_CMD;
    //FlowControlMessage.Command.IsResponse = 0;
    FlowControlMessage.Command = FLOW_CONTROL_CMD;
    FlowControlMessage.MEAddress = MeClient->ClientId;
    FlowControlMessage.HostAddress = MeClient->HostAddress;

    Length = sizeof(HBM_FLOW_CONTROL);
    //
    // Send Flow Control Ready to ME
    //
    Status = mHeci->SendMsg (0,(UINT32 *) &FlowControlMessage, Length, HBM_PROTOCOL_ADDRESS, HBM_PROTOCOL_ADDRESS);
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable to send Client Read Flow Control Message to %d - %r\n", MeClient->ClientId, Status);
        return Status;
    }

    DebugPrint(L"Sent Client Read Flow Control message\n");

    return Status;
}

/*++

Routine Description:

Wait for a message from the ME FWU Client.
Because the ME could be busy, we do a read and if receive a timeout,
repeat the operation 2 additional times.
Because we can receive a flow control message from the ME out of order,
we check if the message received is a flow control.  If it is, we update
the flow control variable and then re-issue the read request.

Arguments:

Returns:

--*/
static EFI_STATUS ReadMessageWithFlow (IN ME_CLIENT_PROPERTIES  *MeClient, IN UINT8* Buffer, IN UINT32 Length)
{
    EFI_STATUS          Status = EFI_SUCCESS;
    HECI_BUS_MESSAGE    *HeciBusMessage = NULL;
    UINT32              ReadLength = 0;
    UINT32              Retry = 0;
    UINT32              i = 0;

    DebugPrint(L"Issuing Read Message for %x bytes\n", Length);

    //
    // Read Message
    //
    Retry = 3; // Retry 3 times with a 5 second timeout
    do
    {
        ReadLength = Length;
        Status = mHeci->ReadMsg (0,BLOCKING,(UINT32 *) Buffer,&ReadLength);

        DebugPrint(L"Read Msg from read message with flow : 0x%02x%02x%02x%02x\n", *(Buffer+3), *(Buffer+2), *(Buffer+1), *(Buffer) );

        if (Status != EFI_TIMEOUT)
        {
            if (EFI_ERROR (Status))
            {
                DebugPrint(L"Unable to Read Message - %r\n", Status);
                return Status;
            }
            //
            // Check if we got a flow control instead of the desired response
            //
            HeciBusMessage = (HECI_BUS_MESSAGE *)Buffer;
            DebugPrint(L"Read Command: %x\n", HeciBusMessage->Command);

            if (HeciBusMessage->Command == FLOW_CONTROL_CMD && HECI_FLOW_CTRL_MSG_LEN == ReadLength)
            {
                HBM_FLOW_CONTROL  *FlowControlMessage = (HBM_FLOW_CONTROL *)Buffer;
                DebugPrint(L"Got ME Flow Control message via Request for %x\n", FlowControlMessage->MEAddress);

                for (i = 0; i < NUM_ME_CLIENTS; i++)
                {
                    if (FlowControlMessage->MEAddress == MeClients[i].ClientId)
                    {
                        MeClients[i].HaveFlowControl = TRUE;
                    }
                }
                ReadLength = MeClients[FWU_CLIENT].MaxMessageLength;
                Status = mHeci->ReadMsg (0,BLOCKING,(UINT32 *) Buffer,&ReadLength);

                DebugPrint(L"Read Msg from read message with flow : 0x%02x%02x%02x%02x\n", *(Buffer+3), *(Buffer+2), *(Buffer+1), *(Buffer) );

                if (EFI_ERROR (Status))
                {
                    DebugPrint(L"Unable to Read Message - %r\n", Status);
                    return Status;
                }
                return Status;
            } else
            {
                Status = GetFlowControl(MeClient);
                if (EFI_ERROR (Status))
                {
                    return Status;
                }
                return Status; // We got our message
            }
        } else
        {
            Retry--; // Only apply retry timeout if we got EFI_TIMEOUT
        }
        DebugPrint(L"Re-Issuing Read Message for %x bytes\n", Length);
    } while (Retry);

    if (Status == EFI_TIMEOUT)
    {
        DebugPrint(L"Unable to Read Message - %r\n", Status);
        return Status;
    }

    return Status;
}

/*++

Routine Description:

Send a message to the ME FWU Client.
A Flow Control response is required from the ME before we can send a message.
Thus if we have not received one, send a request for one.

Arguments:

Returns:

--*/
static EFI_STATUS SendMessageWithFlow (IN ME_CLIENT_PROPERTIES  *MeClient, IN UINT8      *Buffer, IN UINT32     Length)
{
    EFI_STATUS  Status = EFI_SUCCESS;
    UINT32      Retry = 0;

    if (Length > MeClient->MaxMessageLength)
    {
        return EFI_BUFFER_TOO_SMALL;
    }
    //Do send flow control
    /*Status = GetFlowControl(MeClient);
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Status GetFLowControl(MeCLient) : %r \n", Status);
        return Status;
    }*/

    Status = SendFlowControl (&MeClients[FWU_CLIENT]);
    if (EFI_ERROR (Status))
    {
        return Status;
    }

    Retry = 3; // Retry 3 times with 5 second timeout on each message
    do
    {
        //
        // Send message to Local FW Update Client
        //
        DebugPrint(L"Send Message DWord: 0x%02x%02x%02x%02x\n", *(Buffer+3), *(Buffer+2), *(Buffer+1), *(Buffer) );
        Status = mHeci->SendMsg (0,(UINT32 *) Buffer,Length,MeClient->HostAddress,MeClient->ClientId);
        if (Status != EFI_TIMEOUT)
        {
            if (EFI_ERROR (Status))
            {
                DebugPrint(L"Unable send message to Local FW Update Client - %r\n", Status);
                return Status;
            }
        }

        if (!EFI_ERROR (Status))
        {
            return Status; // Everything okay - message sent
        }

        Retry--;
    } while (Retry);

    if (Status == EFI_TIMEOUT)
    {
        DebugPrint(L"Unable to send Message - %r\n", Status);
    }

    return Status;
}

/*++

Routine Description:

Wait for a message from the ME Client without a flow control.
Because the ME could be busy, we do a read and if receive a timeout,
repeat the operation 2 additional times.

Arguments:

Returns:

--*/
static EFI_STATUS ReadMessageWithoutFlow (IN ME_CLIENT_PROPERTIES  *MeClient, IN UINT8      *Buffer, IN UINT32     Length)
{
    EFI_STATUS Status = EFI_SUCCESS;
    UINT32 ReadLength= 0;
    UINT32 Retry = 0;

    DebugPrint(L"Issuing Read Message for %x bytes\n", Length);

    //
    // Read Message
    //
    Retry = 3; // Retry 3 times with a 5 second timeout
    do
    {
        ReadLength = Length;
        Status = mHeci->ReadMsg (0, BLOCKING, (UINT32 *) Buffer, &ReadLength);

        DebugPrint(L"Read Msg from read message with flow : 0x%02x%02x%02x%02x\n", *(Buffer+3), *(Buffer+2), *(Buffer+1), *(Buffer) );

        if (Status != EFI_TIMEOUT)
        {
            if (EFI_ERROR (Status))
            {
                DebugPrint(L"Unable to Read Message - %r\n", Status);
                return Status;
            }
        }

        if (!EFI_ERROR(Status))
        {
            return Status;
        }

        Retry--;
    } while (Retry);

    if (Status == EFI_TIMEOUT)
    {
        DebugPrint(L"Unable to Read Message - %r\n", Status);
        return Status;
    }

    return Status;
}

/*++

Routine Description:

Send a message to the ME Client without a flow control.

Arguments:

Returns:

--*/
static EFI_STATUS SendMessageWithoutFlow (IN ME_CLIENT_PROPERTIES  *MeClient, IN UINT8      *Buffer, IN UINT32     Length)
{
    EFI_STATUS Status = EFI_SUCCESS;
    UINT32 Retry = 0;

    if (Length > MeClient->MaxMessageLength)
    {
        return EFI_BUFFER_TOO_SMALL;
    }

    Retry = 3; // Retry 3 times with 5 second timeout on each message
    do
    {
        //
        // Send message to Local FW Update Client
        //
        DebugPrint(L"Send Message DWord: 0x%02x%02x%02x%02x\n", *(Buffer+3), *(Buffer+2), *(Buffer+1), *(Buffer) );
        Status = mHeci->SendMsg (0, (UINT32 *) Buffer, Length, MeClient->HostAddress, MeClient->ClientId);
        if (Status != EFI_TIMEOUT)
        {
            if (EFI_ERROR (Status))
            {
                DebugPrint(L"Unable send message to Local FW Update Client - %r\n", Status);
                return Status;
            }
        }

        if (!EFI_ERROR (Status))
        {
            return Status; // Everything okay - message sent
        }

        Retry--;
    } while (Retry);

    if (Status == EFI_TIMEOUT)
    {
        DebugPrint(L"Unable to send Message - %r\n", Status);
    }

    return Status;
}

/**++
********************************************************************************
*
** FUNCTION:
**   GetImageFWVersionMsg
**
** DESCRIPTION:
**   This function retrieves GET_IMAGE_FW_VERSION using HECI MKHI client
**
** ARGUMENTS:
**        [IN]  partition_id - id of wanted partition
**        [OUT] module_version_data - inforamtion on the partition
**
** RETURNS:
**        status
**
********************************************************************************
--*/
static UINT32 GetImageFWVersionMsg(uint32_t partition_id, MODULE_VERSION_DATA *module_version_data)
{
    UINT32 Status = SUCCESS;
    EFI_STATUS Status_efi = EFI_SUCCESS;
    ME_CLIENT_PROPERTIES HciClient = {0};
    UINT32 MeMode = 0;
    GET_IMAGE_FW_VERSION req;
    GET_IMAGE_FW_VERSION_ACK resp;
    UINT32 reqLen = sizeof(req);

    // size of response should be, according to GET_IMAGE_FW_VERSION_ACK structure:
    // mkhi_message_header_t +
    // uint32_t +
    // num_of_modules * MODULE_VERSION_DATA
    // num_of_modules is 1, in this case, since we get version of only 1 partition.
    // GET_IMAGE_FW_VERSION_ACK is defined with 1 MODULE_VERSION_DATA, as place holder, so:
    UINT32 respLen = sizeof(GET_IMAGE_FW_VERSION_ACK);


    HciClient.HostAddress = HBM_PROTOCOL_ADDRESS;
    HciClient.ClientId = HECI_ADDRESS_FIX_MKHI;
    HciClient.MaxMessageLength = respLen;

    if (!g_fwulib_heci_initialized){
        Status_efi = gBS->LocateProtocol (
            &gEfiHeciProtocolGuid,
            NULL,
            &mHeci
            );

        if (EFI_ERROR (Status_efi)) {
            DebugPrint(L"InitializeHeci: Could not locate HECI protocol!\n");
            return FWU_GET_PATTRIB_ERROR;
        }
    }
    g_fwulib_heci_initialized = TRUE;

    Status_efi = mHeci->GetMeMode(&MeMode);
    if (EFI_ERROR (Status_efi)) {
        DebugPrint(L"InitializeHeci: Could not obtain ME Mode!\n");
        return FWU_GET_PATTRIB_ERROR;
    }

    if (MeMode != ME_MODE_NORMAL) {
        DebugPrint(L"Local FW Update only supported when ME Mode=Normal\n");
        return FWU_GET_PATTRIB_ERROR;
    }


    ZeroMem(&req, sizeof(req));
    ZeroMem(&resp, sizeof(resp));

    req.header.fields.group_id = MKHI_GROUP_ID_GEN;
    req.header.fields.command = GEN_GET_IMAGE_FW_VERSION_CMD;
    req.partition_id = partition_id;

    Status_efi = SendMessageWithoutFlow(&HciClient, (UINT8 *)&req, reqLen);
    if (EFI_ERROR(Status_efi))
    {
        DebugPrint(L"Unable send message GEN_GET_IMAGE_FW_VERSION_CMD - %r\n", Status_efi);
        Status = FWU_GET_PATTRIB_ERROR;
        return Status;
    }

    DebugPrint(L"Sent GEN_GET_IMAGE_FW_VERSION_CMD - %r\n", Status_efi);

    Status_efi = ReadMessageWithoutFlow(&HciClient, (UINT8 *)&resp, respLen);
    if (EFI_ERROR(Status_efi))
    {
        DebugPrint(L"Unable read message from GEN_GET_IMAGE_FW_VERSION_CMD - %r\n", Status_efi);
        Status = FWU_GET_PATTRIB_ERROR;
        return Status;
    }

    if (SUCCESS != resp.header.fields.result)
    {
        Status = resp.header.fields.result;
        return Status;
    }

    if (resp.num_of_modules != 1)
    {
        Status = FWU_PART_NOT_PRESENT;
        return Status;
    }

    CopyMem(module_version_data, resp.module_entries, sizeof(MODULE_VERSION_DATA));

    return Status;
}

/*++

Routine Description:

Compares the two version data and based on policy
returns result indicated below

Arguments:

Returns:

return < 0 if a < b
0 if a = b
> 0 if a > b

--*/

static int VersionCompare (IN FWVersion *a, IN FWVersion *b)
{
    DebugPrint(L"a = %d.%d.%d.%d\n", a->Major, a->Minor, a->Hotfix, a->Build);
    DebugPrint(L"b = %d.%d.%d.%d\n", b->Major, b->Minor, b->Hotfix, b->Build);

    if (a->Major != b->Major)
    {
        return ((int)a->Major - (int)b->Major);
    }

    if (a->Minor != b->Minor)
    {
        return ((int)a->Minor - (int)b->Minor);
    }

    if (a->Hotfix != b->Hotfix)
    {
        return ((int)a->Hotfix - (int)b->Hotfix);
    }

    if (a->Build != b->Build)
    {
        return ((int)a->Build - (int)b->Build);
    }

    return 0;
}

/*++

Routine Description:

Retrieves the Update params from the ME

NOTE: A connection to the client is required before this routine is
executed.

Arguments:

Returns:

EFI_STATUS

--*/
static EFI_STATUS GetFWUpdateClientParams ()
{
    EFI_STATUS Status = EFI_SUCCESS;
    DebugPrint(L"GetFWUpdateClientParams (): mInitialized: %d\n", mInitialized);

    if (!mInitialized)
    {
        Status = GetFwVersion ();
        if (EFI_ERROR (Status))
        {
            DebugPrint(L"Unable to get current FW version - %r\n", Status);
            return EFI_VOLUME_CORRUPTED;
        }
    }

    return Status;
}

/*++

Routine Description:

Start the update process by asking the ME FW Update client
to allocate a buffer for the image.  Note that the ME does not update the flash until
the FWU_END message is received.

If the current ME Flash image is password protected, then that password is required to unlock
the update process

Because the process of downloading the image is time consuming, a callback is provided for each
chunk of data sent to the ME FW Update client.  If desired, the caller can specify the number of steps
that the total data send will have (e.g. 100 means a total of 100 callbacks for the entire duration). The
number of steps supplied will be divided by the total result of (FileLength / Chunksize) + 1.

Arguments:

UINT8 pointer to FW Update Image
UINT32 FW Update image length
CHAR16 pointer to password (if no password then NULL)
VOID pointer to callback function (if no callback then NULL)
UINT32 callback steps  (1 means each packet sent will trigger a callback)
UINT32 pointer to ME FW Update Client return status

Returns:

EFI_STATUS

--*/
static UINT32
    _startUpdate(
    IN  UINT32            FileLength,
    IN UINT32                updSize,
    IN  CHAR             *Password,
    IN  UINT32            UpdateType,
    IN  UINT32            IpuIdToBeUpdated,
    IN  UINT32            UpdateEnvironment,
    IN  _UUID             OemID,
    IN  UPDATE_FLAGS_LIB  UpdateFlags
    )
{
    EFI_STATUS                Status = EFI_SUCCESS;
    UINT32              fwuError = FWU_ERROR_SUCCESS;
    FWU_START_MSG             *pStartMsg = NULL;
    FWU_HECI_MSG              *ReplyMsg = NULL;
    UINT32                    Total_Struct_Size = 0;
    UINT8                     PwdLength = 0;
    CHAR                      *Pwd = Password;
    UINT32                     i = 0;
    UINT32 dwReplyBufferSize = 0;

    //
    // Since the password can be variable length,
    // we need to dynamically allocate the FWU_START_MSG structure
    // to hold the data
    //
    if (Password != NULL)
    {
        while (*Pwd++)
        {
            PwdLength++;
        }
        if (PwdLength > FWU_PWD_MAX_SIZE)
        {
            PwdLength = FWU_PWD_MAX_SIZE;
        }

        Total_Struct_Size += PwdLength;
    }

    Total_Struct_Size = (UINT32)sizeof(FWU_START_MSG);

    DebugPrint(L"Total_Struct_Size Length: %d\n", Total_Struct_Size);

    pStartMsg = (FWU_START_MSG *)AllocatePool(Total_Struct_Size);
    if (pStartMsg == NULL)
    {
        return FWU_NO_MEMORY;
    }

    ZeroMem (pStartMsg, Total_Struct_Size);

    pStartMsg->MessageType = FWU_START;
    pStartMsg->Length = FileLength;
    pStartMsg->UpdtLength = updSize;
    pStartMsg->UpdateType = (UINT8)UpdateType; // 0 = FULL, 1 = PARTIAL
    pStartMsg->PassWordLength = PwdLength;
    pStartMsg->IpuIdTobeUpdated = IpuIdToBeUpdated;
    pStartMsg->UpdateEnvironment = UpdateEnvironment;

    CopyMem (&pStartMsg->UpdateFlags, &UpdateFlags, sizeof (UPDATE_FLAGS));
    CopyMem (&pStartMsg->OemId, &OemID, sizeof (_UUID));

    if (PwdLength)
    {
        CopyMem (pStartMsg->PassWordData, Password, pStartMsg->PassWordLength);
        DebugPrint(L"Password from user: ");
        for (i = 0; i < PwdLength; i++)
        {
            DebugPrint(L"%c", pStartMsg->PassWordData[i]);
        }
        DebugPrint(L"\nPassword Length: %d\n", pStartMsg->PassWordLength);
    }
    //New
    //  bufferSize = GetBufferSize(&fwuclt);
    dwReplyBufferSize = sizeof(FWU_START_MSG_REPLY);

    Status = SendMessageWithFlow (&MeClients[FWU_CLIENT], (UINT8 *)pStartMsg, (UINT32)Total_Struct_Size);
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable send message FWU_START - %r\n", Status);
        return FWU_IME_NOT_READY;
    }

    DebugPrint(L"Sent FWU_START - %r\n", Status);

    FreePool (pStartMsg);

    //
    // Before receiving, we send a flow control to the ME indicating we have a buffer
    //
    /*Status = SendFlowControl (&MeClients[FWU_CLIENT]);
    if (EFI_ERROR (Status))
    {
        return FWU_IME_NOT_READY;
    }*/

    Status = ReadMessageWithFlow (&MeClients[FWU_CLIENT], mpRxBuff, MeClients[FWU_CLIENT].MaxMessageLength);
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable read message from FWU_START - %r\n", Status);
        return FWU_IME_NOT_READY;
    }

    ReplyMsg = (FWU_HECI_MSG *)mpRxBuff;
    if (ReplyMsg->MessageData.MessageType != FWU_START_REPLY) {
        DebugPrint(L"Start Reply Message Type Incorrect: %x\n", ReplyMsg->MessageData.MessageType);
        return FWU_IME_UN_SUP_MESS;
    }

    if (ReplyMsg->MessageData.StartReply.Status != FWU_STATUS_SUCCESS) {
        DebugPrint(L"Start Reply Status: %x\n", ReplyMsg->MessageData.StartReply.Status);
        fwuError = FromCseToFwuErrorCode(ReplyMsg->MessageData.StartReply.Status);

        return fwuError;
    }

    return fwuError;
}

/*++
********************************************************************************
**
** FUNCTION:
**   _sendData
**
** DESCRIPTION:
**   This function sends the required partitions gathered in a buffer in bytes to the FW.
**
** ARGUMENTS:
**   rxBuff            - command buffer to be sent to the FW
**   dataBuff        - buffer with actual bytes to the FW
**     len            - length of the data buffer currently being sent
**
** RETURNS:
**  TBD
**
** NOTE:
**
********************************************************************************
--*/
static UINT32
    _sendData (
    IN  UINT8   *rxBuff,
    IN  UINT8   *dataBuff,
    IN  int     len
    )
{
    EFI_STATUS        Status = EFI_SUCCESS;
    FWU_HECI_MSG      *ReplyMsg = NULL;
    UINT32            Length = 0;

    Length = len + sizeof(FWU_DATA_MESSAGE);
    Status = SendMessageWithFlow (&MeClients[FWU_CLIENT], mpTxBuff, Length);

    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable to send Data Update Message of %x bytes - %r\n", Length, Status);
        return FWU_IME_NOT_READY;
    }

    DebugPrint(L"Sent FWU_DATA of %x bytes - %r\n", Length, Status);

    //
    // Before receiving, we send a flow control to the ME indicating we have a buffer
    //
    /*Status = SendFlowControl (&MeClients[FWU_CLIENT]);
    if (EFI_ERROR (Status))
    {
        return FWU_IME_NOT_READY;
    }*/

    Status = ReadMessageWithFlow (&MeClients[FWU_CLIENT], mpRxBuff, Length);
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable read message from FWU_DATA - %r\n", Status);
        return FWU_IME_NOT_READY;
    }

    ReplyMsg = (FWU_HECI_MSG *)mpRxBuff;
    if (ReplyMsg->MessageData.MessageType != FWU_DATA_REPLY)
    {
        DebugPrint(L"Send Data Update Reply Message Type Incorrect: %x\n", ReplyMsg->MessageData.MessageType);
        return FWU_IME_UN_SUP_MESS;
    }
    DebugPrint(L"Send Data Update Reply Status: %x\n", ReplyMsg->MessageData.DataReply.Status);
    return FWU_ERROR_SUCCESS;
}
/*++
********************************************************************************
**
** FUNCTION:
**   _endUpdate
**
** DESCRIPTION:
**   This function sends the FWU_END message to the FW.
**
** ARGUMENTS:
**   _forceResetLib    - flag indicating auto system restart after update
**
** RETURNS:
**  TBD
**
** NOTE:
**
********************************************************************************
--*/
static UINT32 _endUpdate (IN  int   _forceResetLib)
{
    EFI_STATUS        Status = EFI_SUCCESS;
    FWU_END_MESSAGE   EndMsg = {0};

    EndMsg.MessageType = FWU_END;

    Status = SendMessageWithFlow (&MeClients[FWU_CLIENT], (UINT8 *)&EndMsg, sizeof(EndMsg));
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable send message FWU_END - %r\n", Status);
        //return FWU_IME_NOT_READY;
        return FWU_GENERAL;
    }

    DebugPrint(L"Sent FWU_END - %r\n", Status);

    /*Status = SendFlowControl (&MeClients[FWU_CLIENT]);
    if (EFI_ERROR (Status))
    {
        //return FWU_IME_NOT_READY;
        return FWU_GENERAL;
    }*/

    //
    // Allow FWU Client to process the message before polling begins
    //
    gBS->Stall(25000); // Wait 1 sec before polling again

    return FWU_ERROR_SUCCESS;
}

/*++
********************************************************************************
**
** FUNCTION:
**   Update
**
** DESCRIPTION:
**   This function sends the data to the FW Update Client
**
** ARGUMENTS:
**
** RETURNS:
**
** NOTE:
**
********************************************************************************
--*/
static UINT32
    Update (
    IN  CHAR8              **Buffer,
    IN  UINT32              BufSize,
    IN  UINT32              updtSize,
    IN  CHAR8*             _pwd,
    IN  int               _forceResetLib,
    IN  UINT32            UpdateType,
    IN  UINT32            IpuIdToBeUpdated,
    IN  UINT32            UpdateEnvironment,
    IN  _UUID             OemID,
    IN  UPDATE_FLAGS_LIB  UpdateFlags,
    IN  void (**func)(float, float)
    )
{
    int               bytesCopied = 0;
    UCHAR             *dataField = 0;
    UINT32            msgBufSize = mChunkSize;
    UINT32            BufPtr = 0;
    UINT32            totalBufferSize = BufSize + updtSize;
    UINT32             statusCount = 0;
    FWU_DATA_MSG      *msg = NULL;
    UINT32             AccumulatedBytesRead = 0;
    UINT32      fwuError = 0;

    ZeroMem ((void *)mpRxBuff, msgBufSize);
    ZeroMem ((void *)mpTxBuff, msgBufSize);

    // Send out start Update messages to the FW
    fwuError =_startUpdate(BufSize,updtSize, _pwd,UpdateType,
        IpuIdToBeUpdated, UpdateEnvironment, OemID,UpdateFlags);
    if (FWU_ERROR_SUCCESS != fwuError)
    {
        return fwuError;
    }

    msg = (FWU_DATA_MSG*)mpTxBuff;
    msg->MessageType = FWU_DATA;

    // Start of the Data field
    dataField = msg->Data;

    statusCount = 0;
    AccumulatedBytesRead = 0;
    BufPtr = 0;
    bytesCopied = msgBufSize - sizeof(FWU_DATA_MSG);
    while (AccumulatedBytesRead < totalBufferSize)
    {
        // The following IF stmt will make sure
        // that the right last bytes are copied
        if ((BufPtr + bytesCopied) > totalBufferSize)
        { //update number of bytes copied to the remaining bytes
            bytesCopied = totalBufferSize - BufPtr;
        }

        // Copy from buffer to Heci message data
        CopyMem (dataField, *Buffer + BufPtr, bytesCopied);

        // Compute total bytes copied so far
        AccumulatedBytesRead += bytesCopied;
        BufPtr = AccumulatedBytesRead;

        msg->Length = bytesCopied;
        ZeroMem((UINT8 *)mpRxBuff,msgBufSize);
        fwuError = _sendData (mpRxBuff, mpTxBuff, sizeof(FWU_DATA_MSG) + bytesCopied);
        if (FWU_ERROR_SUCCESS != fwuError)
        {
            return fwuError;
        }

        if (NULL != *func)
        {
            // Call function to display send percentage
            (*func)((float)AccumulatedBytesRead, (float)totalBufferSize);
        }
    }


    fwuError = _endUpdate (_forceResetLib);
    if (FWU_ERROR_SUCCESS != fwuError)
    {
        return fwuError;
    }
    DebugPrint(L"Update FWU_ERROR_SUCCESS\n");
    return fwuError;
}

/*++

Routine Description:

Get the Flash Partition Table from the image

Arguments:

Returns:

--*/
static UINT32
    ParsingImageForFPT (
    IN  UCHAR* buffer,
    OUT UINT32 *StartOffsetAfterLastActivePartition)
{
    fptt_ver1_t*  fptTable   = NULL;
    UINT32        tableSize  = 0;
    UINT32        i = 0;

    DebugPrint(L"ParsingImageForFPT()\n");
    fptTable = (fptt_ver1_t*)(buffer + FPT_FLASH_INITIAL_BASE);

    //If the fpttype Version2 structure is loaded already return ASAP
    if (NULL != g_fptt)
    {
        goto compute;
    }

    //If FPT is found in the image
    if (FPT_PARTITION_NAME_FPT != fptTable->header.header_marker)
    {
        DebugPrint(L"No FPT in the Image provided\n");
        return NO_FPT_IMAGE;
    }

    DebugPrint(L"fptTable header num_of_entries: 0x%x\n",fptTable->header.num_of_entries);

    tableSize = sizeof(fpt_header_ver_1_t) + (fptTable->header.num_of_entries*sizeof(fpt_entry_ver1_t));

    //found FPT
    //The only place this global structure gets allocated memory
    g_fptt = (fptt_ver1_t*)AllocatePool (tableSize);

    if (NULL == g_fptt)
    {
        DebugPrint(L"No Memory Error\n");
        return FWU_NO_MEMORY;
    }

    ZeroMem(g_fptt, tableSize);

    //Copy the Header and Entries into the global structure
    CopyMem((CHAR8*)g_fptt, (CHAR8*)fptTable, tableSize);

compute:
    for (i = 0; i < g_fptt->header.num_of_entries; i++)
    {
        UINT32 paratitionEndOffset = 0;

        DebugPrint(L"fptt->entry[]->partition_name: 0x%08X\n",g_fptt->entry[i].partition_name);
        DebugPrint(L"fptt->entry[]->offset:         0x%08X\n",g_fptt->entry[i].offset);
        DebugPrint(L"fptt->entry[]->length:         0x%08X\n",g_fptt->entry[i].length);

        //The start of the NON-RUNNABLE part of the image is computed
        //by adding the start and the length of the last region in the
        //runnable portion
        paratitionEndOffset = fptTable->entry[i].offset + fptTable->entry[i].length;
        if (paratitionEndOffset > *StartOffsetAfterLastActivePartition)
        {
            *StartOffsetAfterLastActivePartition = paratitionEndOffset;
        }
    }

    return FWU_ERROR_SUCCESS;
}

/*++
*******************************************************************************c*
**
** FUNCTION:
**   FlashUpdateFindManifestExtension
**
** DESCRIPTION:
**   This function will locate a manifest extension based on the signature and
**   return a reference pointer to the caller.
**
** ARGUMENTS:
**   ManifestExtSignature - Extension to be searched
**   pManifest            - Pointed Mafifest buffer area
**   ppExt                - Will be pointed to extension found or NULL.
**
** RETURNS:
**   Data Format Version (UINT32)
**   0: INVALID_DATA_FORMAT_VERSION
**
**
** NOTE: FlashUpdateFindManifestExtension is common between HECI and AMT based
**            updates. Any changes to this function should be propogated into
**            CheckPolicy for AMTUpdater and HECIUpdater.
**
********************************************************************************
--*/
static void
    FlashUpdateFindManifestExtension(
    IN UINT32              ManifestExtTypeId,
    OUT manifest_header_t * manifestHdr,
    OUT void          **ppExt
    )
{
    mft_ext_header* manifestExtHdr = NULL;
    UINT32 mfstHeaderSize = 0;
    UINT32 mfstExtSize = 0;

    // Initialize extension pointer.
    *ppExt = NULL;

    DebugPrint(L"FlashUpdateFindManifestExtension() to find 0x%x\n", ManifestExtTypeId);

    // Verify manifest pointer and manifest header ID.
    if(manifestHdr == NULL)
    {
        DebugPrint(L"manifestHdr is null \n");
        return;
    }

    if(MANIFEST_HEADER_ID != manifestHdr->header_id)
    {
        DebugPrint("Header ID: [%08X]  Expecting: [%08X] \n",
            manifestHdr->header_id, MANIFEST_HEADER_ID);
        return;
    }

    // Verify if we have an extension in the manifest.
    if((manifestHdr->size) <= manifestHdr->header_length)
    {
        DebugPrint(L"manifestHdr->size:          [%08X] \n",manifestHdr->size);
        DebugPrint(L"manifestHdr->header_length: [%08X] \n",manifestHdr->header_length);
        return;
    }

    // Get a pointer to the first manifest extension header.
    mfstHeaderSize = manifestHdr->header_length * 4;
    manifestExtHdr = (mft_ext_header *)((UINT8 *)manifestHdr + mfstHeaderSize);

    // Loop until the desired extension is found or pass the buffer limit.
    do
    {
        DebugPrint(L"manifestExtHdr->type:          [%08X] \n",manifestExtHdr->type);
        DebugPrint(L"ManifestExtTypeId:             [%08X] \n",ManifestExtTypeId);
        // Check for matching extension header ID.
        if(manifestExtHdr->type == ManifestExtTypeId)
        {
            DebugPrint(L"FlashUpdateFindManifestExtension() found the ManifestExt: 0x%0x\n", ManifestExtTypeId);
            *ppExt = (VOID *)manifestExtHdr;
            break;
        }

        // Advance to the next extension block.
        mfstExtSize = (manifestExtHdr->length);
        manifestExtHdr = (mft_ext_header *)((UINT8 *)manifestExtHdr + mfstExtSize);
    } while((UINT32)((UINT8*)manifestExtHdr - (UINT8*)manifestHdr) < (manifestHdr->size * 4));

    return;
}

/** Returns the offset to the manifest with respect to the FPT base */
static UINT32 FindManifestOffset(UCHAR* buffer,UINT32 FTPROffsetFromFPT)
{
    UINT32 entryIdx = 0;
    UINT32 manifestListIdx = 0;
    UINT32 manifestOffset = 0;
    manifest_header_t* manifestHdr = NULL;
    partition_directory_header_t*      cpdHdr = (partition_directory_header_t*)(buffer + FTPROffsetFromFPT);
    code_partition_directory_entry_t*  cpdEntry = NULL;

    if (PARTITION_DIRECTORY_HEADER_MARKER != cpdHdr->header_marker)
    {
        DebugPrint(L"Unable to validate CPD header\n");
        return INVALID_PARTITION_START;
    }

    // Get first cpd entry
    cpdEntry = (code_partition_directory_entry_t*)(cpdHdr + 1);

    // parse cpd entries until
    for (entryIdx = 0; entryIdx < cpdHdr->num_of_entries; entryIdx++)
    {
        for (manifestListIdx = 0; manifestListIdx < FwUpdPartitionNameCount; manifestListIdx++)
        {
            if (!strncmp((const CHAR8*)cpdEntry[entryIdx].name,
                FwUpdManifestFileNames[manifestListIdx],
                MAX_CPD_ENTRY_NAME_SIZE))
            {
                //manifestOffset = (UINT32)((CHAR8*)FTPROffsetFromFPT + cpdEntry[entryIdx].offset.offset);
                manifestOffset = FTPROffsetFromFPT + cpdEntry[entryIdx].offset.offset;
                manifestHdr = (manifest_header_t*)(manifestOffset + buffer);
                if (MANIFEST_HEADER_ID != manifestHdr->header_id)
                {
                    DebugPrint(L"CPD Entry points to an invalid manifest\n");
                    return INVALID_PARTITION_START;
                }
                return manifestOffset;
            }
        }
    }

    return INVALID_PARTITION_START;
}

/*++
********************************************************************************
**
** FUNCTION:
**   GetImageDataFormatVersion
**
** DESCRIPTION:
**   This function will extract the data format version from an update image
**   and return it to the caller.
**
** ARGUMENTS:
**   strFileName - pointer to the update image filename.
**
** RETURNS:
**   Data Format Version (UINT32)
**   0: INVALID_DATA_FORMAT_VERSION
**
** NOTE: GetImageDataFormatVersion is common between HECI and AMT based updates. Any
**       changes to this function should be propogated into CheckPolicy for AMTUpdater
**       and HECIUpdater.
**
********************************************************************************
--*/
static UINT32 GetImageDataFormatVersion (
    IN  UCHAR          *buffer,
    IN  UINT32 FTPROffsetFromFPT,
    FlashVersion *pImgVer
    )
{
    mft_ifwi_part_man_ext_t* ifwiPartExt = NULL;
    manifest_header_t*   manifestHdr = NULL;
    UINT32 manifestOffset = 0;

    if (NULL == buffer)
    {
        return FWU_INVALID_POINTER;
    }

    manifestOffset = FindManifestOffset(buffer,FTPROffsetFromFPT);

    if (INVALID_PARTITION_START == manifestOffset)
    {
        return INVALID_DATA_FORMAT_VERSION;
    }
    // Update manifest pointer
    manifestHdr = (manifest_header_t *)((UINT8*)buffer + manifestOffset);

    // Update Code Version from manifest and return to caller.
    pImgVer->code.Major  = GetVersionMajor (manifestHdr->version);
    pImgVer->code.Minor  = GetVersionMinor (manifestHdr->version);
    pImgVer->code.Hotfix = GetVersionHotfix(manifestHdr->version);
    pImgVer->code.Build  = GetVersionBuild (manifestHdr->version);

    // Obtain a pointer to the partition info extension
    FlashUpdateFindManifestExtension(MFT_EXT_TYPE_IFWI_PART_MAN, manifestHdr,(void**)&ifwiPartExt);
    if (!ifwiPartExt)
    {
        return INVALID_DATA_FORMAT_VERSION;
    }

    // Return the data format version to caller
    return ifwiPartExt->data_format_version;
}

/*++
********************************************************************************
**
** FUNCTION:
**   GetMeFwCaps
**
** DESCRIPTION:
**   This function will return the MEFWCAPS_SKU for the FW image stored in
**   buffer.
**
** ARGUMENTS:
**        buffer                - Pointer to FW image
**        FTPROffsetFromFTP    - FTP relative offset from FTP
**        meFwCapsSku            - Pointer used to return MEFWCAPS_SKU
**
** RETURNS:
**        meFwCapsSku by call-by-reference
**
**        0:       FWU_ERROR_SUCCESS
**        error: FWU_SKU_MISMATCH
**
********************************************************************************
--*/

static UINT32 GetMeFwCaps(
    UCHAR *buffer,
    UINT32 FTPROffsetFromFPT,
    //mft_fw_sku_capabilities_t** meFwCapsSku,
    mft_fw_sku_attributes_t** meFwAttrSku)
{
    mft_client_sys_info_ext_t* clientSysInfoExt  = NULL;
    manifest_header_t*  manifestHdr = NULL;
    UINT32 manifestOffset = 0;

    if (NULL == buffer)
    {
        return FWU_INVALID_POINTER;
    }

    manifestOffset = FindManifestOffset(buffer,FTPROffsetFromFPT);
    if (INVALID_PARTITION_START == manifestOffset)
    {
        return FWU_SKU_MISMATCH;
    }

    // Update manifest pointer
    manifestHdr = (manifest_header_t *)((UINT8 *)buffer + manifestOffset);

    // Obtain a pointer to the system info extension
    FlashUpdateFindManifestExtension(MFT_EXT_TYPE_CLIENT_SYS_INFO, manifestHdr,(void **)&clientSysInfoExt);
    if (!clientSysInfoExt)
    {
        return FWU_SKU_MISMATCH;
    }

    //*meFwCapsSku = &clientSysInfoExt->fw_sku_capabilities;
    *meFwAttrSku = &clientSysInfoExt->fw_sku_attributes;
    return FWU_ERROR_SUCCESS;
}

/*
++
********************************************************************************
**
** FUNCTION:
**   _CheckPolicy
**
** DESCRIPTION:
**   This function will determine to upgrade or downgrade and check for the right
**        data format version.
**
** ARGUMENTS:
**        img            - the object of FWImage Class
**        allowsv        - Allow same version flag
**        strFileName    - input file
**        Upd_Type        - Update type output
**
** RETURNS:
**        Upd_Type by call-by-reference
**
**        0:    FWU_ERROR_SUCCESS
**        8714: FWU_FILE_OPEN
**        8710: FWU_NO_MEMORY
**
** NOTE: CheckPolicy is common between HECI and AMT based updates.
**            Any changes to this function should be propogated into CheckPolicy
**            for AMTUpdater and HECIUpdater
**
********************************************************************************
--*/
static UINT32
    _CheckPolicy (
    IN  UINT32 allowsv,
    IN  UCHAR *buffer,
    IN  UPDATE_TYPE **Upd_Type,
    IN  VersionLib **ver)
{
    FlashVersion    imgVer = {0};
    INT32           CodeVersion = 0;
    UINT32          StartOffsetAfterLastActivePartition = 0;
    UINT32          FTPROffsetFromFPT = 0;
    UINT32          ImageDataFormatVersion = 0;
    BOOLEAN            noFPT = FALSE;
    EFI_STATUS      Status = EFI_SUCCESS;
    UINT32    fwuError = 0;
    UINT32    i = 0;
    //mft_fw_sku_capabilities_t* imgMeFwCapsSku = NULL;
    //mft_fw_sku_attributes_t* imgMeFwAttrSku = NULL;

    //UINT32 IsLpPlatform = 0;
    mft_fw_sku_attributes_t* imgMeFwAttrSku = NULL;
    //ME_PLATFORM_TYPE platformType = {0};  

    Status = GetFWUpdateClientParams ();
    if (EFI_ERROR (Status)) {
        DebugPrint(L"Unable to get current FW version - %r\n", Status);
        return FWU_GET_ME_FWU_INFO;
    }

    // Read FPT and check for its presence in the image
    fwuError = ParsingImageForFPT (buffer, &StartOffsetAfterLastActivePartition);
    if (fwuError != FWU_ERROR_SUCCESS)
    {
        DebugPrint(L"ParsingImageForFPT() returned %d\n", Status);
        if (NO_FPT_IMAGE == fwuError)
        {
            noFPT = TRUE;
        }
        else
        {
            return fwuError;
        }
    }

    //
    // With Flash partition table
    //
    if (!noFPT)
    {
        for (i = 0; i < g_fptt->header.num_of_entries; i++)
        {
            //Trying to find FTP, NFTP, and/or WCOD/LOCL/FPSP
            if (FPT_PARTITION_NAME_FTPR == g_fptt->entry[i].partition_name)
            {
                FTPROffsetFromFPT = g_fptt->entry[i].offset;
                break;
            }
        }
    }
    else
    {
        DebugPrint(L"Check Policy noFPT \n");
        return FWU_FILE_INVALID;
    }

    DebugPrint(L"FTPROffsetFromFPT: 0x%x\n", FTPROffsetFromFPT);

    // once LPT-LP should not support LPT-H this code should change.
    // For now LPT-LP tools should support both LPT-H and LPT-LP.
    // Allow LPT-H to be updated to LPT-H capable fw sku.
    // Allow LPT-LP to be updated to LPT-LP capable fw sku
    // Don't allow LPT-H -> LPT-LP or LPT-LP -> LPT-H upgrades

    fwuError = GetMeFwCaps(buffer, FTPROffsetFromFPT, &imgMeFwAttrSku);
    if (FWU_ERROR_SUCCESS != fwuError)
    {
        return fwuError;
    }

    /*fwuError = GetSkuFromHeciDid(&IsLpPlatform);
    if (FWU_ERROR_SUCCESS != fwuError)
    {
        DebugPrint(L"Error from GetSkuFromHEciDid");
        return fwuError;
    }

    DebugPrint(L"imgMeFwAttrSku: (H-0, LP-1) = %d\n", imgMeFwAttrSku->pch_type);
    DebugPrint(L"imgMeFwAttrSku: (H-0, LP-1) = %d\n",imgMeFwAttrSku->pch_type);
    DebugPrint(L"IsLpPlatform: (H-0, LP-1) = %d\n", IsLpPlatform);

    //pch_type: LP - 1, H - 0
    //Checking if given update image PCH type LP or H is equal to current machine PCH
    if (imgMeFwAttrSku->pch_type != IsLpPlatform)
    {
        fwuError = FWU_SKU_MISMATCH;
        return fwuError;
    }*/

/*#if !DISABLE_MEFWCAPS_CHECK
    if (!((_mefwcaps_sku.h && imgMeFwCapsSku->h) ||
        (_mefwcaps_sku.lp && imgMeFwCapsSku->lp)))
    {
        fwuError = FWU_SKU_MISMATCH;
        return fwuError;
    }
#endif*/

    DebugPrint(L"FTPROffsetFromFPT: 0x%x\n", FTPROffsetFromFPT);

    // Skipping MeFWCapsSKu here. Exists in Eng version
    // Get data format version from update image.
    // Note: this also retrieves the image code version and returns to "imgVer".
    //GetImageDataFormatVersion function exists outside
    ImageDataFormatVersion = GetImageDataFormatVersion(buffer, FTPROffsetFromFPT, &imgVer);

    (*ver)->Major = imgVer.code.Major;
    (*ver)->Minor = imgVer.code.Minor;
    (*ver)->Hotfix = imgVer.code.Hotfix;
    (*ver)->Build = imgVer.code.Build;

    DebugPrint(L"ImageDataFormatVersion: 0x%x\n", ImageDataFormatVersion);
    DebugPrint(L"mDataFormatVersion: 0x%x\n", mDataFormatVersion);
    DebugPrint(L"ImageDataFormatVersion: 0x%x\n", ImageDataFormatVersion);
    DebugPrint(L"mDataFormatVersion: 0x%x\n", mDataFormatVersion);

    // Get version from running FW.
    CodeVersion = VersionCompare (&imgVer.code, &mVer.code);

    DebugPrint(L"VersionCompare returned %d\n", CodeVersion);

    // Check for Downgrade
    if (0 > CodeVersion)
    {
        // Verify SAME data format version of image and current FW for Downgrade.
        if (mDataFormatVersion != ImageDataFormatVersion)
        {
            **Upd_Type = DOWNGRADE_FAILURE;
            fwuError = FWU_DOWNGRADE_NOT_ALLOWED_DATA_MISMATCH;
            DebugPrint(L"Downgrade not allowed\n");
            return fwuError;
        }

        // Successful downgrade
        **Upd_Type = DOWNGRADE_SUCCESS;
        fwuError = FWU_ERROR_SUCCESS;
        DebugPrint(L"Downgrade allowed\n");
        return fwuError;
    }

    // Verify data format version for same version and upgrade
    if (0 <= CodeVersion)
    {
        **Upd_Type = UPGRADE_SUCCESS;

        // Verify SAME data format version of image and current FW for Downgrade.
        if (mDataFormatVersion < ImageDataFormatVersion)
        { // Successful upgrade with prompt
            **Upd_Type = UPGRADE_PROMPT;
            fwuError = FWU_ERROR_SUCCESS;
            DebugPrint(L"Upgrade not allowed\n");
            return fwuError;
        }
    }

    // Check for Same Version
    if (0 == CodeVersion)
    {
        **Upd_Type = SAMEVERSION_SUCCESS;

        if (!allowsv)
        {
            **Upd_Type = SAMEVERSION_FAILURE;
            fwuError = FWU_USAGE_ERROR_B;
            DebugPrint(L"Same version not allowed\n");
            return fwuError;
        }
    }
    DebugPrint(L"Upd_Type = %x\n", **Upd_Type);
    fwuError = FWU_ERROR_SUCCESS;

    return fwuError;
}

/*++
********************************************************************************
**
** FUNCTION:
**   _countNonDuplicates
**
** DESCRIPTION:
**   Count the number of non duplicate images
**
** ARGUMENTS:
**
** RETURNS:
**
** NOTE:
**
********************************************************************************
--*/
static UINT32
    _countNonDuplicates (
    IN  ImagePartition *imgPart,
    IN  UINT32 NumParts
    )

{
    UINT32    count = 0;
    ImagePartition  *Temp = NULL;
    UINT32    tempPName = 0;
    UINT32    tempInstId = 0;
    UINT32   i = 0, j = 0;

    if (NumParts <=1 || NULL == imgPart)
    {
        return NumParts;
    }

    Temp = (ImagePartition*)AllocatePool(NumParts*sizeof(ImagePartition));

    if (NULL == Temp) {
        return count;
    }

    ZeroMem (Temp, NumParts*sizeof(ImagePartition));
    CopyMem(Temp,imgPart,NumParts*sizeof(ImagePartition));

    for(i=0;i<NumParts-1;i++)
    {
        for(j=0;j<NumParts-i-1;j++)
        {
            if (Temp[j].PartitionName > Temp[j+1].PartitionName ||
                Temp[j].InstanceId >= Temp[j+1].InstanceId)
            {
                tempPName = Temp[j].PartitionName;
                tempInstId = Temp[j].InstanceId;

                Temp[j].PartitionName = Temp[j+1].PartitionName;
                Temp[j].InstanceId = Temp[j+1].InstanceId;

                Temp[j+1].PartitionName = tempPName;
                Temp[j+1].InstanceId = tempInstId;
            }
        }
    }

    tempPName = Temp[0].PartitionName;
    tempInstId = Temp[0].InstanceId;
    count = 1;
    for(i=1;i<NumParts;i++)
    {
        if (Temp[i].PartitionName != tempPName ||
            Temp[i].InstanceId != tempInstId)
        {
            count++;
        }
        tempPName = Temp[i].PartitionName;
        tempInstId = Temp[i].InstanceId;

    }
    FreePool(Temp);

    return count;
}

/*++
********************************************************************************
**
** FUNCTION:
**   ReadParitionNameManifestType
**
** DESCRIPTION:
**   This function will read the partition name from begining of the Inout File
**
** ARGUMENTS:
**   InputFile            - Input File to read
**   PartitionName   - Buffer area to copy the extracted PartitionName
**
** RETURNS:
**        PartitionName by call-by-reference
**        0:    FWU_ERROR_SUCCESS
**        8714: FWU_FILE_OPEN
**        8710: FWU_NO_MEMORY
**
** NOTE:
**
********************************************************************************
--*/
static UINT32
    ReadParitionNameManifestType (
    IN  UCHAR          *buffer,
    OUT UINT32  *PartitionName
    )
{
    partition_directory_header_t* cpdHdr = (partition_directory_header_t*)buffer;

    //Copy the Name from the Manifest to the PartitionName buffer returned
    CopyMem(PartitionName,
        &cpdHdr->partition_name,
        sizeof(cpdHdr->partition_name));
    return FWU_ERROR_SUCCESS;
}

/*++
********************************************************************************
**
** FUNCTION:
**   DetermineImageType
**
** DESCRIPTION:
**   This function will determine the type of file
**
** ARGUMENTS:
**   InputFile            - Input File to read
**   imgType            - ENUM that stores the image type
**                                FULL = Full image with FPT, FTP, and NFTP
**                                PARTIAL = NO FPT, FTP, and NFTP in the file
**                                RESTORE =  NO FPT
**                                INVALID = None of the above types
**
** RETURNS:
**        imgType by call-by-reference
**        0:    FWU_ERROR_SUCCESS
**        8714: FWU_FILE_OPEN
**        8710: FWU_NO_MEMORY
**
** NOTE:
**
********************************************************************************
--*/
static UINT32
    DetermineImageType (
    IN  CHAR8          *buffer,
    IN  UINT32  bufLen,
    OUT IMAGE_TYPE    *imgType
    )
{
    fptt_ver1_t* fptTable = NULL;
    UINT32      PartitionName = 0;
    UINT32      ftp_nftp_found = 0;
    UINT32      fwuError = 0;
    UINT32      i = 0;

    if (NULL == buffer)
    {
        return FWU_INVALID_POINTER;
    }

    //read the Flash Partition Table
    fptTable = (fptt_ver1_t*)(buffer + FPT_FLASH_INITIAL_BASE);

    //Check for presence of FPT in the file
    if (FPT_PARTITION_NAME_FPT == fptTable->header.header_marker)
    {

        //Flag to determine both FTP and NFTP are present and valid
        ftp_nftp_found = 2;        //(0000 0010)
        for(i =0; i < fptTable->header.num_of_entries; i++)
        {
            if (FPT_PARTITION_NAME_FTPR == fptTable->entry[i].partition_name ||
                FPT_PARTITION_NAME_NFTP == fptTable->entry[i].partition_name)
            {
                DebugPrint(L"[DetermineImageType] PartitionName: 0x%x [%c%c%c%c]\n",
                           fptTable->entry[i].partition_name,
                           fptTable->entry[i].partition_name & 0xFF,
                           (fptTable->entry[i].partition_name >> 8) & 0xFF,
                           (fptTable->entry[i].partition_name >> 16) & 0xFF,
                           fptTable->entry[i].partition_name >> 24);
                DebugPrint(L"[DetermineImageType] OffsetFromFPT: 0x%x \n",fptTable->entry[i].offset);
                DebugPrint(L"[DetermineImageType] PartitionLength: 0x%x \n",fptTable->entry[i].length);
                DebugPrint(L"[DetermineImageType] Offset + PartitionLength: 0x%x \n", fptTable->entry[i].offset + fptTable->entry[i].length);
                DebugPrint(L"[DetermineImageType] BufferLength: 0x%x \n",bufLen);
                //If FTP or NFTP make sure that they are within the file size
                if(((fptTable->entry[i].offset + fptTable->entry[i].length) >  fptTable->entry[i].offset) &&
                    ((fptTable->entry[i].offset + fptTable->entry[i].length) <= bufLen))
                {
                    //Right shift the flag by one everytime you find a FTP or NFTP
                    ftp_nftp_found = ftp_nftp_found >> 1;
                }
            }

        } //for(int i =0;i<fptv2.NumFptEntries;i++)

        DebugPrint(L"[DetermineImageType] ftp_nftp_found: 0x%x \n",ftp_nftp_found);

        if (0 == ftp_nftp_found)
        { //both FTP and NFTP are present
            *imgType = FULL;
        }
        else
        {
            *imgType = INVALID;
        }
    }
    else
    {//No FPT in the Image

        //Read Partition name
        fwuError = ReadParitionNameManifestType((UCHAR*) buffer, &PartitionName);
        if(FWU_ERROR_SUCCESS != fwuError)
        {
            //FreePool(PartitionName);
            return fwuError;
        }

        if (FPT_PARTITION_NAME_WCOD == PartitionName ||
            FPT_PARTITION_NAME_LOCL == PartitionName ||
            FPT_PARTITION_NAME_ISHC == PartitionName ||
            FPT_PARTITION_NAME_IUNP == PartitionName ||
            FPT_PARTITION_NAME_IOMP == PartitionName ||
            FPT_PARTITION_NAME_MGPP == PartitionName ||
            FPT_PARTITION_NAME_TBTP == PartitionName)
        {   //if the image starts with IUP (wcod or locl ...)
            *imgType = PARTIAL;
        }
        else if (FPT_PARTITION_NAME_FTPR == PartitionName ||
            FPT_PARTITION_NAME_NFTP == PartitionName)
        {//if the image starts with FTP or NFTP
            *imgType = RESTORE;
        }
        else
        {//None of the above
            *imgType = INVALID;
        }
    }//if (FPT_PARTITION_NAME_FPT == fptHdr.HeaderMarker)

    return FWU_ERROR_SUCCESS;
}

/*++
********************************************************************************
**
** FUNCTION:
**   ExtractPartitionsToBuffer
**
** DESCRIPTION:
**   This function will use the loaded ImagePartition to extract and load the
**        partitions in expected order into buffer before sending it to FW
**
** ARGUMENTS:
**        filename            - Input File to read
**        Buffer            - Buffer to copy the partitions
**        imgPart            -    Data structure with required partition Information
**        NumParts            - Number of required partitions in the structure
**        imgType            - ENUM that stores the image type
**                                FULL = Full image with FPT, FTP, and NFTP
**                                PARTIAL = NO FPT, FTP, and NFTP in the file
**                                RESTORE =  NO FPT
**                                INVALID = None of the above types
**
** RETURNS:
**        DestBuffer by call-by-reference
**        0:    FWU_ERROR_SUCCESS
**        8806: FWU_INVALID_POINTER
**        8710: FWU_NO_MEMORY
**
** NOTE:
**
********************************************************************************
--*/
static UINT32
    ExtractPartitionsToBuffer (
    IN   CHAR8             *buffer,
    OUT  CHAR8             **DestBuffer,
    IN   UINT32     DestBufferLength,
    OUT  ImagePartition   *imgPart,
    OUT  UINT32     NumParts,
    OUT  IMAGE_TYPE       imgType
    )
{
    UINT32 i =0;
    UINT32 j = 0;
    UINT32  order = 1;
    UINT32  CurrentBufferPtr = 0;
    BOOLEAN          found = FALSE;

    if (NULL == buffer)
    {
        return FWU_INVALID_POINTER;
    }

    DebugPrint(L"\n----------------Image Order--------------\n");

    while (j < NumParts)
    {
        for (i =0;i<NumParts;i++)
        {
            if (order == imgPart[i].ImageOrder)
            {
                DebugPrint(L"\n*DestBuffer + CurrentBufferPtr: 0x%x\n", *DestBuffer + CurrentBufferPtr);
                DebugPrint(L"CurrentBufferPtr: 0x%x\n", CurrentBufferPtr);
                DebugPrint(L"Partition Name: 0x%x\n", imgPart[i].PartitionName);
                DebugPrint(L"Offset: 0x%x\n", imgPart[i].OffsetFromFPT);
                DebugPrint(L"Length: 0x%x\n", imgPart[i].PartitionLength);
                DebugPrint(L"UpvVersion: 0x%x\n", imgPart[i].UpvVersion);
                DebugPrint(L"Instance Id: 0x%x\n", imgPart[i].InstanceId);
                DebugPrint(L"Image Order: 0x%x\n", imgPart[i].ImageOrder);

                CopyMem(*DestBuffer + CurrentBufferPtr,
                    buffer + imgPart[i].OffsetFromFPT,
                    imgPart[i].PartitionLength);

                //Increment buffer pointer to copy the next partition
                CurrentBufferPtr = CurrentBufferPtr + imgPart[i].PartitionLength;

                //Current order is found
                found = TRUE;
                j++;
            } //if (order == imgPart[i].ImageOrder)
        } //for (i =0;i<NumParts;i++)

        if (!found)
        {
            DebugPrint(L"Image Order not found\n");
            //return FWU_ERROR_FW;
        }

        found = FALSE;
        order++;

    }// Loop Over J

    return FWU_ERROR_SUCCESS;

}

/*++
********************************************************************************
**
** FUNCTION:
**   GetPartVersion
**
** DESCRIPTION:
**   This function retrieves the current version of specific partition using Heci.
**
** ARGUMENTS:
**        major        - OUT parameter; Major version of the partition
**        major        - OUT parameter; Major version of the partition
**        hotfix       - OUT parameter; Hotfix version of the partition
**        build        - OUT parameter; Build version of the partition
**
** RETURNS:
**        status
**
********************************************************************************
--*/
int GetPartVersion(UINT32 partID, UINT16 *major, UINT16 *minor, UINT16 *hotfix, UINT16 *build)
{
    UINT32 status;
    MODULE_VERSION_DATA module_version_data = { 0 };
    status = GetImageFWVersionMsg(partID, &module_version_data);
    if (SUCCESS != status)
    {
        return status;
    }

    *major = module_version_data.module_version.Major;
    *minor = module_version_data.module_version.Minor;
    *hotfix = module_version_data.module_version.Hotfix;
    *build = module_version_data.module_version.Build;

    return FWU_ERROR_SUCCESS;
}

/*++
********************************************************************************
**
** FUNCTION:
**   GetPartVendorID
**
** DESCRIPTION:
**   This function retrieves the vendor ID specific partition using Heci.
**
** ARGUMENTS:
**        vendor_id    - OUT parameter; Vendor ID of the partition
**
** RETURNS:
**        status
**
********************************************************************************
--*/
int GetPartVendorID(UINT32 partID, UINT32 *vendor_id)
{
    UINT32 status;
    MODULE_VERSION_DATA module_version_data = { 0 };
    status = GetImageFWVersionMsg(partID, &module_version_data);
    if (SUCCESS != status)
    {
        return status;
    }

    *vendor_id = module_version_data.vendor_id;

    return FWU_ERROR_SUCCESS;
}

/*++
********************************************************************************
**
** FUNCTION:
**   GetUpvAndInstanceID
**
** DESCRIPTION:
**   This function will read the Partition name
**
** ARGUMENTS:
**        fp                    - file pointer to the input file
**        offset            - Offset within the file read manifest
**        PartitionName    - Buffer to copy the partition name
**        UpvVersion        - store output UPV Version
**        InstanceId        - store output Instance Id
**        RegionLength    - Store output Region Length
**
** RETURNS:
**        PartitionName by call-by-reference
**        UpvVersion by call-by-reference
**        InstanceId by call-by-reference
**        RegionLength by call-by-reference
**        0:    FWU_ERROR_SUCCESS
**        8714: FWU_FILE_OPEN
**        8710: FWU_NO_MEMORY
**
** NOTE:
**
********************************************************************************
--*/
static UINT32 GetUpvAndInstanceID (
    IN  UCHAR*         buffer,
    IN  UINT32 Offset,
    IN  UINT32  *PartitionName,
    OUT UINT32  *UpvVersion,
    OUT UINT32  *InstanceId,
    OUT UINT32  *RegionLength
    )

{
    mft_ifwi_part_man_ext_t* ifwiPartExt = NULL;
    manifest_header_t* manifestHdr = NULL;
    UINT32 manifestOffset = 0;
    UINT32 fwuError = 0;

    DebugPrint(L"\nGetUpvAndInstanceID()\n");
    DebugPrint(L"\nOffset: 0x%x\n", Offset);
    DebugPrint(L"\nPartitionName: 0x%x\n", PartitionName);


    //Set the file pointer position to start of the file
    if (NULL == buffer)
    {
        DebugPrint(L"Invalid buffer GetUpvAndInsatnceId\n");
        return FWU_FILE_OPEN;
        //return FWU_NULL_BUFFER;
    }

    manifestOffset = FindManifestOffset(buffer,Offset);

    if (INVALID_PARTITION_START == manifestOffset)
    {
        DebugPrint(L"Invalid manifest format at offset 0x%08X\n", Offset);
        return INVALID_MANIFEST_DATA;
    }

    manifestHdr = (manifest_header_t*)((UINT8 *)buffer + manifestOffset);

    // Obtain a pointer to the partition info extension
    FlashUpdateFindManifestExtension(MFT_EXT_TYPE_IFWI_PART_MAN, manifestHdr,(void **)&ifwiPartExt);
    if(!ifwiPartExt)
    {
        DebugPrint(L"Unable to get ifwi extension\n");
        return INVALID_MANIFEST_DATA;
    }

    fwuError = ReadParitionNameManifestType(buffer,PartitionName);

    if(FWU_ERROR_SUCCESS != fwuError)
    {
        DebugPrint(L"Unable to read partition name\n");
        return fwuError;
    }

    *UpvVersion   = (ifwiPartExt->partition_version_major << 16 ) | (ifwiPartExt->partition_version_minor);
    *InstanceId   = ifwiPartExt->instance_id;
    *RegionLength = ifwiPartExt->complete_partition_length;
    *PartitionName = ifwiPartExt->partition_name;

    DebugPrint(L"PartitionName: 0x%x\n", *PartitionName);
    DebugPrint(L"UpvVersion:    0x%x\n", *UpvVersion);
    DebugPrint(L"RegionLength:  0x%x\n", *RegionLength);
    DebugPrint(L"InstanceId:    0x%x\n", *InstanceId);
    DebugPrint(L"Offset:        0x%x\n",  Offset);

    return FWU_ERROR_SUCCESS;
}

/*++
********************************************************************************
**
** FUNCTION:
**   LoadUpvAndInstanceID
**
** DESCRIPTION:
**   This function will read the file for the non-runnable region to gather and
**        look for expected IPUs and copy them into output structure
**
** ARGUMENTS:
**        strFileName                    - the inout file
**        offset                        - Offset within the file read manifest
**        ExpectedPartitionName    - Expected partition name too look for
**        ExpectedInstanceId        - Expected InstanceId too look for
**        ExpectedUpvVersion        - Expected UpvVersion too look for
**        _imgPartition                - Structure to capture the found partitions
**        PartitionCount                - current size of the structure
**
** RETURNS:
**        _imgPartition by call-by-reference
**
**        0:    FWU_ERROR_SUCCESS
**        8714: FWU_FILE_OPEN
**        8710: FWU_NO_MEMORY
**
** NOTE:
**
********************************************************************************
--*/
static UINT32
    LoadUpvAndInstanceID (
    IN  CHAR8*           buffer,
    IN  UINT32    bufferLength,
    IN  UINT32   Offset,
    IN  UINT32    ExpectedPartitionName,
    IN  UINT32    ExpectedInstanceId,
    IN  UINT32    ExpectedUpvVersion,
    OUT ImagePartition  **_imgPartition,
    OUT UINT32    *PartitionCount,
    OUT UINT32    *foundPart,
    OUT UINT32    IsFullUpdate
    )

{
    UINT32    PartitionName = 0;
    UINT32    UpvVersion = 0;
    UINT32    InstanceId = 0;
    UINT32    RegionLength = 0;
    UINT32    fwuError = 0;
    UINT32    currentPartIdx = 0;
    BOOLEAN ignorePartition = FALSE;
    UINT32 partIdx = 0;

    *foundPart = 0;

    // Open input image file

    if (NULL == buffer)
    {
        DebugPrint(L"Invalid buffer LoadUpvAndInstanceId\n");
        return FWU_FILE_OPEN;
    }

    //Loop through the file until EOF
    while (Offset < bufferLength && !*foundPart)
    {
        //Initialize data structures for every iteration
        PartitionName = 0;
        UpvVersion = 0;
        InstanceId = 0;
        RegionLength = 0;

        //Read Partition Info from file starting from offset
        fwuError = GetUpvAndInstanceID((UCHAR*)buffer,Offset,
            &PartitionName,&UpvVersion,&InstanceId,&RegionLength);
        if (FWU_ERROR_SUCCESS != fwuError)
        {
            // This may not happen but for completion we check for error with
            // FlashUpdateFindManifestExtension  routine
            if (INVALID_MANIFEST_DATA == fwuError)
            {
                fwuError = FWU_GENERAL;
            }

            return fwuError;
        }

        DebugPrint(L"LoadUpvAndInstanceID ExpectedPartitionName: 0x%x \n", ExpectedPartitionName);
        DebugPrint(L"LoadUpvAndInstanceID ExpectedInstanceId: 0x%x \n", ExpectedInstanceId);
        DebugPrint(L"LoadUpvAndInstanceID PartitionName: 0x%x \n",PartitionName);
        DebugPrint(L"LoadUpvAndInstanceID InstanceId: 0x%x \n",InstanceId);
        DebugPrint(L"LoadUpvAndInstanceID UpvVersion: 0x%x \n",UpvVersion);
        DebugPrint(L"LoadUpvAndInstanceID ExpectedPartitionName: 0x%x\n", ExpectedPartitionName);
        DebugPrint(L"LoadUpvAndInstanceID ExpectedInstanceId: 0x%x\n", ExpectedInstanceId);
        DebugPrint(L"LoadUpvAndInstanceID PartitionName: 0x%x\n",PartitionName);
        DebugPrint(L"LoadUpvAndInstanceID InstanceId: 0x%x\n",InstanceId);
        DebugPrint(L"LoadUpvAndInstanceID UpvVersion: 0x%x\n",UpvVersion);

        if (IsFullUpdate) {
            //If it is Full update then we don't have to check for UpvVersion match
            //unless there was a partition with the same name in then runnable part
            //of the image, in which case we make sure that this partition has the
            //same UPV. If we were to allow UPV mismatch, the full update would fail
            //in the FW side.
            if ((PartitionName == ExpectedPartitionName) && (InstanceId == ExpectedInstanceId))
            {
                // Found a Matching Partition
                currentPartIdx = *PartitionCount;
                ignorePartition = FALSE;
                // check current list to see if there is already an entry for
                // this partition name.
                for (partIdx = 0; partIdx < *PartitionCount; partIdx++)
                {
                    if ((*_imgPartition)[partIdx].PartitionName == PartitionName)
                    {
                        currentPartIdx = partIdx;
                        DebugPrint(L"Found duplicate entry for this partition. Using the best candidate\n");
                        // Only allow matching UPVs at this point. The UPV comes from
                        // the active region in the update image.
                        if((*_imgPartition)[partIdx].UpvVersion != UpvVersion)
                        {
                            DebugPrint(L"This partition does not match UPD image expected UPV. Skipping partition.\n");
                            ignorePartition = TRUE;
                            break;
                        }
                        break;
                    }
                }
                // Here we are making sure we are not adding new partition entries
                // to _imgPartition. This is because during full update, we only
                // parse the non-runnable region to find matching instanceIDs.
                if ((ignorePartition) || (currentPartIdx == *PartitionCount))
                {
                    //Move to the next Region
                    Offset = Offset + RegionLength;

                    //Free up allocated buffer
                    /* if (PartitionName) {
                    FreePool (PartitionName);
                    }*/
                    continue;
                }

                (*_imgPartition)[currentPartIdx].PartitionName = PartitionName;
                (*_imgPartition)[currentPartIdx].OffsetFromFPT = Offset;
                (*_imgPartition)[currentPartIdx].PartitionLength = RegionLength;
                (*_imgPartition)[currentPartIdx].UpvVersion = UpvVersion;
                (*_imgPartition)[currentPartIdx].InstanceId = InstanceId;

                *foundPart = 1;
            }
        }
        else
        { //If it is Partial Update then check for UpvVersion match also.
            //If it is Partial Update then check for UpvVersion match also.
            if((PartitionName == ExpectedPartitionName) &&
                (InstanceId == ExpectedInstanceId) &&
                (UpvVersion>>16 == ExpectedUpvVersion>>16))
            {
                (*_imgPartition)[*PartitionCount].PartitionName = PartitionName;
                (*_imgPartition)[*PartitionCount].OffsetFromFPT = Offset;
                (*_imgPartition)[*PartitionCount].PartitionLength = RegionLength;
                (*_imgPartition)[*PartitionCount].UpvVersion = UpvVersion;
                (*_imgPartition)[*PartitionCount].InstanceId = InstanceId;
                (*_imgPartition)[*PartitionCount].ImageOrder = 0;

                *PartitionCount = *PartitionCount + 1;
                *foundPart = 1;

            }
        }
        DebugPrint(L"\nFOUND - PARTIAL Update\n");
        DebugPrint(L"PartitionName: 0x%x\n", PartitionName);
        DebugPrint(L"UpvVersion: 0x%x\n", UpvVersion);
        DebugPrint(L"RegionLength: 0x%x\n", RegionLength);
        DebugPrint(L"InstanceId: 0x%x\n", InstanceId);
        DebugPrint(L"Offset: 0x%x\n", Offset);
        DebugPrint(L"\n");

        // Move to the next Region
        Offset = Offset + RegionLength;

        //Free up allocated buffer
        /*if (PartitionName) {
        FreePool (PartitionName);
        }*/

    } // while( Offset < bufferLength)

    return FWU_ERROR_SUCCESS;
}

/*++
********************************************************************************
**
** FUNCTION:
**   LoadUpvAndInstanceID
**
** DESCRIPTION:
**   This function will read the file for the non-runnable region to gather and
**        look for expected IPUs and copy them into output structure
**
** ARGUMENTS:
**        strFileName              - the inout file
**        offset                   - Offset within the file read manifest
**        ExpectedPartitionName    - Expected partition name too look for
**        ExpectedInstanceId       - Expected InstanceId too look for
**        ExpectedUpvVersion       - Expected UpvVersion too look for
**        _imgPartition            - Structure to capture the found partitions
**        PartitionCount           - current size of the structure
**
** RETURNS:
**        _imgPartition by call-by-reference
**
**        0:    FWU_ERROR_SUCCESS
**        8714: FWU_FILE_OPEN
**        8710: FWU_NO_MEMORY
**
** NOTE:
**
********************************************************************************
--*/

static UINT32 LoadPreUpdatePartition(UCHAR* buffer,
    UINT32 bufferLength,
    UINT32 Offset,
    ImagePartition *imgPartition,
    UINT32 *foundPart)
{
    const UINT32 ExpectedPartitionName = PARTITION_NAME_UPDT;
    UINT32  PartitionName = 0;
    UINT32  partitionLength = 0;
    UINT32  UpvVersion = 0;
    UINT32  InstanceId = 0;
    UINT32 fwuError;

    *foundPart = 0;

    // Open input image file

    if(NULL == buffer)
    {
        DebugPrint(L"Invalid buffer \n");
        return FWU_FILE_OPEN;
    }

    //Loop through the file until EOF
    for (; Offset < bufferLength; Offset += partitionLength)
    {
        //Initialize data structures for every iteration
        PartitionName = PARTITION_NAME_UPDT;
        partitionLength = 0;

        //Read Partition Info from file starting from offset
        fwuError = GetUpvAndInstanceID(buffer, Offset, &PartitionName, &UpvVersion, &InstanceId, &partitionLength);
        if(FWU_ERROR_SUCCESS != fwuError)
        {
            if(INVALID_MANIFEST_DATA == fwuError)
            {
                //This is not a valid paratition, cannot continue scanning, assuming that such paratition does not exist.
                *foundPart = 0;
                return FWU_ERROR_SUCCESS;
            }
            DebugPrint(L"Error searching for UPDT starting at offset 0x%08X \n", Offset);
            return fwuError;
        }

        //If it is Full update then we don't have to check for UpvVersion match
        //unless there was a partition with the same name in then runnable part
        //of the image, in which case we make sure that this partition has the
        //same UPV. If we were to allow UPV mismatch, the full update would fail
        //in the FW side.
        if (PartitionName == ExpectedPartitionName)
        {
            // Found a Matching Partition
            imgPartition->PartitionName = PartitionName;
            imgPartition->OffsetFromFPT = Offset;
            imgPartition->PartitionLength = partitionLength;
            imgPartition->UpvVersion = UpvVersion;
            imgPartition->InstanceId = InstanceId;

            *foundPart = 1;

            DebugPrint(L"\nFOUND Pre Update Module\n");
            DebugPrint(L"PartitionName: 0x%x\n", PartitionName);
            DebugPrint(L"UpvVersion:    0x%x\n",  UpvVersion);
            DebugPrint(L"RegionLength:  0x%x\n",  partitionLength);
            DebugPrint(L"InstanceId:    0x%x\n",  InstanceId);
            DebugPrint(L"Offset:        0x%x\n\n",Offset);

            break;
        }
        if (partitionLength == 0)
        {
            //Not possible. will cause an infinit loop.
            return FWU_GENERAL;
        }

    }


    return FWU_ERROR_SUCCESS;
}

/*++
********************************************************************************
**
** FUNCTION:
**   DisplayBufferContent
**
** DESCRIPTION:
**   This function will display the buffer sent to FW in that order
**
** ARGUMENTS:
**        Buffer            - Buffer to copy the partitions
**        length            - Length of the buffer
**
** RETURNS:
**        Void
**
** NOTE:
**
********************************************************************************
--*/

#ifdef _DEBUG
static void DisplayBufferContent(CHAR8 **buffer, UINT32 length)
{
    mft_ifwi_part_man_ext_t* ifwiPartExt = NULL;
    manifest_header_t*   manifestHdr   = NULL;
    UINT32 partitionName = 0;
    UINT32 manifestOffset = 0;
    CHAR8* bufPtr = (CHAR8*)*buffer;
    CHAR8* bufEnd = ((CHAR8*)(*buffer) + length);

    Print(L"\n\n-----------------Current Buffer to FW : display Buffer content----------------\n");
    //Loop through the buffer
    while(bufPtr < bufEnd)
    {
        if(ReadParitionNameManifestType((UCHAR*)bufPtr, &partitionName))
            return;

        Print(L"\nName: 0x%08X\n", partitionName);

        manifestOffset = FindManifestOffset((UCHAR*)bufPtr,0);
        if (INVALID_PARTITION_START == manifestOffset)
            return;

        manifestHdr = (manifest_header_t*)((UINT8*)bufPtr + manifestOffset);

        // Obtain a pointer to the partition info extension
        FlashUpdateFindManifestExtension(MFT_EXT_TYPE_IFWI_PART_MAN, manifestHdr, (void **)&ifwiPartExt);
        if(!ifwiPartExt)
        {
            Print(L"Unable to find ifwi extension\n");
            return;
        }

        Print(L"UpvVersion:   0x%x\n", partInfoExt->partition_version_major << 16 | partInfoExt->partition_version_minor);
        Print(L"RegionLength: 0x%x\n", partInfoExt->complete_partition_length);
        Print(L"InstanceId:   0x%x\n", partInfoExt->instance_id);
        bufPtr += partInfoExt->complete_partition_length;
    }
}
#endif /* _DEBUG */

/*++

Routine Description:
The ME FW when created has an OEM ID embedded in it.  For security
reasons, we need to verify that the OEM ID of the application (mOemID)
matches the OEM ID of the ME FW.

NOTE: A connection to the client is required before this routine is
executed.

Arguments:

UUID to check


Returns:

TRUE or FALSE indicating if the match was succesful

--*/
BOOLEAN VerifyOemId( IN _UUID  id)
{
    EFI_STATUS                 Status = EFI_SUCCESS;
    FWU_VERIFY_OEMID_MSG Msg = {0};
    CHAR8 RxBuf[1024] = {0};
    FWU_HECI_MSG *ReplyMsg = NULL;

    if (!mInitialized)
    {
        Status = GetFwVersion ();
        if (EFI_ERROR (Status))
        {
            DebugPrint(L"Unable to get current FW version - %r\n", Status);
            return FALSE;
        }
    }

    ZeroMem(&Msg, sizeof(Msg));
    Msg.MessageType = FWU_VERIFY_OEMID;
    CopyMem (&Msg.OemId, &id, sizeof(_UUID));

    Status = SendMessageWithFlow(&MeClients[FWU_CLIENT], (UINT8 *)&Msg, sizeof(Msg));
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable send message to VERIFY OEMID - %r\n", Status);
        return FALSE;
    }

    //
    // Before receiving, we send a flow control to the ME indicating we have a buffer
    //
    /*Status = SendFlowControl (&MeClients[FWU_CLIENT]);
    if (EFI_ERROR (Status))
    {
        return FALSE;
    }*/

    Status = ReadMessageWithFlow (&MeClients[FWU_CLIENT], RxBuf, sizeof(RxBuf));
    ReplyMsg = (FWU_HECI_MSG *)RxBuf;

    if (ReplyMsg->MessageData.MessageType != FWU_VERIFY_OEMID_REPLY)
    {
        DebugPrint(L"Version Reply Message Type Incorrect: %x\n", ReplyMsg->MessageData.MessageType);
        Status = EFI_PROTOCOL_ERROR;
        return FALSE;
    }

    DebugPrint(L"VerifyOem Reply Status: %x\n", ReplyMsg->MessageData.VerifyOemIdReply.Status);

    if (ReplyMsg->MessageData.VerifyOemIdReply.Status != FWU_STATUS_SUCCESS)
    {
        switch(ReplyMsg->MessageData.VerifyOemIdReply.Status)
        {
        case STATUS_INVALID_OEM_ID:
            Status = EFI_ACCESS_DENIED;
            break;
        case STATUS_UPDATE_GENERAL_FAILURE: //fall through
        default:
            Status = EFI_DEVICE_ERROR;
            break;
        }
        return FALSE;
    }

    return TRUE;
}

/*++

Routine Description:

Get the last status of the FW Update Client - this is used to determine if a previous update
requires a reboot

Arguments:

UINT32 pointer for results of previous update.

Returns:

EFI Error code

--*/
EFI_STATUS GetLastStatus (OUT UINT32            *LastFwUpdateStatus)
{
    EFI_STATUS      Status = EFI_SUCCESS;

    Status = GetFWUpdateClientParams();
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable to get current FW version - %r\n", Status);
        return EFI_VOLUME_CORRUPTED;
    }

    *LastFwUpdateStatus = mLastFwUpdateStatus;

    return Status;
}


/*++

Routine Description:

Get the last status of the FW Update Client - this is used to determine if a previous update
requires a reboot

Arguments:

UINT32 pointer for results of previous update.

Returns:

EFI Error code

--*/
EFI_STATUS GetLastUpdateResetType (OUT UINT32            *LastResetType)
{
    EFI_STATUS      Status = EFI_SUCCESS;

    Status = GetFWUpdateClientParams();
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable to get current FW version - %r\n", Status);
        return EFI_VOLUME_CORRUPTED;
    }

    *LastResetType = mLastResetType;

    return Status;
}

/*++

Routine Description:

Returns the interfaces supported by the ME

Arguments:

Returns:

--*/
EFI_STATUS GetInterfaces (OUT UINT16        *Interfaces)
{
    EFI_STATUS      Status = EFI_SUCCESS;

    Status = GetFWUpdateClientParams();
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable to get current FW version - %r\n", Status);
        return EFI_VOLUME_CORRUPTED;
    }

    *Interfaces = mInterfaces;

    return Status;
}

/*++

Routine Description:

Get the FW version of the currently running ME FW

NOTE: A connection to the client is required before this routine is
executed.

Arguments:

Returns:

--*/
static EFI_STATUS GetFwVersion()

{
    EFI_STATUS                Status = EFI_SUCCESS;
    FWU_GET_INFO_MESSAGE      GetInfoMsg = {0};
    UINT8                     Buffer[1024] = {0};
    FWU_HECI_MSG              *ReplyMsg = NULL;

    DebugPrint(L"Fw version mInitialized: %d\n", mInitialized);

    Status = ConnectFWU ();

    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Cannot connect to clients - %r\n", Status);
        return Status;
    }
    GetInfoMsg.MessageType = FWU_GET_VERSION;

    //Added GetFlowControl. Initially it was not
    /*Status = GetFlowControl(MeClient);
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Status GetFLowControl(MeCLient) : %r \n", Status);
        return Status;
    }*/

    Status = SendMessageWithFlow (&MeClients[FWU_CLIENT], (UINT8 *)&GetInfoMsg, sizeof(FWU_GET_INFO_MESSAGE));
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable send message FWU_START - %r\n", Status);
        return Status;
    }

    DebugPrint(L"Sent FWU_GET_VERSION - %r\n", Status);

    //
    // Before receiving, we send a flow control to the ME indicating we have a buffer
    //
    //Try Remove this
    /*Status = SendFlowControl (&MeClients[FWU_CLIENT]);
    if (EFI_ERROR (Status))
    {
        return Status;
    }*/

    Status = ReadMessageWithFlow (&MeClients[FWU_CLIENT], (UINT8 *)Buffer, MeClients[FWU_CLIENT].MaxMessageLength);
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable read message from FWU_GET_VERSION - %r\n", Status);
        return Status;
    }

    ReplyMsg = (FWU_HECI_MSG *)Buffer;

    if (ReplyMsg->MessageData.MessageType != FWU_GET_VERSION_REPLY)
    {
        DebugPrint(L"Version Reply Message Type Incorrect: %x\n", ReplyMsg->MessageData.MessageType);
        Status = DisConnectFWU();
        if (EFI_ERROR (Status))
        {
            DebugPrint(L"Cannot connect to clients - %r\n", Status);
            return Status;
        }
        Status = EFI_PROTOCOL_ERROR;
        mInitialized = FALSE;
        return Status;
    }

    DebugPrint(L"Version Reply Status: %x\n", ReplyMsg->MessageData.VersionReply.Status);

    if (ReplyMsg->MessageData.VersionReply.Status == FWU_STATUS_SUCCESS)
    {
        mVer.code.Major = ReplyMsg->MessageData.VersionReply.CodeVersion.Major;
        mVer.code.Minor = ReplyMsg->MessageData.VersionReply.CodeVersion.Minor;
        mVer.code.Hotfix = ReplyMsg->MessageData.VersionReply.CodeVersion.Hotfix;
        mVer.code.Build = ReplyMsg->MessageData.VersionReply.CodeVersion.Build;
        DebugPrint(L"\n Currently Flashed Version: %d.%d.%d.%d \n", mVer.code.Major, mVer.code.Minor,mVer.code.Hotfix,mVer.code.Build);

        mDataFormatVersion = ReplyMsg->MessageData.VersionReply.DataFormatVersion;
        //mHwSku.PchVer = ReplyMsg->MessageData.VersionReply.PCHVer;

        //mVendor = ReplyMsg->MessageData.VersionReply.Vendor;

        //DebugPrint(L"\n HW SKu PCH Version and Vendor : 0x%x 0x%x \n", mHwSku.PchVer, mVendor);

        mInterfaces = ReplyMsg->MessageData.VersionReply.EnabledUpdateInterfaces;
        mLastFwUpdateStatus = ReplyMsg->MessageData.VersionReply.LastFwUpdateStatus;
        mLastResetType = ReplyMsg->MessageData.VersionReply.LastUpdateResetType;
        mInitialized = TRUE; // we have the variables, no need to call this routine again
    }
    else
    {
        Status = DisConnectFWU();
        if (EFI_ERROR (Status))
        {
            DebugPrint(L"Cannot connect to clients - %r\n", Status);
            return Status;
        }
        Status = EFI_DEVICE_ERROR;
        mInitialized = FALSE;
    }

    return Status;
}

/*++
********************************************************************************
**
** FUNCTION:
**   CheckPolicyBuffer
**
** DESCRIPTION:
**   This function will determine to upgrade or downgrade and check for the right
**        data format version.
**
** ARGUMENTS:
**
** RETURNS:
**
********************************************************************************
--*/
UINT32 CheckPolicyBuffer (
    IN  CHAR8*       buffer,
    IN  int         bufferLength,
    IN  int         AllowSV,
    IN  UPDATE_TYPE *Upd_Type,
    IN  VersionLib  *ver
    )
{
    UINT32    fwuError = 0;

    fwuError = _CheckPolicy (AllowSV, (UCHAR*)buffer,&Upd_Type,&ver);

    if (FWU_ERROR_SUCCESS != fwuError)
    {
        return fwuError;
    }

    return FWU_ERROR_SUCCESS;
}

// Check for end status when CSE return 127 during FWSTS polling
static UINT32 _CheckFwuEndErrorStatus(UINT32* lastResetType)
{
    UINT32 status = SUCCESS;
    FWU_END_MSG_REPLY * msgReply = NULL;
    UINT32 outBuffSize = sizeof(FWU_END_MSG_REPLY);
    EFI_STATUS Status_efi = EFI_SUCCESS;

    if (NULL == lastResetType)
    {
        DebugPrint(L"Invalid lastResetType Pointer \n");
        status = STATUS_UPDATE_OUT_OF_RESOURCES;
        goto end;
    }

    DebugPrint(L"Heci max buffer size: %d\n", MeClients[FWU_CLIENT].MaxMessageLength);
    msgReply = (FWU_END_MSG_REPLY *)AllocatePool(MeClients[FWU_CLIENT].MaxMessageLength);
    if (!msgReply)
    {
        status = STATUS_UPDATE_OUT_OF_RESOURCES;
        goto end;
    }

    ZeroMem(msgReply, MeClients[FWU_CLIENT].MaxMessageLength);

    Status_efi = ReadMessageWithFlow(&MeClients[FWU_CLIENT], (UINT8 *)(msgReply), outBuffSize);
    if (EFI_ERROR(Status_efi))
    {
        status = STATUS_UPDATE_NOT_READY;
        goto end;
    }

    status = msgReply->Status;
    *lastResetType = msgReply->ResetType;

end:
    FreePool(msgReply);
    return status;
}

/*++
********************************************************************************
**
** FUNCTION:
**   FWUpdate_QueryStatus_Get_Response
**
** DESCRIPTION:
**   This function queries the ME Kernel for the update status
**
** ARGUMENTS:
**
** RETURNS:
**
** NOTE:
**
********************************************************************************
--*/

UINT32
    FWUpdate_QueryStatus_Get_Response (
    IN  UINT32  *UpdateStatus,
    IN  UINT32  *TotalStages,
    IN  UINT32  *PercentWritten,
    IN  UINT32  *LastUpdateStatus,
    IN  UINT32  *LastResetType
    )
{
    UINT32 fwuError = 0;
    heci1_cse_gs1_t fwstatus = {0};
    UINT32 fwsts2 = 0;
    UINT32 status = 0;
    UINT32 fwstsbitsmax = 0x7F; //(0x7F represnts Error value)
    EFI_STATUS Status_efi = EFI_SUCCESS;

    *LastResetType = 0x0;


    if (!gPciRootBrdgIo)
    {
        EFI_STATUS PciInitStatus = 0;
        PciInitStatus = gBS->LocateProtocol(&gEfiPciRootBridgeIoProtocolGuid, NULL, (VOID **)&gPciRootBrdgIo);

        if (EFI_ERROR(PciInitStatus))
        {
            DebugPrint(L"\n Failed to register EFI PCI protocol \n");
            fwuError = FWU_IME_NOT_READY;
            return fwuError;
        }
    }
    if (!mInitialized) {
        Status_efi = GetFwVersion ();
        if (EFI_ERROR(Status_efi)) {
            DebugPrint(L"Unable to get current FW version - %r\n", Status);
            return FWU_IME_NOT_READY;
        }
    }

    status = ReadFwStatusRegValue(2, (UINT32*)&fwsts2);
    if (status)
    {
        DebugPrint(L"Error occurred GetFwStatus: %x\n", fwsts2);
        fwuError = FWU_IME_NOT_READY;
        /* *** TODO: why comment out?
        Status_efi = DisConnectFWU();
        if (EFI_ERROR(Status_efi))
        {
            DebugPrint(L"Cannot connect to clients - %r\n", Status_efi);
            return FWU_IME_NOT_READY;
        }
        mInitialized = FALSE;*/
        return fwuError;
    }

    fwstatus.ul = fwsts2;
    *UpdateStatus = fwstatus.r.phase;

    //Adding debug messages, see note below
    DebugPrint("we are beffor the IF states:\n");
    DebugPrint("fwstatus.r.phase:%x\n", fwstatus.r.phase);
    DebugPrint("GetFwStatus:%x\n", fwstatus.ul);
    DebugPrint("GetFwStatus:PercentWritten %d\n", fwstatus.r.status & fwstsbitsmax);

    //during the progress of the FW update process we sould receive: fwstsbitsmax != (fwstatus.r.status & fwstsbitsmax)
    //however if we get fwstsbitsmax == (fwstatus.r.status & fwstsbitsmax) (e.g.0x7f) it means we have an error

    if (fwstsbitsmax != (fwstatus.r.status & fwstsbitsmax) &&
        (HECI1_CSE_GS1_PHASE_FWUPDATE == fwstatus.r.phase))
    {
        *LastUpdateStatus = STATUS_UPDATE_SUCCESS;
        fwuError = FWU_ERROR_SUCCESS;
    }
    //we are still at FW-UPDATE Phase, checking what's the FW status and exit
    else if(HECI1_CSE_GS1_PHASE_FWUPDATE == fwstatus.r.phase)
    {
        DebugPrint("Maybe Wrong Heci1 state:%x\n", fwstatus.r.phase);
        status = _CheckFwuEndErrorStatus(LastResetType);
        *LastUpdateStatus = status;
        fwuError = FromCseToFwuErrorCode(status);
        /* *** TODO: why comment out?
        if(Status_efi != STATUS_UPDATE_NOT_READY)
        {
            Status = DisConnectFWU();
            if (EFI_ERROR(Status_efi))
            {
                DebugPrint(L"Cannot connect to clients - %r\n", Status_efi);
                return FWU_IME_NOT_READY;
            }
            mInitialized = FALSE;
        }*/
        return fwuError;
    }
    //we are NOT at FW-UPDATE Phase, checking what's the FW status and exit
    else
    {
        DebugPrint(L"Wrong Heci1 state:%x\n", fwstatus.r.phase);
        status = _CheckFwuEndErrorStatus(LastResetType);
        *LastUpdateStatus = status;
        fwuError = FromCseToFwuErrorCode(status);
        /* *** TODO: why comment out?
        if (status != STATUS_UPDATE_NOT_READY)
        {
            Status_efi = DisConnectFWU();
            if (EFI_ERROR(Status_efi))
            {
                DebugPrint(L"Cannot connect to clients - %r\n", Status_efi);
                return FWU_IME_NOT_READY;
            }
            mInitialized = FALSE;
        }*/
        return fwuError;
    }

    DebugPrint(L"GetFwStatus:%x\n", fwstatus.ul);
    DebugPrint(L"GetFwStatus:PercentWritten %d\n", fwstatus.r.status & fwstsbitsmax);

    *PercentWritten = (fwstatus.r.status & fwstsbitsmax);

    *TotalStages = 0;

    if (*PercentWritten == 100)
    {
        status = _CheckFwuEndErrorStatus(LastResetType);
        *LastUpdateStatus = status;

        if (STATUS_UPDATE_SUCCESS != status)
        {
            fwuError = FromCseToFwuErrorCode(status);
        }
        else
        {
            fwuError = FWU_ERROR_SUCCESS;
        }

        Status_efi = DisConnectFWU();
        if (EFI_ERROR(Status_efi))
        {
            DebugPrint(L"Cannot connect to clients - %r\n", Status_efi);
            return FWU_IME_NOT_READY;
        }
        mInitialized = FALSE;
    }
    return fwuError;
}

/*++
********************************************************************************
**
** FUNCTION:
**   GetIpuPartitionAttributes
**
** DESCRIPTION:
**   This function gets the attributes of the partition
**
** ARGUMENTS:
**
** RETURNS:
**
** NOTE:
**
********************************************************************************
--*/

UINT32
    GetExtendedIpuPartitionAttributes (
    IN  FWU_GET_IPU_PT_ATTRB_MSG_REPLY  *FwuGetIpuAttrbMsgInfo, UINT32 updateOp
    )
{
    EFI_STATUS                  Status = EFI_SUCCESS;
    FWU_GET_IPU_PT_ATTRB_MSG        IpuPtAttribMsg = {0};
    UINT8                       Buffer[1024] = {0};
    FWU_HECI_MSG                *ReplyMsg = {0};
    UINT32 fwuError = 0;
#ifdef _DEBUG
    UINT32 i =0;
#endif

    //Diffreant
    //SetMem (&IpuPtAttribMsg, 0, sizeof (IpuPtAttribMsg);
    ZeroMem(&IpuPtAttribMsg,sizeof (IpuPtAttribMsg));

    IpuPtAttribMsg.MessageType = FWU_GET_IPU_PT_ATTRB;
    IpuPtAttribMsg.UpdateType = updateOp;
    //Diffreant
    Status = SendMessageWithFlow (&MeClients[FWU_CLIENT], (UINT8 *)&IpuPtAttribMsg, sizeof(FWU_GET_IPU_PT_ATTRB_MSG));
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable send message FWU_START - %r\n", Status);
        return FWU_STATUS_INTERNAL_ERROR;
    }

    DebugPrint(L"Sent FWU_GET_IPU_PT_ATTRB - %r\n", Status);

    //
    // Before receiving, we send a flow control to the ME indicating we have a buffer
    //
    /*Status = SendFlowControl (&MeClients[FWU_CLIENT]);
    if (EFI_ERROR (Status))
    {
        return FWU_STATUS_INTERNAL_ERROR;
    }*/

    Status = ReadMessageWithFlow (&MeClients[FWU_CLIENT], (UINT8 *)Buffer, MeClients[FWU_CLIENT].MaxMessageLength);
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable read message from FWU_GET_IPU_PT_ATTRB - %r\n", Status);
        return FWU_STATUS_INTERNAL_ERROR;
    }
    //ReplyMsg.UpdateType = updateOp;

    ReplyMsg = (FWU_HECI_MSG *)Buffer;
    if (ReplyMsg->MessageData.MessageType != FWU_GET_IPU_PT_ATTRB_REPLY)
    {
        DebugPrint(L"Version Reply Message Type Incorrect: %x\n", ReplyMsg->MessageData.MessageType);
        fwuError = FWU_GET_PATTRIB_ERROR;
        return fwuError;
    }

    DebugPrint(L"Version Reply Status: %x\n", ReplyMsg->MessageData.VersionReply.Status);

    if (ReplyMsg->MessageData.VersionReply.Status != FWU_STATUS_SUCCESS)
    {
#if BYPASS_FWU_GET_IPU_PT_ATTRB
        return FWU_STATUS_SUCCESS;
#endif
        fwuError = FromCseToFwuErrorCode(ReplyMsg->MessageData.VersionReply.Status);
        //FreeMemory(rxBuff);
        //HeciUninitialize(&fwuclt);

        return fwuError;
    }

    // Copy the Heci Message Info into appropriate data structures
    FwuGetIpuAttrbMsgInfo->MessageType =
        ReplyMsg->MessageData.GetIpuPtAttrbReply.MessageType;
    FwuGetIpuAttrbMsgInfo->Status = ReplyMsg->MessageData.GetIpuPtAttrbReply.Status;
    FwuGetIpuAttrbMsgInfo->FtpFwVer =
        ReplyMsg->MessageData.GetIpuPtAttrbReply.FtpFwVer;
    FwuGetIpuAttrbMsgInfo->NumOfPartition =
        ReplyMsg->MessageData.GetIpuPtAttrbReply.NumOfPartition;
    FwuGetIpuAttrbMsgInfo->SizeoOfPtAttrib =
        ReplyMsg->MessageData.GetIpuPtAttrbReply.SizeoOfPtAttrib;

#ifdef _DEBUG
    for (i = 0; i < ReplyMsg->MessageData.GetIpuPtAttrbReply.NumOfPartition; i++)
    {
        DebugPrint(L"ExpectedUpvVersion PartID %d: 0x%x \n",i,ReplyMsg->MessageData.GetIpuPtAttrbReply.PtAttribute[i].ExpectedUpvVer);
        DebugPrint(L"ExpectedInstId PartID %d: 0x%x \n", i, ReplyMsg->MessageData.GetIpuPtAttrbReply.PtAttribute[i].ExpectedInstId);
    }
#endif

    //Copy method diff
    /* CopyMem (FwuGetIpuAttrbMsgInfo->PtAttribute,
                ReplyMsg->MessageData.GetIpuPtAttrbReply.PtAttribute,
                sizeof(PT_ATTRB)*MAXIMUM_IPU_SUPPORTED);*/
    CopyMem(FwuGetIpuAttrbMsgInfo->PtAttribute,ReplyMsg->MessageData.GetIpuPtAttrbReply.PtAttribute,sizeof(PT_ATTRB)*MAXIMUM_IPU_SUPPORTED);

    if(FwuGetIpuAttrbMsgInfo->PtAttribute)
    {
        FreePool(FwuGetIpuAttrbMsgInfo->PtAttribute);
    }

    return FWU_ERROR_SUCCESS;
}

UINT32 GetIpuPartitionAttributes(FWU_GET_IPU_PT_ATTRB_MSG_REPLY* FwuGetIpuAttrbMsgInfo)
{
    return GetExtendedIpuPartitionAttributes(FwuGetIpuAttrbMsgInfo, FWU_FULL_UPDATE_OPERATION);
}

/*++
********************************************************************************
**
** FUNCTION:
**   FwUpdatePartialBuffer
**
** DESCRIPTION:
**   This function sends the data to the FW Update client
**
** ARGUMENTS:
**
** RETURNS:
**
** NOTE:
**
********************************************************************************
--*/
UINT32
    FwUpdatePartialBuffer (
    IN  CHAR8             *buffer,
    IN  UINT32            bufferLength,
    IN  UINT32            PartitionID,
    IN UINT32             Flags,
    IN IPU_UPDATED_INFO   *IpuUpdatedInfo,
    IN  CHAR8              *_pwd,
    //IN  int             _forceResetLib,
    IN  UINT32            UpdateEnvironment,
    IN  _UUID             OemID,
    IN  UPDATE_FLAGS_LIB  update_flags,
    IN  void(*func)(float, float)
    )
{
    FWU_GET_IPU_PT_ATTRB_MSG_REPLY  FwuGetIpuAttrbMsgInfo = {0};
    ImagePartition                  *imgPartitions = NULL;
    UINT32                          PartitionName = 0;
    IMAGE_TYPE                      imgType = {0};
    UINT32                          StartOffsetAfterLastActivePartition =0;
    UINT32                          UpvVersion = 0;
    UINT32                          InstanceId = 0;
    UINT32                          RegionLength = 0;
    UINT32                          PartitionCount=0;
    UINT32                          ExpectedNumOfPartitions = 0;
    UINT32                          ExpectedPartitionId = 0;
    UINT32                          ExpectedInstanceId = 0;
    UINT32                          ExpectedUpvVersion = 0;
    UINT32                          totalBufSize = 0;
    UINT32                          ElementCount = 0;
    CHAR8                           *fwUpdBuffer = NULL;
    UINT32                          fwuError = 0;
    UINT32                          i, j;
    EFI_STATUS                      Status = EFI_SUCCESS;
    UINT32                          foundPart = 0;
    BOOLEAN                         foundExpected = FALSE;

    if (!mInitialized)
    {
        Status = GetFwVersion ();
        if (EFI_ERROR (Status))
        {
            DebugPrint(L"Unable to get current FW version - %r\n", Status);
            return FWU_IME_NOT_READY;
        }
    }

    //Determine the type of Image
    if (FWU_ERROR_SUCCESS != (fwuError = (UINT32)DetermineImageType(buffer,bufferLength, &imgType)))
    {
        return fwuError;
    }

    //For Partial update accept only FULL or PARTIAL Image
    if (imgType == RESTORE || imgType == INVALID)
    {
        DebugPrint(L"\nImageType Failed: 0x%x\n", imgType);
        return FWU_FILE_INVALID;
    }

    //Get Partition Attribute from Firmware
    fwuError = GetExtendedIpuPartitionAttributes(&FwuGetIpuAttrbMsgInfo,FWU_IPU_UPDATE_OPERATION);
    if (FWU_ERROR_SUCCESS != fwuError)
    {
        //Heci gets deinitialized on both error or success in GetIpuPartitionAttributes
        return fwuError;
    }

    /*if (NULL == g_fptt)
    { //Image with FPT is required
        return FWU_FILE_INVALID;
    }*/
    if(0 != PartitionID)
    {//If Partition ID requested
        ExpectedNumOfPartitions = 1; //only one partition requested

        //Search for requested Partition ID in expected list from FW
        for(j=0;j<FwuGetIpuAttrbMsgInfo.NumOfPartition;j++)
        {
            if(PartitionID == FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId)
            {//Determine the expected IDs
                ExpectedPartitionId = FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId;
                ExpectedInstanceId = FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedInstId;
                ExpectedUpvVersion = FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedUpvVer;

                foundExpected = TRUE;

                DebugPrint(L"Expected PartitionId %d : 0x%x \n", j, FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId);
                DebugPrint(L"Expected ExpectedInstId %d : 0x%x \n",j,FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedInstId);
                DebugPrint(L"Expected ExpectedUpvVer %d : 0x%x \n",j,FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedUpvVer);
            }
        }

        //the passed in PID must be expected as per the FW
        if(!foundExpected)
        {//IF not found error out
            return FWU_PID_NOT_EXPECTED;
        }
    }
    else
    {//No Partition ID provided
        //Bug Fixed: 192029 removing support for "-ipu"
        return FWU_INVALID_PARTID;
    }

    //Allocate memory to capture the required partition information for send to FW
    imgPartitions =
        (ImagePartition*)AllocatePool(ExpectedNumOfPartitions*sizeof(ImagePartition));
    if(NULL == imgPartitions)
    {
        DebugPrint(L"No Memory Error\n");
        return FWU_NO_MEMORY;
    }
    ZeroMem(imgPartitions,ExpectedNumOfPartitions*sizeof(ImagePartition));


    PartitionCount=0;
    if(imgType == FULL)
    {//IF image provided is full then search the Runnable part of the image
        //Read FPT and check for its presence in the image
        fwuError = ParsingImageForFPT((UCHAR*)buffer,&StartOffsetAfterLastActivePartition);
        if(FWU_ERROR_SUCCESS != fwuError)
        {
            DebugPrint(L"ParsingImageForFPT Failed\n");
            //FreeMemory(imgPartitions);
            return fwuError;
        }
        DebugPrint(L"MARK StartOffsetAfterLastActivePartition: 0x%x\n",StartOffsetAfterLastActivePartition);

        //Loop through the list of entries from the image to find
        // the right IUP Instance ID.
        for(i=0; i<g_fptt->header.num_of_entries; i++)
        {//Trying to find WCOD/LOCL/ISHC
            if (g_fptt->entry[i].length != 0 &&
                (FPT_PARTITION_NAME_WCOD == g_fptt->entry[i].partition_name ||
                 FPT_PARTITION_NAME_LOCL == g_fptt->entry[i].partition_name ||
                 FPT_PARTITION_NAME_MDMV == g_fptt->entry[i].partition_name ||
                 FPT_PARTITION_NAME_IOMP == g_fptt->entry[i].partition_name ||
                 FPT_PARTITION_NAME_MGPP == g_fptt->entry[i].partition_name ||
                 FPT_PARTITION_NAME_TBTP == g_fptt->entry[i].partition_name ||
                 FPT_PARTITION_NAME_IUNP == g_fptt->entry[i].partition_name ||
                 FPT_PARTITION_NAME_ISHC == g_fptt->entry[i].partition_name))
            {
                DebugPrint(L"\nEntry 0x%x, Lpartition_name: 0x%x [%c%c%c%c]\n",
                           i,
                           g_fptt->entry[i].partition_name,
                           g_fptt->entry[i].partition_name & 0xFF,
                           (g_fptt->entry[i].partition_name >> 8) & 0xFF,
                           (g_fptt->entry[i].partition_name >> 16) & 0xFF,
                           g_fptt->entry[i].partition_name >> 24);
                DebugPrint(L"Entry 0x%x, partition_offet: 0x%x\n",i, g_fptt->entry[i].offset);

                //One file
                if (NULL == buffer)
                {
                    DebugPrint(L"Invalid Buffer \n");
                    //FreeMemory(imgPartitions);
                    fwuError = FWU_FILE_OPEN;
                    //fwuError = FWU_NULL_BUFFER;
                    return fwuError;
                }

                //get the Partition Info from the image for the current offset
                fwuError = GetUpvAndInstanceID((UCHAR*)buffer, g_fptt->entry[i].offset,
                    &PartitionName,&UpvVersion,&InstanceId,&RegionLength);
                if(FWU_ERROR_SUCCESS != fwuError)
                {
                    //FreeMemory(imgPartitions);

                    if (INVALID_MANIFEST_DATA == fwuError)
                    {
                        fwuError = FWU_GENERAL;
                    }
                    return fwuError;
                }

                if (0 != PartitionID)
                {//Expected partition info is passed to the function
                    if (InstanceId == ExpectedInstanceId &&
                        PartitionName == ExpectedPartitionId &&
                        UpvVersion >> 16 == ExpectedUpvVersion >> 16)
                    {
                        imgPartitions[PartitionCount].PartitionName =
                            g_fptt->entry[i].partition_name;
                        imgPartitions[PartitionCount].OffsetFromFPT =
                            g_fptt->entry[i].offset;
                        imgPartitions[PartitionCount].PartitionLength =
                            g_fptt->entry[i].length;
                        imgPartitions[PartitionCount].UpvVersion = UpvVersion;
                        imgPartitions[PartitionCount].InstanceId = InstanceId;

                        imgPartitions[PartitionCount].ImageOrder =  1;
                        //Note: the ImageOrder field will be zero for partition
                        //other than FTP and NFTP

                        PartitionCount++;
                    }
                }
            }//if(FPT_PARTITION_NAME_WCOD == fptt->entry[i].partition_name
        }// for(int i=0;i<fptHdr.num_of_entries;i++)
    }//if(imgType == FULL)

    //Count number of non duplicate elements in structure
    ElementCount = _countNonDuplicates(imgPartitions,PartitionCount);
    DebugPrint(L"ElementCount: %d\n",ElementCount);

    //Check to see if found the expected number of Partitions
    if(ElementCount < ExpectedNumOfPartitions)
    {//meaning all expected partitions are not found yet
        //so begin searching the NON-RUNNABLE part of the Image

        if (0 != PartitionID)
        {
            fwuError = LoadUpvAndInstanceID(buffer,bufferLength,
                StartOffsetAfterLastActivePartition,
                ExpectedPartitionId,
                ExpectedInstanceId,
                ExpectedUpvVersion,
                &imgPartitions,&PartitionCount, &foundPart,0);
            if (FWU_ERROR_SUCCESS != fwuError)
            {
                //FreeMemory(imgPartitions);
                return fwuError;
            }

            if (foundPart)
            {
                imgPartitions[PartitionCount-1].ImageOrder =  1;
            }
        }
    }

    //Count number of non duplicate elements in structure
    ElementCount = _countNonDuplicates(imgPartitions,PartitionCount);

    if(ElementCount != ExpectedNumOfPartitions)
    {//Error Case not all expected Partitions are found
        DebugPrint(L"ElementCount: %d\tExpectedNumOfPartitions: %d\n",ElementCount,ExpectedNumOfPartitions);
        //FreeMemory(imgPartitions);
        return FWU_FILE_INVALID;
    }

    //Compute Required Buffer Size to Store required and found Partitions
    for(i=0;i<PartitionCount;i++)
    {
        totalBufSize = totalBufSize + imgPartitions[i].PartitionLength;
    }

    DebugPrint(L"TotalBufSize: %d\n",totalBufSize);

    //allocate memory space to hold all partitions to be sent to FW
    fwUpdBuffer = (CHAR8*)AllocatePool(totalBufSize);
    if(NULL == fwUpdBuffer)
    {
        DebugPrint(L"No Memory Error\n");
        //FreeMemory(imgPartitions);
        return FWU_NO_MEMORY;
    }
    ZeroMem(fwUpdBuffer,totalBufSize);

    DebugPrint(L"PartitionCount: 0x%x\n",PartitionCount);
    //Extract the right partitions from the image to the buffer to be sent to the
    //firmware
    fwuError = ExtractPartitionsToBuffer(buffer,&fwUpdBuffer,totalBufSize,
        imgPartitions, PartitionCount,imgType);
    if(FWU_ERROR_SUCCESS != fwuError)
    {
        // *** TODO: why in comment ?
        //FreeMemory(fwUpdBuffer);
        //FreeMemory(imgPartitions);
        return fwuError;
    }

/*
#ifdef _DEBUG
    DebugPrint(L"Buffer Address F: 0x%x\n",fwUpdBuffer);
    DebugPrint(L"totalBufSize: 0x%x\n",totalBufSize);
    DisplayBufferContent(&fwUpdBuffer,totalBufSize);
#endif
*/

    //Call the actual update function to start-send-end fw update
    //using the buffer
    fwuError = Update(&fwUpdBuffer, totalBufSize, FWUPD_SKIP_UPDT_SIZE,
        _pwd, FWU_IPU_UPDATE_OPERATION,1,PartitionID,UpdateEnvironment,OemID,
        update_flags,&func);

    //Store the UpvVersion updated for return to caller
    IpuUpdatedInfo->UpdatedUpvVer = ExpectedUpvVersion;

    FreePool(fwUpdBuffer);
    FreePool(imgPartitions);

    gBS->Stall(100000); // WA for 1209142436 Doesn't return correct percentage via FWUpdate_QueryStatus_Get_Response.
                        // A delay for FW to update the percentage on FWSTS2.

    return fwuError;
}

/*++
********************************************************************************
**
** FUNCTION:
**   FwUpdateFullBuffer
**
** DESCRIPTION:
**   This function sends the data to the FW Update client
**
** ARGUMENTS:
**
** RETURNS:
**
** NOTE:
**
********************************************************************************
--*/
UINT32
    FwUpdateFullBuffer (
    IN  CHAR8                *buffer,
    IN  UINT32        bufferLength,
    IN  CHAR8                *_pwd,
    IN  int                 _forceResetLib,
    IN  UINT32        UpdateEnvironment,
    IN  _UUID               OemID,
    IN  UPDATE_FLAGS_LIB    update_flags,
    IN  void(*func)(float, float)
    )
{
    FWU_GET_IPU_PT_ATTRB_MSG_REPLY  FwuGetIpuAttrbMsgInfo = {0};
    ImagePartition                  *imgPartition = NULL;
    ImagePartition  preUpdatePartition = {0};
    IMAGE_TYPE                      imgType = {0};
    UINT32                   StartOffsetAfterLastActivePartition = 0;
    UINT32                    PartitionName = 0;
    UINT32                    UpvVersion = 0;
    UINT32                    InstanceId = 0;
    UINT32                    RegionLength = 0;
    UINT32                   PartitionCount=0;
    UINT32                    ExpectedNumOfPartitions = 0;
    UINT32                    totalBufSize = 0;
    CHAR8                            *fwUpdBuffer = NULL;
    INT32 NumOfIUPsNotToUpdate = 0;
    UINT32                    instanceOrder = NUM_OF_MANDATORY_PARTITIONS; // 1, 2, 3 are for FTPR, NFTP, RBEP respectively
    UINT32                    foundPart = 0;
    //UINT32                    imgIndexOffset = 0;
    UINT32                    fwuError = 0;
    UINT32                    i = 0, j = 0;
    EFI_STATUS                      Status = EFI_SUCCESS;
    UINT32                    currentPartIdx = 0;
    BOOLEAN                            ignorePartition = FALSE;
    UINT32                    partIdx = 0;

    // Used to keep track of expected instance Ids
    UINT32                    expectedNumofInstIdMatches = 0;
    UINT32                    InstIdMatchesCount = 0;
    UINT32 preUpdSize = 0; // Size of pre-update partition.
    BOOLEAN IsOEMP_inOldFW = FALSE;
    BOOLEAN IsIUNP_inOldFW = FALSE;
    BOOLEAN IsISHC_inOldFW = FALSE;
    BOOLEAN IsOEMP_inNewFW = FALSE;
    BOOLEAN IsIUNP_inNewFW = FALSE;
    BOOLEAN IsISHC_inNewFW = FALSE;

    DebugPrint(L"\nStart FwupdateFull Buffer mInitialized Value: %d\n", mInitialized);

    if (!mInitialized)
    {
        Status = GetFwVersion ();
        if (EFI_ERROR (Status))
        {
            DebugPrint(L"Unable to get current FW version - %r\n", Status);
            return FWU_IME_NOT_READY;
        }
    }

    //Determine the type of Image
    if (FWU_ERROR_SUCCESS != (fwuError = DetermineImageType(buffer,bufferLength, &imgType)))
    {
        return fwuError;
    }

    //The Image MUST be FULL image with FTP and NFTP to do FULL update
    if (imgType != FULL)
    {
        DebugPrint(L"\nImage Type is not Full, Type: 0x%x\n",imgType);
        return FWU_FILE_INVALID;
    }

    //Read FPT and check for its presence in the image
    if ((fwuError =
        ParsingImageForFPT((UCHAR*)buffer,&StartOffsetAfterLastActivePartition)))
    {
        return fwuError;
    }

    DebugPrint(L"StartOffsetAfterLastActivePartition: 0x%x\n",StartOffsetAfterLastActivePartition);

    //Get Partition Attribute from Firmware
    fwuError = GetExtendedIpuPartitionAttributes(&FwuGetIpuAttrbMsgInfo,FWU_FULL_UPDATE_OPERATION);
    if (FWU_ERROR_SUCCESS != fwuError)
    {
        //Heci gets deinitialized on both error or success in GetIpuPartitionAttributes
        return fwuError;
    }

    if (NULL == g_fptt)
    { //Image with FPT is required
        return FWU_FILE_INVALID;
    }

/*#if !BYPASS_FWU_GET_IPU_PT_ATTRB
    // FOR FULL FW Update number of expected IPUs plus 3 (FTP, NFTP, RBEP)
    ExpectedNumOfPartitions = FwuGetIpuAttrbMsgInfo.NumOfPartition + NUM_OF_MANDATORY_PARTITIONS;
    expectedNumofInstIdMatches = FwuGetIpuAttrbMsgInfo.NumOfPartition;
#else
    FwuGetIpuAttrbMsgInfo.NumOfPartition = 0;
    ExpectedNumOfPartitions = 2;
    expectedNumofInstIdMatches = 0;
#endif*/

    DebugPrint(L"FwuGetIpuAttrbMsgInfo->NumOfPartition: 0x%x\n",FwuGetIpuAttrbMsgInfo.NumOfPartition);
    // determine if the current FW image(on the platform) contains OEMP, ISHC and / or IUNP
    for(i=0;i<FwuGetIpuAttrbMsgInfo.NumOfPartition;i++)
    {
        if (FPT_PARTITION_NAME_OEMP == FwuGetIpuAttrbMsgInfo.PtAttribute[i].PtNameId)
        {
            IsOEMP_inOldFW = TRUE;
        }
        else if (FPT_PARTITION_NAME_IUNP == FwuGetIpuAttrbMsgInfo.PtAttribute[i].PtNameId)
        {
            IsIUNP_inOldFW = TRUE;
        }
        else if (FPT_PARTITION_NAME_ISHC == FwuGetIpuAttrbMsgInfo.PtAttribute[i].PtNameId)
        {
            IsISHC_inOldFW = TRUE;
        }

        DebugPrint(L"FwuGetIpuAttrbMsgInfo.PtAttribute[%d].PtNameId: 0x%x [%c%c%c%c]\n",
                    i,
                    FwuGetIpuAttrbMsgInfo.PtAttribute[i].PtNameId,
                    FwuGetIpuAttrbMsgInfo.PtAttribute[i].PtNameId & 0xFF,
                    (FwuGetIpuAttrbMsgInfo.PtAttribute[i].PtNameId>> 8) & 0xFF,
                    (FwuGetIpuAttrbMsgInfo.PtAttribute[i].PtNameId>>16) & 0xFF,
                    FwuGetIpuAttrbMsgInfo.PtAttribute[i].PtNameId>>24);
        DebugPrint(L"FwuGetIpuAttrbMsgInfo.PtAttribute[%d].ExpectedInstId: 0x%x \n",i,FwuGetIpuAttrbMsgInfo.PtAttribute[i].ExpectedInstId);
        DebugPrint(L"FwuGetIpuAttrbMsgInfo.PtAttribute[%d].ExpectedUpvVer: 0x%x\n",i,  FwuGetIpuAttrbMsgInfo.PtAttribute[i].ExpectedUpvVer);

    }
    DebugPrint(L"IsOEMP_inOldFW: %d, IsIUNP_inOldFW: %d, IsISHC_inOldFW: %d\n", IsOEMP_inOldFW, IsIUNP_inOldFW, IsISHC_inOldFW);

    //In the following section we define which IUPs must exist in the new FW:
    //Few lines above we discovered which IUPs are in the old FW (on the platform).
    //Now we loop on the new FW image (the FPT) and resolve which IUPs are provided.
    //next we will define if we are missing any IUP
    //Note: down the code we will do again this loop over the FPT, however since
    //the array is small and in order to keep the code readble we are doing it also here
    for (i = 0; i<g_fptt->header.num_of_entries; i++)
    {
        if (FPT_PARTITION_NAME_OEMP == g_fptt->entry[i].partition_name && g_fptt->entry[i].length != 0)
        {
            IsOEMP_inNewFW = TRUE;
        }
        else if (FPT_PARTITION_NAME_IUNP == g_fptt->entry[i].partition_name && g_fptt->entry[i].length != 0)
        {
            IsIUNP_inNewFW = TRUE;
        }
        else if (FPT_PARTITION_NAME_ISHC == g_fptt->entry[i].partition_name && g_fptt->entry[i].length != 0)
        {
            IsISHC_inNewFW = TRUE;
        }
    }
    DebugPrint(L"IsOEMP_inNewFW: %d, IsIUNP_inNewFW: %d, IsISHC_inNewFW: %d\n", IsOEMP_inNewFW, IsIUNP_inNewFW, IsISHC_inNewFW);

    //checking if any IUP is missing from the new given FW image
    if (IsOEMP_inOldFW)//the old FW image holds EOMP
    {
        if (IsOEMP_inNewFW)// the new FW image comes with OEMP,
        {   //Old and new images have OEMP => hence we must provide ISHC and INUP (if they exist in the old FW)
            if ((IsIUNP_inOldFW) && (!IsIUNP_inNewFW))// we already have IUNP on the platform but we didn't get a new version in the new FW image - error
            {
                DebugPrint(L"Error: Provided image is missing IUNP partition\n");
                return FWU_FILE_INVALID_MISSING_IUNP;
            }
            if ((IsISHC_inOldFW) && (!IsISHC_inNewFW))// we already have ISHC on the platform but we didn't get a new version in the new FW image - error
            {
                DebugPrint(L"Error: Provided image is missing ISHC partition\n");
                return FWU_FILE_INVALID_MISSING_ISHC;
            }
            if ((!IsIUNP_inOldFW) && (IsIUNP_inNewFW))// we don't have IUNP on the platform but we have a new version in the new FW image - error
            {
                DebugPrint(L"Error: new provided image contains IUNP partition but the existing image on the platform doesn't contain it\n");
                return FWU_FILE_INVALID_MISSING_IUNP_IN_FLASH;
            }
            if ((!IsISHC_inOldFW) && (IsISHC_inNewFW))// we don't have ISHC on the platform but we have a new version in the new FW image - error
            {
                DebugPrint(L"Error: new provided image contains ISHC partition but the existing image on the platform doesn't contain it\n");
                return FWU_FILE_INVALID_MISSING_ISHC_IN_FLASH;
            }
        }
        else //new image doesn't contain OEMP => in this case we must not update ISHC and IUNP
        {
            if (IsISHC_inNewFW || IsIUNP_inNewFW)
            {
                DebugPrint(L"Error: new image contains ISHC or/and IUNP partitions but not the OEM Key manifest patition\n");
                return FWU_FILE_INVALID_MISSING_OEMP;
            }

            // building the exact number of partitions to update:
            // If we are here it means that we will have OEMP, now we check if we have ISHC or IUNP or both of them
            // about the logic:
            // when we retrive the number of partions from the FW we get the number of partition we already have inside.
            // however, there is a chance we may update *less* partitions hence here we want to reduce the number of IUPs

            NumOfIUPsNotToUpdate += 1;// we know that OEMP is in the platform but we will not update it
            if (IsISHC_inOldFW)// here we check that ISHC partition is in the old FW imgae but we will NOT update it
            {
                NumOfIUPsNotToUpdate += 1;
            }
            if (IsIUNP_inOldFW)// here we check that IUNP partition is in the old FW imgae but we will NOT update it
            {
                NumOfIUPsNotToUpdate += 1;
            }

        }
    }
    else //No OEMP in old image
    {
        if (IsOEMP_inNewFW) // an OEMP was provided with new image but old image don't contain such => this is an error
        {
            DebugPrint(L"Error: new image contains OEM Key manifest partition but the current platform image don't contain such partition\n");
            return FWU_FILE_INVALID_MISSING_OEMP_IN_FLASH;
        }

        if (IsISHC_inNewFW)
        {
            if (!IsISHC_inOldFW)// an ISHC was provided with new image but old image don't contain such => this is an error
            {
                DebugPrint(L"Error: new image contains ISHC partition but the current platform image - no\n");
                return FWU_FILE_INVALID_MISSING_ISHC_IN_FLASH;
            }
        }
        else //No ISHC in the new provided image
        {
            if (IsISHC_inOldFW)// we have ISHC in the old FW, but we don't have a new verion in the new FW image=> we wont update this partition
            {
                NumOfIUPsNotToUpdate += 1;
            }
        }

        if (IsIUNP_inNewFW)
        {
            if (!IsIUNP_inOldFW)// an IUNP was provided with new image but old image don't contain such => this is an error
            {
                DebugPrint(L"Error: new image contains IUNP partition but the current platform image - no\n");
                return FWU_FILE_INVALID_MISSING_IUNP_IN_FLASH;
            }
        }
        else //No IUNP in the new provided image
        {
            if (IsIUNP_inOldFW)// we have IUNP in the old FW, but we don't have a new verion in the new FW image=> we wont update this partition
            {
                NumOfIUPsNotToUpdate += 1;
            }
        }
    }

#if !BYPASS_FWU_GET_IPU_PT_ATTRB
    // FOR FULL FW Update number of expected IPUs plus 3 (FTP, NFTP, RBEP), additionally we remove the number of IUPs that are on the
    // platform but wont be updated (NumOfIUPsNotToUpdate is zero or negative)
    ExpectedNumOfPartitions = FwuGetIpuAttrbMsgInfo.NumOfPartition + NUM_OF_MANDATORY_PARTITIONS - NumOfIUPsNotToUpdate;
    expectedNumofInstIdMatches = FwuGetIpuAttrbMsgInfo.NumOfPartition - NumOfIUPsNotToUpdate;

    DebugPrint(L"ExpectedNumOfPartitions: %d\n", ExpectedNumOfPartitions);
    DebugPrint(L"expectedNumofInstIdMatches: %d\n", expectedNumofInstIdMatches);
    DebugPrint(L"NumOfIUPsNotToUpdate: %d\n", NumOfIUPsNotToUpdate);
#else
    FwuGetIpuAttrbMsgInfo.NumOfPartition = 0;
    ExpectedNumOfPartitions = 2;
    expectedNumofInstIdMatches = 0;
#endif

    //Allocate memory to capture the required partition information for send to FW
    imgPartition = (ImagePartition*)AllocatePool(ExpectedNumOfPartitions*sizeof(ImagePartition));
    if (NULL == imgPartition)
    {
        DebugPrint(L"No Memory Error\n");
        return FWU_NO_MEMORY;
    }

    ZeroMem(imgPartition, ExpectedNumOfPartitions * sizeof(ImagePartition));

    PartitionCount=0;
    DebugPrint(L"PartitionCount: %d\n", PartitionCount);

    //Loop through the list of entries read from FPT to find
    //FTP, NFTP, and/or WCOD | LOCL |FPSP | ISHC
    for(i=0;i<g_fptt->header.num_of_entries;i++)
    {
        if (FPT_PARTITION_NAME_FTPR == g_fptt->entry[i].partition_name||
            FPT_PARTITION_NAME_NFTP == g_fptt->entry[i].partition_name ||
            FPT_PARTITION_NAME_RBEP == g_fptt->entry[i].partition_name)
        {
            imgPartition[PartitionCount].PartitionName =
                g_fptt->entry[i].partition_name;
            imgPartition[PartitionCount].OffsetFromFPT =
                g_fptt->entry[i].offset;
            imgPartition[PartitionCount].PartitionLength =
                g_fptt->entry[i].length;
            imgPartition[PartitionCount].UpvVersion = 1;//Defaults for FTP and NFTP
            imgPartition[PartitionCount].InstanceId = 1;//Defaults for FTP and NFTP

            DebugPrint(L"imgPartition[PartitionCount].PartitionName = 0x%x [%c%c%c%c]\n",
                        g_fptt->entry[i].partition_name,
                        g_fptt->entry[i].partition_name & 0xFF,
                        (g_fptt->entry[i].partition_name>> 8) & 0xFF,
                        (g_fptt->entry[i].partition_name>>16) & 0xFF,
                        g_fptt->entry[i].partition_name>>24);
            DebugPrint(L"imgPartition[PartitionCount].OffsetFromFPT = 0x%x \n", g_fptt->entry[i].offset);
            DebugPrint(L"imgPartition[PartitionCount].PartitionLength = 0x%x \n", g_fptt->entry[i].length);

            if (FPT_PARTITION_NAME_FTPR == g_fptt->entry[i].partition_name) {
                // for full update FTP will be first with order number 1
                imgPartition[PartitionCount].ImageOrder = 1;
            } else if (FPT_PARTITION_NAME_NFTP == g_fptt->entry[i].partition_name) {
                //for full update NFTP will follow FTP with order number 2
                imgPartition[PartitionCount].ImageOrder = 2;
            } else if (FPT_PARTITION_NAME_RBEP == g_fptt->entry[i].partition_name) {
                //for full update RBEP will follow NFTP with order number 3
                imgPartition[PartitionCount].ImageOrder = 3;
            }
            PartitionCount++;

        } else if (g_fptt->entry[i].length != 0 &&
            (FPT_PARTITION_NAME_WCOD == g_fptt->entry[i].partition_name ||
             FPT_PARTITION_NAME_LOCL == g_fptt->entry[i].partition_name ||
             FPT_PARTITION_NAME_MDMV == g_fptt->entry[i].partition_name ||
             FPT_PARTITION_NAME_PMCP == g_fptt->entry[i].partition_name ||
             FPT_PARTITION_NAME_OEMP == g_fptt->entry[i].partition_name ||
             FPT_PARTITION_NAME_IUNP == g_fptt->entry[i].partition_name ||
             FPT_PARTITION_NAME_ISHC == g_fptt->entry[i].partition_name))
        {
            DebugPrint(L"Partition Name: 0x%x [%c%c%c%c]\n",
                        g_fptt->entry[i].partition_name,
                        g_fptt->entry[i].partition_name & 0xFF,
                        (g_fptt->entry[i].partition_name>> 8) & 0xFF,
                        (g_fptt->entry[i].partition_name>>16) & 0xFF,
                        g_fptt->entry[i].partition_name>>24);

            //Read the Manifest to get UpvVersion and Instance ID of the region
            fwuError = GetUpvAndInstanceID((UCHAR*)buffer,
                                           g_fptt->entry[i].offset,
                                           &PartitionName,
                                           &UpvVersion,
                                           &InstanceId,
                                           &RegionLength);
            if (FWU_ERROR_SUCCESS != fwuError)
            {
                DebugPrint(L"GetUpvAndInstanceID() failed: %d\n", fwuError);
                FreePool (imgPartition);

                if (INVALID_MANIFEST_DATA == fwuError)
                {
                    fwuError = FWU_GENERAL;
                }
                return fwuError;
            }

            //Find the right Partition matching Partition ID, Instance ID, and
            // Major part of the UPV version
            for(j=0;j<FwuGetIpuAttrbMsgInfo.NumOfPartition;j++)
            {
                DebugPrint(L"FwUpdateFull ExpectedPartitionName: 0x%x [%c%c%c%c]\n",
                            FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId,
                            FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId & 0xFF,
                            (FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId>> 8) & 0xFF,
                            (FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId>>16) & 0xFF,
                            FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId>>24);
                DebugPrint(L"FwUpdateFull ExpectedInstanceId: 0x%x \n",FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedInstId);
                DebugPrint(L"FwUpdateFull ExpectedUpvVersion: 0x%x \n",FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedUpvVer);

                DebugPrint(L"FwUpdateFull PartitionName: 0x%x [%c%c%c%c]\n",PartitionName,
                            PartitionName & 0xFF,
                            (PartitionName>> 8) & 0xFF,
                            (PartitionName>>16) & 0xFF,
                            PartitionName>>24);
                DebugPrint(L"FwUpdateFull InstanceId: 0x%x \n",InstanceId);
                DebugPrint(L"FwUpdateFull UpvVErsion: 0x%x \n",UpvVersion);

                if (PartitionName == FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId)
                {
                    DebugPrint(L"\n FOUND PARTITION \n");
                    DebugPrint(L"PartitionName: 0x%x [%c%c%c%c]\n",FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId,
                                FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId & 0xFF,
                                (FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId>> 8) & 0xFF,
                                (FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId>>16) & 0xFF,
                                FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId>>24);

                    DebugPrint(L"InstanceId: 0x%x \n", FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedInstId);
                    DebugPrint(L"UpvVersion: 0x%x \n",UpvVersion);

                    currentPartIdx = PartitionCount;
                    ignorePartition = FALSE;
                    // check current list to see if there is already an entry for
                    // this partition name.
                    for(partIdx = 0; partIdx < PartitionCount; partIdx++)
                    {
                        if(imgPartition[partIdx].PartitionName == PartitionName)
                        {
                            currentPartIdx = partIdx;
                            DebugPrint(L"Found duplicate partition in the runnable region. Using best candidate\n");

                            // Will ignore this partition if we already have an entry
                            // for it with matching instanceId.
                            if(imgPartition[partIdx].InstanceId == FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedInstId)
                            {
                                DebugPrint(L"Ignoring this partition. Already have matching instanceId \n");

                                ignorePartition = TRUE;
                                break;
                            }

                            // Make sure UPV of this partition matches the first found. This is done
                            // to ensure that the Full image will not have any issues processing
                            // the image.
                            if(imgPartition[partIdx].UpvVersion != UpvVersion)
                            {
                                DebugPrint(L"Ignoring this partition. Previous matching partition has different UPV.\n  Will keep previous one.\n");

                                ignorePartition = TRUE;
                                break;
                            }

                            // InstanceID bookeeping.
                            if(FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedInstId == InstanceId)
                            {
                                InstIdMatchesCount++;
                            }
                            break;
                        }
                    }

                    if(ignorePartition)
                        continue;

                    // If we couldn't find an entry with the same partition name,
                    // then increment the number of partitions. We also check the instanceId
                    if(currentPartIdx == PartitionCount)
                    {
                        if(InstanceId == FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedInstId)
                        {
                            InstIdMatchesCount++;
                        }
                        PartitionCount++;
                        imgPartition[currentPartIdx].ImageOrder = instanceOrder + ++j;
                        // Note: the ImageOrder field will be zero for FTP and NFTP.
                        // Only populate the order if this is a newly found partition.
                    }

                    imgPartition[currentPartIdx].PartitionName =
                        g_fptt->entry[i].partition_name;
                    imgPartition[currentPartIdx].OffsetFromFPT =
                        g_fptt->entry[i].offset;
                    imgPartition[currentPartIdx].PartitionLength =
                        g_fptt->entry[i].length;
                    imgPartition[currentPartIdx].UpvVersion = UpvVersion;
                    imgPartition[currentPartIdx].InstanceId = InstanceId;


                }
            }//for(int j=0;j<FwuGetIpuAttrbMsgInfo.NumOfPartition;j++)
        }//else if(FPT_PARTITION_NAME_WCOD == fptt->entry[i].partition_name ||
        //FPT_PARTITION_NAME_LOCL == fptt->entry[i].partition_name)  ||
        //FPT_PARTITION_NAME_MDMV == g_fptt->entry[i].partition_name ||
        //FPT_PARTITION_NAME_IUNP == g_fptt->entry[i].partition_name ||
        //(FPT_PARTITION_NAME_ISHC...
    }// for(int i=0;i<fptHdr.num_of_entries;i++)

//    DebugPrint(L"ElementCount: %d/%d\n",ElementCount, ExpectedNumOfPartitions);
    DebugPrint(L"ExpectedNumOfPartitions: %d\n",ExpectedNumOfPartitions);
    DebugPrint(L"Matching instance IDs: %d/%d\n",InstIdMatchesCount, expectedNumofInstIdMatches);
    DebugPrint(L"PartitionCount %d \n", PartitionCount);

    if(PartitionCount != ExpectedNumOfPartitions)
    {
        DebugPrint(L"Error: PartitionCount does not match Expected number of partitions! \n");
        FreePool (imgPartition);
        return FWU_FILE_INVALID; //Seven Commented for test on KBLR
    }

    //Did we find all the expected partitions, if not continue to
    //search for partitions in the non-runnable part of the image
    //if (ElementCount < ExpectedNumOfPartitions)
    if(InstIdMatchesCount < expectedNumofInstIdMatches)
    {
        //meaning all expected partitions are not found yet
        //Search non-runnable regions
        for(j=0;j<FwuGetIpuAttrbMsgInfo.NumOfPartition;j++)
        {
            //imgIndexOffset = j;
            fwuError = LoadUpvAndInstanceID(buffer,bufferLength,
                StartOffsetAfterLastActivePartition,
                FwuGetIpuAttrbMsgInfo.PtAttribute[j].PtNameId,
                FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedInstId,
                FwuGetIpuAttrbMsgInfo.PtAttribute[j].ExpectedUpvVer,
                &imgPartition,&PartitionCount, &foundPart,1);
            if (FWU_ERROR_SUCCESS != fwuError)
            {
                FreePool(imgPartition);
                return fwuError;
            }

            // At this point the matching criteria is enforcing instanceId matches
            // so we can increment the counter whenever we find a partition.
            if((InstIdMatchesCount != expectedNumofInstIdMatches) && (foundPart))
            {
                expectedNumofInstIdMatches++;
            }
            // No need to continue the search once we have found all the
            // partitions.
            if(InstIdMatchesCount == expectedNumofInstIdMatches)
            {
                break;
            }
        }
    }

#ifdef _DEBUG
    Print(L"Image Order Results:");

    for(i = 0; i < PartitionCount; i++)
    {
        Print(L"Partition: 0x%08x - order: %d\n", imgPartition[i].PartitionName, imgPartition[i].ImageOrder);
    }
#endif

    //Even after searching both Runnable and Non-Runnable part of the image
    // if you don't find the expected number of partitons Error out.
    // Here we don't check for number of instanceId matches because we allow
    // the active region partitions to take precedence over the FW specific
    // instance IDs requested by FW.
    if(PartitionCount != ExpectedNumOfPartitions)
    {
        //this should not happen
        //DebugPrint(L"ElementCount: %d\nExpectedNumOfPartitions: %d\n",ElementCount,ExpectedNumOfPartitions);
        DebugPrint(L"Incorrect or bad upd file entered.\n");
        //FreePool (imgPartition); //Seven Commented for test on KBLR
        //return FWU_FILE_INVALID;//Seven Commented for test on KBLR
    }
    DebugPrint(L"TotalBufSize: %d\n",totalBufSize);
    // Search for the pre-update FW module PARTITION_NAME_UPDT
    DebugPrint(L"Searging for Pre-Update Partition PARTITION_NAME_UPDT\n");

    ZeroMem((UCHAR*)&preUpdatePartition, sizeof(preUpdatePartition));

    fwuError = LoadPreUpdatePartition((UCHAR*)buffer, bufferLength, StartOffsetAfterLastActivePartition,
        &preUpdatePartition, &foundPart);

    if (FWU_ERROR_SUCCESS != fwuError)
    {
        DebugPrint(L"Error looking for UPDT partition\n");
        FreePool(imgPartition);
        return fwuError;
    }

    if (foundPart)
    {
        preUpdSize = preUpdatePartition.PartitionLength;
    }
    // TODO remove this after done //DEBUGging
    //preUpdSize = 0;

    //Compute Required Buffer Size to Store Partitions
    for(i=0; i<PartitionCount; i++)
    {
        totalBufSize = totalBufSize + imgPartition[i].PartitionLength;
    }
    DebugPrint(L"TotalBufSize: %d\n", totalBufSize);
    DebugPrint(L"preUpdSize: %d\n", preUpdSize);

    //allocate memory space to hold all partitions to be sent to FW
    fwUpdBuffer = (CHAR8*)AllocatePool(totalBufSize + preUpdSize);
    if (NULL == fwUpdBuffer)
    {
        DebugPrint(L"No Memory Error\n");
        FreePool (imgPartition);
        return FWU_NO_MEMORY;
    }
    ZeroMem(fwUpdBuffer, totalBufSize + preUpdSize);

    //Extract the right partitions from the image to the buffer to be sent to the
    //firmware
    fwuError = ExtractPartitionsToBuffer(buffer,&fwUpdBuffer,totalBufSize,
        imgPartition, PartitionCount,imgType);
    if (FWU_ERROR_SUCCESS != fwuError)
    {
        FreePool(fwUpdBuffer);
        FreePool(imgPartition);
        return fwuError;
    }

/*
#ifdef _DEBUG
    DebugPrint(L"Buffer Address: 0x%x  \n",fwUpdBuffer);
    DebugPrint(L"totalBufSize: 0x%x \n",totalBufSize);

    _ShowBuffer((UINT8*)&fwUpdBuffer,totalBufSize);
    DisplayBufferContent(&fwUpdBuffer,totalBufSize);
#endif
*/

    //Add UPDT data to the end of the Buffer
    if (preUpdSize > 0)
    {
        CopyMem(fwUpdBuffer + totalBufSize,
            buffer + preUpdatePartition.OffsetFromFPT,
            preUpdSize);
    }
    //skipEndResponseMsgOnSuccess = 1;
    fwuError = Update(&fwUpdBuffer, totalBufSize, preUpdSize, _pwd,_forceResetLib,0,0,UpdateEnvironment,OemID,update_flags,&func);

    FreePool(fwUpdBuffer);
    FreePool(imgPartition);

    gBS->Stall(100000); // WA for 1209142436 Doesn't return correct percentage via FWUpdate_QueryStatus_Get_Response.
                        // A delay for FW to update the percentage on FWSTS2.

    return fwuError;
}

/*++

Routine Description:

Connect to the enumerated client that supports FW Update
A valid client ID is required for this routine

Arguments:

Returns:

EFI_STATUS

--*/
static EFI_STATUS
    ClientConnect (
    IN  ME_CLIENT_PROPERTIES   *MeClient
    )
{
    EFI_STATUS                          Status = EFI_SUCCESS;
    HBM_CLIENT_CONNECT_REQUEST          ClientConnectRequest = {0};
    HBM_CLIENT_CONNECT_RESPONSE         *ClientConnectResponse = {0};
    UINT32                              Length = 0;
    CHAR8                               Buffer[1024] = {0};

    ClientConnectRequest.MEAddress = MeClient->ClientId;
    ClientConnectRequest.HostAddress = CLIENT_HOST_ADDRESS;
    ClientConnectRequest.Command = CLIENT_CONNECT_REQ_CMD;
    ClientConnectRequest.Reserved = 0;

    Length = sizeof(HBM_CLIENT_CONNECT_REQUEST);
    //
    // Send Client Connect Request to ME
    //
    Status = mHeci->SendMsg (0,
        (UINT32 *) &ClientConnectRequest,
        Length,
        HBM_PROTOCOL_ADDRESS,
        HBM_PROTOCOL_ADDRESS
        );
    if (EFI_ERROR (Status)) {
        DebugPrint(L"Unable to send Client Connect Request message to %d - %r\n", MeClient->ClientId, Status);
        return Status;
    }

    //
    // Read response
    //
    //Status = ReadMessageWithFlow (MeClient, Buffer, sizeof (Buffer);
    Length = sizeof(HBM_CLIENT_CONNECT_RESPONSE);
    Status = mHeci->ReadMsg (0,BLOCKING,(UINT32 *) Buffer,&Length);

    DebugPrint(L"Read Msg from Connect : 0x%02x%02x%02x%02x\n", *(Buffer+3), *(Buffer+2), *(Buffer+1), *(Buffer) );

    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable to read Client Connection Response - %r\n", Status);
        return Status;
    }

    ClientConnectResponse = (HBM_CLIENT_CONNECT_RESPONSE *)Buffer;
    if (ClientConnectResponse->Status || ClientConnectResponse->Command != CLIENT_CONNECT_RES_CMD)
    {
        DebugPrint(L"Unable to connect to client - %r\n", Status);
        return Status;
    }
    Status = GetFlowControl(MeClient);
    if (EFI_ERROR (Status))
    {
        return Status;
    }

    MeClient->Connected = TRUE;
    MeClient->HostAddress = ClientConnectResponse->HostAddress;

    return Status;
}

/*++

Routine Description:

Disconnect from the connected client

Arguments:

Returns:
EFI_STATUS

--*/
static EFI_STATUS
    ClientDisconnect (
    IN  ME_CLIENT_PROPERTIES   *MeClient
    )
{
    EFI_STATUS                          Status = EFI_SUCCESS;
    HBM_CLIENT_DISCONNECT_REQUEST       ClientDisconnectRequest = {0};
    HBM_CLIENT_DISCONNECT_RESPONSE      ClientDisconnectResponse = {0};
    UINT32                              Length = 0;
    CHAR8                              Buff[1024] = {0};

    if (MeClient->Connected == FALSE) {
        return EFI_SUCCESS;
    }

    ClientDisconnectRequest.MEAddress = MeClient->ClientId;
    ClientDisconnectRequest.HostAddress = CLIENT_HOST_ADDRESS;
    ClientDisconnectRequest.Command = CLIENT_DISCONNECT_REQ_CMD;
    ClientDisconnectRequest.Reserved[0] = 0;

    Length = sizeof(HBM_CLIENT_DISCONNECT_REQUEST);

    //
    // Send Client Disconnect Request to ME
    //
    Status = mHeci->SendMsg (0,
        (UINT32 *) &ClientDisconnectRequest,
        Length,
        HBM_PROTOCOL_ADDRESS,
        HBM_PROTOCOL_ADDRESS
        );
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable to send Client Disconnect Request message to %d - %r\n", MeClient->ClientId, Status);
        return Status;
    }

    //
    // Read response - not sure why response is 8 bytes instead of 4, so we use
    // use cheat buffer
    //
    Length = sizeof(HBM_CLIENT_DISCONNECT_RESPONSE);

    Status = mHeci->ReadMsg (0,
        BLOCKING,
        (UINT32 *) Buff,
        &Length
        );

    DebugPrint(L"Read Msg from Disconnect : 0x%02x%02x%02x%02x\n", *(Buff+3), *(Buff+2), *(Buff+1), *(Buff) );

    if (EFI_ERROR (Status))
    {
        DebugPrint(L"Unable to read Client Disconnect Response - %r\n", Status);
        return Status;
    }

    CopyMem (&ClientDisconnectResponse, Buff, sizeof (HBM_CLIENT_DISCONNECT_RESPONSE));

    if (ClientDisconnectResponse.Status || ClientDisconnectResponse.Command != CLIENT_DISCONNECT_RES_CMD)
    {
        DebugPrint(L"Unable to disconnect from Client - %x\n", ClientDisconnectResponse.Status);
    } else
    {
        MeClient->Connected = FALSE;
        DebugPrint(L"Disconnected from client\n");
    }
    return Status;
}

/*++

Routine Description:

Initialize the HECI interface.
Get a handle to the protocol, verify the ME Mode and current intialization state
Reset the HECI pointers in case a message is still left in the queue

Arguments:

Returns:

EFI_STATUS

--*/
static EFI_STATUS
    InitializeHeci ()
{
    EFI_STATUS                          Status = EFI_SUCCESS;
    EFI_STATUS                          TempStatus = EFI_SUCCESS;
    UINT32                              MeMode = 0;
    UINT32                              MeStatus = 0;
    UINT32                              NumTries = 10;
#ifdef MKHI_SUPPORT
    if (!g_fwulib_heci_initialized){
#endif
    Status = gBS->LocateProtocol (
        &gEfiHeciProtocolGuid,
        NULL,
        &mHeci
        );

    if (EFI_ERROR (Status)) {
        DebugPrint(L"InitializeHeci: Could not locate HECI protocol!\n");
        return EFI_UNSUPPORTED;
    }
#ifdef MKHI_SUPPORT
        g_fwulib_heci_initialized = TRUE;
    }
#endif

    Status = mHeci->GetMeMode(&MeMode);
    if (EFI_ERROR (Status)) {
        DebugPrint(L"InitializeHeci: Could not obtain ME Mode!\n");
        return Status;
    }

    if (MeMode != ME_MODE_NORMAL) {
        DebugPrint(L"Local FW Update only supported when ME Mode=Normal\n");
        return EFI_UNSUPPORTED;
    }

    //
    // We need to make sure the Local FW Update client is loaded, thus wait for FW_INIT_DONE
    //
    TempStatus = mHeci->GetMeStatus (&MeStatus);
    while (NumTries-- && !ME_STATUS_IS_ME_FW_INIT_COMPLETE (MeStatus)) {
        gBS->Stall (100);
        TempStatus = mHeci->GetMeStatus (&MeStatus);
    }

    Status = mHeci->ResetHeci(0);

    return Status;

}

/*++

Routine Description:

Ask the ME for a list of dynamic clients.
For each dynamic client returned in the array, ask the client for
the properties and if the FW Update GUID is found, we have our
client.  Each client will have its own MeAddress and HostAddress that is
required for HECI messages to that client.

Arguments:

Returns:

EFI_STATUS

--*/
static EFI_STATUS
    FindClients ()
{
    EFI_STATUS                          Status = EFI_SUCCESS;
    HBM_HOST_ENUMERATION_REQUEST        EnumRequestMessage = {0};
    HBM_HOST_ENUMERATION_RESPONSE       EnumResponse = {0};
    UINT8                               i = 0, j = 0;
    UINT32                              Length = 0;
    HBM_HOST_CLIENT_PROPERTIES_REQUEST  ClientPropertiesMessage = {0};
    HBM_HOST_CLIENT_PROPERTIES_RESPONSE ClientPropertiesResponse = {0};

    //EnumRequestMessage.Command.IsResponse = 0;
    EnumRequestMessage.Command = HOST_ENUM_REQ_CMD;
    EnumRequestMessage.Reserved[0] = 0;
    EnumRequestMessage.Reserved[1] = 0;
    //EnumRequestMessage.Reserved[2] = 0;

    Length = sizeof (HBM_HOST_ENUMERATION_REQUEST);

    //
    // Send Host Enumeration to ME
    //
    Status = mHeci->SendMsg (0,
        (UINT32 *) &EnumRequestMessage,
        Length,
        HBM_PROTOCOL_ADDRESS,
        HBM_PROTOCOL_ADDRESS
        );
    if (EFI_ERROR (Status)) {
        DebugPrint(L"Unable to send Host Enumeration Request message - %r\n", Status);
        return Status;
    }

    //
    // Read response
    //
    Length = sizeof (HBM_HOST_ENUMERATION_RESPONSE);
    Status = mHeci->ReadMsg (0,
        BLOCKING,
        (UINT32 *) &EnumResponse,
        &Length
        );
    if (EFI_ERROR (Status)) {
        DebugPrint(L"Unable to read Host Enumeration Response - %r\n", Status);
        return Status;
    }

    // Find the FWU Client
    //
    for (i = 0; i < sizeof(EnumResponse.ValidAddresses); i ++) {
        for (j = 0; j < 8; j ++) {
            if ((EnumResponse.ValidAddresses[i] & (1 << j)) != 0) {
                //ClientPropertiesMessage.Command.IsResponse = 0;
                ClientPropertiesMessage.Command = HOST_CLIENT_PROPERTIES_REQ_CMD;
                ClientPropertiesMessage.Address = (i*8)+j;
                ClientPropertiesMessage.Reserved[0] = 0;
                ClientPropertiesMessage.Reserved[1] = 0;

                Length = sizeof(HBM_HOST_CLIENT_PROPERTIES_REQUEST);

                //
                // Send Client Properties Request to ME
                //
                Status = mHeci->SendMsg (0,
                    (UINT32 *) &ClientPropertiesMessage,
                    Length,
                    HBM_PROTOCOL_ADDRESS,
                    HBM_PROTOCOL_ADDRESS
                    );
                if (EFI_ERROR (Status)) {
                    DebugPrint(L"Unable to send Client Properties message to %d - %r\n", i, Status);
                    break;
                }

                //
                // Read response
                //
                Length = sizeof (HBM_HOST_CLIENT_PROPERTIES_RESPONSE);
                Status = mHeci->ReadMsg (0,
                    BLOCKING,
                    (UINT32 *) &ClientPropertiesResponse,
                    &Length
                    );
                if (EFI_ERROR (Status)) {
                    DebugPrint(L"Unable to read Host Enumeration Response - %r\n", Status);
                    break;
                }

                if (!ClientPropertiesResponse.Status) {
                    if (CompareMem(&ClientPropertiesResponse.ClientProperties.ProtocolName, &FWUGuid, 16) == 0) { // Compre GUID
                        DebugPrint(L"Found FWU Client\n");

                        MeClients[FWU_CLIENT].MaxMessageLength = ClientPropertiesResponse.ClientProperties.MaxMessageLength;
                        //
                        // Need to reduce the size of the data sent to include:
                        //  1) HECI Message Header
                        //  2) Flow Control Message and HECI Message Header
                        //
                        // Testing revealed that occassionally the Flow Control Message will be put into the same HECI buffer
                        // from the host and thus iff the data is not restricted in size by the additional bytes, update data
                        // will get corrupted.
                        //
                        MeClients[FWU_CLIENT].MaxMessageLength -= sizeof (HECI_MESSAGE_HEADER);
                        DebugPrint(L"FWU Client Max Message Length: %d\n", MeClients[FWU_CLIENT].MaxMessageLength);
                        if (MeClients[FWU_CLIENT].MaxMessageLength < sizeof(FWU_HECI_MSG)) {
                            return EFI_BUFFER_TOO_SMALL;
                        }
                        // ClientId = (i*8)+j;
                        MeClients[FWU_CLIENT].ClientId = ClientPropertiesResponse.Address;
                    }
                }
            }
        }
    }

    if (!MeClients[FWU_CLIENT].ClientId) {
        DebugPrint(L"Did not find FWU Client - exiting\n");
        return EFI_NOT_FOUND;
    }

    return Status;
}

/*++

Routine Description:

Connect the clients using the list of Client IDs/Host Addresses found
previously.

Arguments:

Returns:

EFI_STATUS

--*/
static EFI_STATUS ConnectClients ()
{
    EFI_STATUS Status = EFI_SUCCESS;

    Status = ClientConnect (&MeClients[FWU_CLIENT]);
    if (EFI_ERROR (Status)) {
        DebugPrint(L"Unable to Connect FWU Client - %r\n", Status);
        return Status;
    }

    DebugPrint(L"Connected to FWU Client: %d, %d\n", MeClients[FWU_CLIENT].HostAddress, MeClients[FWU_CLIENT].ClientId);
    return Status;
}

/*++

Routine Description:

Initiatlize the Update library and connect to the FW Update Client

Arguments:

Returns:

--*/
static EFI_STATUS
    ConnectFWU ()
{
    EFI_STATUS Status = EFI_SUCCESS;

    ZeroMem (MeClients, sizeof(MeClients));

    Status = InitializeHeci();

    if (EFI_ERROR (Status)) {
        DebugPrint(L"Cannot initialize HECI or ME mode is wrong - %r\n", Status);
        return Status;
    }

    //
    // Now find and connect the clients
    //
    Status = FindClients();
    if (EFI_ERROR (Status)) {
        return Status;
    }
    Status = ConnectClients();
    if (EFI_ERROR (Status)) {
        return Status;
    }

    mChunkSize = MeClients[FWU_CLIENT].MaxMessageLength - sizeof(FWU_DATA_MESSAGE);

    mpTxBuff = (UINT8 *)AllocatePool (MeClients[FWU_CLIENT].MaxMessageLength);
    mpRxBuff = (UINT8 *)AllocatePool (MeClients[FWU_CLIENT].MaxMessageLength);

    return Status;
}

/*++

Routine Description:

Disconnect from the FW Update Client

Arguments:

Returns:

EFI_STATUS

--*/
static EFI_STATUS
    DisConnectFWU ()
{
    EFI_STATUS  Status = EFI_SUCCESS;
    UINT32 i = 0;
    DebugPrint(L"Entered Disconnect FWU \n");

    for (i = 0; i < NUM_ME_CLIENTS; i++) {
        Status = ClientDisconnect (&MeClients[i]);
    }

    FreePool (mpTxBuff);
    FreePool (mpRxBuff);

    return Status;
}

/*++

Routine Description:

Get the FW version of the currently running ME FW

NOTE: A connection to the client is required before this routine is
executed.

Arguments:

FWVersion pointer to FW Flash version

Returns:

EFI_STATUS

--*/
EFI_STATUS
    GetFwFlashVersion(
    IN  FWVersion  *fwVersion
    )
{
    EFI_STATUS Status = EFI_SUCCESS;

    if (fwVersion != NULL) {
        if (!mInitialized) {
            Status = GetFwVersion ();
            if (EFI_ERROR (Status)) {
                DebugPrint(L"Unable to get current FW version - %r\n", Status);
                return Status;
            }
        }
        CopyMem (fwVersion, &mVer.code, sizeof(mVer.code));
        Status = DisConnectFWU();
        if (EFI_ERROR (Status))
        {
            DebugPrint(L"Cannot connect to clients - %r\n", Status);
            return Status;
        }
        mInitialized = FALSE;

        return EFI_SUCCESS;
    } else {
        return EFI_INVALID_PARAMETER;
    }
}


/**
This message is sent when we wants to send
PDT Message containing data to ISH.

@param[in] FileName             Filename of bios2ish file.
@param[in] PointerToPdtData     Pointer to PDT Data.
@param[in] PayloadSize          Payload Size.

@exception EFI_UNSUPPORTED      Current ME mode doesn't support this function or EOP was sent
@retval EFI_SUCCESS             Command succeeded
@retval EFI_DEVICE_ERROR        HECI Device error, command aborts abnormally
**/
static EFI_STATUS
    HeciPdtUpdateMsg1(
    IN const CHAR8*                FileName,
    IN UINT32                      *PointerToPdtData,
    IN UINT16                       PayloadSize
    )
{
    EFI_STATUS Status = EFI_SUCCESS;
    UINT16 Length = 0;
    PDT_UPDATE PdtUpdate = {0};
    UINT32 MeMode= 0;
    EFI_HECI_PROTOCOL *Heci = NULL;

    Status = gBS->LocateProtocol(
        &gEfiHeciProtocolGuid,
        NULL,
        (VOID **)&Heci
        );
    if (EFI_ERROR(Status)) {
        return Status;
    }

    Length = ((sizeof (PDT_UPDATE)) - (PDT_UPDATE_MAX_PAYLOAD_SIZE - PayloadSize));

    Status = Heci->GetMeMode(&MeMode);
    if (EFI_ERROR(Status) || (MeMode != ME_MODE_NORMAL)) {
        return EFI_UNSUPPORTED;
    }

    //Section below should probably be removed in order to make this tool work after EOP

    /*
    if (MeIsAfterEndOfPost ()) {
        return EFI_UNSUPPORTED;
    }*/

    PdtUpdate.Command = ISH_SET_FILE;
    PdtUpdate.MessageLength = Length - 4;

    ZeroMem(&PdtUpdate.FileName, sizeof(PdtUpdate.FileName));
    CopyMem(&PdtUpdate.FileName, FileName, sizeof(PdtUpdate.FileName));

    ZeroMem(&PdtUpdate.Data, sizeof(PdtUpdate.Data));
    CopyMem(&PdtUpdate.Data, PointerToPdtData, PayloadSize);

    Status = Heci->SendMsg(
        0,
        (UINT32 *)&PdtUpdate,
        Length,
        BIOS_FIXED_HOST_ADDR,
        HECI_ISH_SERVICE_MESSAGE_ADDR
        );

    return Status;
}

EFI_STATUS
    HeciPdt(
    IN  CHAR8                *buffer,
    IN  UINT32        bufferLength
    )
{
    static const CHAR8   FileName[] = { 'b', 'i', 'o', 's', '2', 'i', 's', 'h', '\0' };
    EFI_STATUS           Status = EFI_SUCCESS;
    PDT_UPDATE_DATA      pdt_update_data = {0};
    UINT16               pdt_update_data_length = 0;
    EFI_HECI_PROTOCOL                             *Heci = NULL;

    Status = gBS->LocateProtocol(
        &gEfiHeciProtocolGuid,
        NULL,
        (VOID **)&Heci
        );
    if (EFI_ERROR(Status)) {
        return Status;
    }
    //SetMem(pdt, PDT_UPDATE_DATA_SIZE, 'A');
    pdt_update_data.NumberOfElements = 1;
    pdt_update_data.DataType = PDT_UPDATE_DATA_TYPE;
    pdt_update_data.PayloadSize = (UINT16)bufferLength;

    //CopyMem (pdt_update_data.Payload, pdt, PDT_UPDATE_DATA_SIZE);
    CopyMem(pdt_update_data.Payload, buffer, bufferLength);

    //pdt_update_data_length = ((sizeof (PDT_UPDATE_DATA)) - (PDT_SERVICE_PAYLOAD_SIZE - PDT_UPDATE_DATA_SIZE)) ;
    pdt_update_data_length = (UINT16)((sizeof (PDT_UPDATE_DATA)) - (PDT_SERVICE_PAYLOAD_SIZE - bufferLength));

    Status = HeciPdtUpdateMsg1(FileName,
        (UINT32*)&pdt_update_data,
        pdt_update_data_length);
    if (EFI_ERROR(Status)) {
        DebugPrint(L"Send Failed \n");
    }
    else {
        DebugPrint(L"Send Succeeded \n");
    }

    return EFI_SUCCESS;
}

/*++
********************************************************************************
**
** FUNCTION:
**   GetPlatformType
**
** DESCRIPTION:
**   This function return platform details like cpuType, chipset and etc.
**
** ARGUMENTS:
**        [IN] platformType
**
** RETURNS:
**        status
**
********************************************************************************
--*/
/*
static UINT32 GetPlatformType(ME_PLATFORM_TYPE *platformType)
{
    EFI_STATUS   Status;
    FWU_GET_INFO_MSG msg;
    UINT8 Buffer[1024];
    //UCHAR *rxBuff = NULL;
    //UINT32 bufSize;
    //UINT32 dwReplyBufferSize;
    //ME_CLIENT_PROPERTIES   *MeClient;
    FWU_GET_PLATFORM_TYPE_MSG_REPLY *repMsg = NULL;
    UINT32 fwuError = 0;
    DebugPrint(L"Entered GePlatformType \n");
    // Initialize HECI interface
    / * if (!fwuclt.initialized)
    {
        if (!_getParamsFromHECI())
        {
            return g_fwuError;
        }
    }

    bufSize = GetBufferSize(&fwuclt);

    if (sizeof(FWU_GET_PLATFORM_TYPE_MSG_REPLY) > bufSize)
    {
        g_fwuError = FWU_IME_SMALL_BUFF;
        goto End;
    }

    rxBuff = (UCHAR *)AllocateMemory(bufSize);
    if (NULL == rxBuff)
    {
        g_fwuError = FWU_NO_MEMORY;
        goto End;
    } * /

    ZeroMem(&msg,sizeof(msg));
    msg.MessageType = FWU_GET_PLATFORM_TYPE;
    //dwReplyBufferSize = sizeof(FWU_GET_PLATFORM_TYPE_MSG_REPLY);

    / * g_fwuError = _sendHeciCmd(&fwuclt, &msg, sizeof(msg), rxBuff, &dwReplyBufferSize, HECI_IO_TIMEOUT, FWU_ERROR_GET_PLATFORM_TYPE);
    if(g_fwuError != SUCCESS)
    {
        goto End;
    } * /
    Status = SendMessageWithFlow (&MeClients[FWU_CLIENT],(UINT8*)&msg,sizeof(msg));
    //Status = SendMessageWithFlow (&MeClients[FWU_CLIENT], (UINT8 *)&msg, sizeof(FWU_GET_PLATFORM_TYPE_MSG_REPLY));

    if (EFI_ERROR (Status))
    {
        DebugPrint(L"SendmessageWith flow Status: %r\n", Status);
        return Status;
    }

    DebugPrint(L"Sent FWU_GET_PLATFORM_TYPE - %r\n", Status);

    Status = ReadMessageWithFlow (&MeClients[FWU_CLIENT], (UINT8 *)Buffer, MeClients[FWU_CLIENT].MaxMessageLength);
    if (EFI_ERROR (Status))
    {
        DebugPrint(L"ReadMessageWithFlow :%r\n",Status);
        return Status;
    }
    repMsg = (FWU_GET_PLATFORM_TYPE_MSG_REPLY *)Buffer;

    if (FWU_GET_PLATFORM_TYPE_REPLY != repMsg->MessageType)
    {
        fwuError = FWU_IME_UN_SUP_MESS;
        DebugPrint(L"8722");
        //goto End;
    }

    if(FWU_STATUS_SUCCESS != repMsg->Status)
    {
        fwuError = FWU_ERROR_GET_PLATFORM_TYPE;
        DebugPrint(L"8968");
        //goto End;
    }
    CopyMem(&platformType->Data, &repMsg->PlatformType, sizeof(UINT32));

    / * End:
    Status = DisConnectFWU();
    if (EFI_ERROR (Status))
    {
        DEBUG ((D_ERROR, "Cannot connect to clients - %r\n", Status));
        //DebugPrint(L"Cannot connect to clients - %r\n", Status);
        return Status;
    } * /
    return fwuError;
}
*/

/**++
********************************************************************************
*
** FUNCTION:
**   GetFwType
**
** DESCRIPTION:
**   This function retrieve the FW type
**
** ARGUMENTS:
**    fwType - UINT32 from the user to be populated by the function
**    fwType can get 4 possible values:
**      0    for 1.5M SKU image (consumer)
**      1    for 5M SKU image (corporate)
**      2    reserved
**      3    Unknown SKU
** Returns 0 on success
**
********************************************************************************
--*/
/*
UINT32 GetFwType(UINT32 *fwType)
{
    UINT32 status;
    ME_PLATFORM_TYPE platformType = {0};

    if(!fwType)
    {
        return ERROR_NULL_CHECK;
    }

    status = GetPlatformType(&platformType);
    if(SUCCESS != status)
    {
        return status;
    }

    switch (platformType.Fields.ImageType)
    {
    case IMAGE_TYPE_FULL_SKU:
        *fwType = SKU_5_MB;
        break;
    case IMAGE_TYPE_SMALL_SKU:
        *fwType = SKU_1_5_MB;
        break;
    case IMAGE_TYPE_SLIM_SKU:
    case IMAGE_TYPE_TEST_SKU:
        *fwType = SKU_RESERVED;
        break;
    case IMAGE_TYPE_NO_FW:
    default:
        *fwType = SKU_INVALID;
        break;
    }

    return SUCCESS;
}
*/

/**++
********************************************************************************
*
** FUNCTION:
**   GetPchSKU
**
** DESCRIPTION:
**   This function retrieve the platform SKU.
**
** ARGUMENTS:
**    sku    - UINT32 from the user to be populated by the function
**    sku can get 3 possible values:
**      0    H
**      1    LP
**      2    Unknown SKU
** Returns 0 on success
**
********************************************************************************
--*/
/*
UINT32 GetPchSKU(UINT32 *sku)
{
    UINT32 fwuError;
    ME_PLATFORM_TYPE platformType = {0};

    if(!sku)
    {
        return ERROR_NULL_CHECK;
    }

    fwuError = GetPlatformType(&platformType);
    if (FWU_ERROR_SUCCESS != fwuError)
    {
        return fwuError;
    }

    switch (platformType.Fields.Chipset)
    {
    case SPT_LP_CHIPSET:
        *sku = PCH_SKU_LP;
        break;
    case SPT_CHIPSET:
    case LBG_CHIPSET:
    case KBP_H_CHIPSET:
        *sku = PCH_SKU_H;
        break;
    default:
        *sku = PCH_SKU_INVALID;
        break;
    }

    return FWU_ERROR_SUCCESS;
}
*/
